package connections

import (
	"context"
	"encoding/json"
	"fmt"
	"os"

	"github.com/VictoriaMetrics/fastcache"
	"gitlab.mashplace.com/mashplace/backend/common/generic/env"
	"gitlab.mashplace.com/mashplace/backend/common/generic/grpc"
	document "gitlab.mashplace.com/mashplace/backend/document/api"
	googleGrpc "google.golang.org/grpc"
)

const (
	ThumbDir = "/thumb"
	WebpDir  = "/webp"
	InfoDir  = "/info"
)

func init() {
	err := os.MkdirAll(MediaCacheDir+InfoDir, 0777)
	if err != nil {
		panic(err)
	}

	err = os.MkdirAll(MediaCacheDir+WebpDir, 0777)
	if err != nil {
		panic(err)
	}

	err = os.MkdirAll(MediaCacheDir+ThumbDir, 0777)
	if err != nil {
		panic(err)
	}
}

var MediaCacheDir = env.ResolveString("MEDIA_CACHE_DIR", "/media")
var MediaInfoCacheSize = env.ResolveInt("MEDIA_INFO_CACHE_SIZE", 1024*1024*128)

type documentServiceCache struct {
	document.DocumentServiceClient
	fileinfoCache *fastcache.Cache
}

var DocumentServiceURL = env.ResolveString("DOCUMENT_SERVICE_URL", "mashroom-document:10000")
var DocumentService = &documentServiceCache{
	DocumentServiceClient: document.NewDocumentServiceClient(grpc.Dial(DocumentServiceURL, "Document")),
	fileinfoCache:         fastcache.New(MediaInfoCacheSize),
}

func (cache *documentServiceCache) Get(
	ctx context.Context,
	req *document.GetRequest,
	opts ...googleGrpc.CallOption,
) (*document.GetResponse, error) {
	res, err := cacheMethod(ctx, "document", req, func() (interface{}, error) {
		return cache.DocumentServiceClient.Get(ctx, req, opts...)
	})
	if err != nil {
		return nil, err
	}

	v, _ := res.(*document.GetResponse)

	return v, nil
}

func (cache *documentServiceCache) FileInfo(
	ctx context.Context,
	in *document.FileInfoRequest,
	opts ...googleGrpc.CallOption,
) (resp *document.FileInfoResponse, err error) {
	bid := in.GetId()
	if in.GetVariant() != nil {
		bid = document.EncodeVariant(bid, in.GetVariant())
	}

	rid := fmt.Sprintf("%s-%d", bid, in.GetSize())

	path := MediaCacheDir + InfoDir + "/" + rid

	id := ([]byte)(rid)

	bs := cache.fileinfoCache.Get(nil, id)
	if len(bs) > 0 {
		finfo := &document.FileInfoResponse{}

		err = json.Unmarshal(bs, finfo)
		if err == nil {
			return finfo, nil
		}
	}

	bs, _ = os.ReadFile(path)
	if len(bs) > 0 {
		resp = &document.FileInfoResponse{}

		err = json.Unmarshal(bs, resp)
		if err == nil {
			cache.fileinfoCache.Set(id, bs)

			return resp, nil
		}
	}

	resp, err = cache.DocumentServiceClient.FileInfo(ctx, in, opts...)
	if err != nil {
		return nil, err
	}

	bs, _ = json.Marshal(resp)
	if len(bs) > 0 {
		cache.fileinfoCache.Set(id, bs)

		_ = os.WriteFile(path, bs, 0666)
	}

	return resp, nil
}
