package impl

import (
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"net/http"
	"os"
	"path/filepath"
	"strconv"
	"strings"

	auth "gitlab.mashplace.com/mashplace/backend/auth/api"

	"gitlab.mashplace.com/mashplace/backend/common/generic/env"

	"github.com/h2non/filetype"
	document "gitlab.mashplace.com/mashplace/backend/document/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
)

var envMediaLimits = env.ResolveString("MEDIA_LIMITS", "{}")

func init() {
	err := json.Unmarshal([]byte(envMediaLimits), &mediaLimits)
	if err != nil {
		panic(err)
	}

	maxMultipartMemory = mediaLimits.Default.Bytes

	for _, m := range mediaLimits.MIME {
		if m.Bytes > maxMultipartMemory {
			maxMultipartMemory = m.Bytes
		}
	}
}

type mediaLimit struct {
	Bytes int64 `json:"bytes"`
}

type mediaLimitConfig struct {
	Default mediaLimit            `json:"default"`
	MIME    map[string]mediaLimit `json:"mime"`
}

var maxMultipartMemory int64

var mediaLimits = mediaLimitConfig{
	Default: mediaLimit{
		Bytes: 1024 * 1024 * 100,
	},
}

func extractMediaAttr(request *http.Request) *document.FileAttribute {
	token := request.Header.Get("Authorization")
	if strings.HasPrefix(strings.ToLower(token), "bearer ") {
		token = token[7:]
	}

	if token == "" {
		c, _ := request.Cookie("accessToken")
		if c != nil {
			token = c.Value
		}
	}

	authresp, _ := connections.AuthService.Auth(request.Context(), &auth.AuthRequest{
		AccessToken: token,
		AdminSite:   strings.EqualFold(request.Host, AdminSiteURL),
	})

	attr := &document.FileAttribute{
		User:   authresp.GetAccount().GetIdValue(),
		Remote: extractRemote(request),
		Data:   make(map[string]string),
	}

	for k, vs := range request.Form {
		if len(vs) == 0 || len(vs[0]) == 0 {
			continue
		}

		attr.Data[k] = vs[0]
	}

	return attr
}

// MediaUploader uploads file to a document service
func MediaUploader(writer http.ResponseWriter, request *http.Request) {
	if request.Method == http.MethodOptions {
		writer.WriteHeader(200)
		return
	}

	err := request.ParseMultipartForm(maxMultipartMemory + 1024*1024*1)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusBadRequest)
		return
	}

	file, header, err := request.FormFile("file")
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	defer func() {
		_ = file.Close()
	}()

	mime := header.Header.Get("Content-Type")

	size, err := file.Seek(0, io.SeekEnd)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	_, err = file.Seek(0, io.SeekStart)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	mtype, _ := filetype.MatchReader(file)

	limit, ok := mediaLimits.MIME[mtype.MIME.Value]
	if !ok {
		limit = mediaLimits.Default
	}

	if limit.Bytes < size {
		http.Error(writer, "File is too large", http.StatusRequestEntityTooLarge)
		return
	}

	_, err = file.Seek(0, io.SeekStart)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	if size != header.Size {
		http.Error(writer, "Invalid file size after transfer", http.StatusBadRequest)
		return
	}

	attr := extractMediaAttr(request)

	resp, err := document.UploadFile(request.Context(), connections.DocumentService, header.Filename, mime, attr, file)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	bs, err := json.Marshal(resp)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	writer.Header().Set("Content-Type", "application/json")
	_, err = writer.Write(bs)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
}

func MediaUpdater(writer http.ResponseWriter, request *http.Request) {
	id := strings.Replace(request.URL.Path, "/media/", "", 1)
	id = strings.TrimSuffix(id, filepath.Ext(id))

	if len(id) < 64 {
		http.Error(writer, "invalid id format", http.StatusBadRequest)
		return
	}

	err := request.ParseMultipartForm(maxMultipartMemory)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusBadRequest)
		return
	}

	file := &document.File{
		Id:   id,
		Name: request.FormValue("name"),
	}

	attr := extractMediaAttr(request)

	resp, err := connections.DocumentService.FileUpdate(request.Context(), &document.FileUpdateRequest{
		File:      file,
		Attribute: attr,
	})

	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	writer.Header().Set("Content-Type", "application/json")

	bs, err := json.Marshal(resp.File)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	_, err = writer.Write(bs)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	return
}

func MediaID(writer http.ResponseWriter, request *http.Request) {
	id := strings.Replace(request.URL.Path, "/media/", "", 1)
	if len(id) == 0 {
		MediaUploader(writer, request)
		return
	}
	switch request.Method {
	case http.MethodGet:
		MediaDownloader(writer, request)
	case http.MethodPatch:
		MediaUpdater(writer, request)
	case http.MethodOptions:
		writer.WriteHeader(200)
	}
}

// MediaDownloader downloads a media file to local cache
func MediaDownloader(writer http.ResponseWriter, request *http.Request) {
	id := strings.Replace(request.URL.Path, "/media/", "", 1)

	var path string

	var infoReq bool

	cacheID := strings.TrimSuffix(id, filepath.Ext(id))

	if strings.Index(id, "info-") == 0 {
		id = strings.Replace(id, "info-", "", 1)
		cacheID = strings.TrimSuffix(id, filepath.Ext(id))
		infoReq = true
	}

	imageSize := document.ImageSize_ORIGINAL
	if strings.Index(id, "thumb-") == 0 {
		id = strings.Replace(id, "thumb-", "", 1)
		imageSize = document.ImageSize_THUMBNAIL
	}

	if strings.Index(id, "webp-") == 0 {
		id = strings.Replace(id, "webp-", "", 1)
		imageSize = document.ImageSize_WEBP
	}

	id = strings.TrimSuffix(id, filepath.Ext(id))
	if len(id) < 64 {
		http.Error(writer, "invalid id format", http.StatusBadRequest)
		return
	}

	err := updateDraftContract(request.Context(), id)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	info, err := connections.DocumentService.FileInfo(request.Context(), &document.FileInfoRequest{
		Id:   id,
		Size: imageSize,
	})
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	if infoReq {
		writer.Header().Set("Content-Type", "application/json")
		var bs []byte
		bs, err = json.Marshal(info.File)
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
		_, err = writer.Write(bs)
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
		return
	}

	path = connections.MediaCacheDir

	switch imageSize {
	case document.ImageSize_THUMBNAIL:
		path = connections.MediaCacheDir + connections.ThumbDir + ID2Filepath(id)
	case document.ImageSize_WEBP:
		path = connections.MediaCacheDir + connections.WebpDir + ID2Filepath(id)
	default:
		path = connections.MediaCacheDir + ID2Filepath(cacheID)
	}

	_, err = os.Stat(path)

	if os.IsNotExist(err) {
		idx := strings.LastIndex(path, "/")

		err = os.MkdirAll(path[:idx], os.FileMode(0755))
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}

		var f *os.File

		f, err = os.OpenFile(path, os.O_RDWR|os.O_CREATE, os.FileMode(0644))
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}

		defer func() {
			_ = f.Close()
			if err != nil {
				_ = os.Remove(path)
			}
		}()

		_, _, err = document.DownloadFile(request.Context(), connections.DocumentService, id, imageSize, f)
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
	}

	writer.Header().Set("Content-Type", info.File.Mime)
	http.ServeFile(writer, request, path)
}

func MediaBundle(writer http.ResponseWriter, request *http.Request) {
	if !strings.EqualFold(request.Host, AdminSiteURL) {
		writer.WriteHeader(http.StatusForbidden)
		return
	}

	property := request.URL.Query().Get("property")

	if property == "" {
		writer.WriteHeader(http.StatusBadRequest)
		return
	}

	client, err := connections.DocumentService.BundleDownload(request.Context(), &document.BundleDownloadRequest{
		Property:  property,
		TypeCodes: []string{document.TypePhoto},
	})
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	ev, err := client.Recv()
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	if ev.Init == nil {
		http.Error(writer, "missing init", http.StatusInternalServerError)
		return
	}

	disposition := fmt.Sprintf("attachment; filename=\"%s\"", ev.GetInit().GetFile().GetName())

	writer.Header().Set("Content-Type", ev.GetInit().GetFile().GetMime())
	writer.Header().Set("Content-Disposition", disposition)
	writer.Header().Set("Content-Length", strconv.FormatInt(ev.GetInit().GetFile().GetSize(), 10))

	writer.WriteHeader(http.StatusOK)

	for {
		ev, err := client.Recv()
		if errors.Is(err, io.EOF) {
			break
		}

		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}

		if ev.Data == nil {
			http.Error(writer, "missing data", http.StatusInternalServerError)
			return
		}

		_, err = writer.Write(ev.Data.Data)
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
	}
}
