package resolvers

import (
	"context"
	"reflect"
	"time"

	"gitlab.mashplace.com/mashplace/backend/offer/api/rto"

	"github.com/graphql-go/graphql"
	account "gitlab.mashplace.com/mashplace/backend/account/api"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	comments "gitlab.mashplace.com/mashplace/backend/comments/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	document "gitlab.mashplace.com/mashplace/backend/document/api"
	expenses "gitlab.mashplace.com/mashplace/backend/expenses/api"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/enrichment"
	insurance "gitlab.mashplace.com/mashplace/backend/insurance/api"
	invitation "gitlab.mashplace.com/mashplace/backend/invitation/api"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	maintenance "gitlab.mashplace.com/mashplace/backend/maintenance/api"
	mortgage "gitlab.mashplace.com/mashplace/backend/mortgage/api"
	notes "gitlab.mashplace.com/mashplace/backend/notes/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	viewing "gitlab.mashplace.com/mashplace/backend/viewing/api"
)

var correlationTypeOffer = "offer"

type scheduledRecordsTransaction struct {
	ScheduledRecords []*expenses.ScheduledRecord
}

func initProductEnrichment() {
	util.BindOutputName(&extras.Service{}, "", util.EnrichmentMap{
		"learn_more": &util.Enrichment{
			DeprecationReason: "Use description",
			OutputNullable:    true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				svc, ok := p.Source.(*extras.Service)
				if !ok || svc.GetDescription() == nil {
					return nil, nil
				}

				return svc.GetDescriptionValue(), nil
			},
		},
	}, "extras_api_Service")

	util.BindOutputName(&extras.ServiceSnapshot{}, "", util.EnrichmentMap{
		"learn_more": &util.Enrichment{
			DeprecationReason: "Use description",
			OutputNullable:    false,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				svc, ok := p.Source.(*extras.ServiceSnapshot)
				if !ok {
					return "", nil
				}

				return svc.GetDescription(), nil
			},
		},
	}, "extras_api_ServiceSnapshot")
}

//nolint:gocyclo, gocognit // this func can have high cyclomatic and cognitive complexity
func initAccountEnrichment() {
	util.BindOutput(&account.AccountDetails{}, "", util.EnrichmentMap{
		"avatar": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&document.File{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				acc, ok := p.Source.(*account.AccountDetails)
				if !ok || acc.GetAvatarValue() == "" {
					return &document.File{}, nil
				}
				resp, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{
					Id: acc.GetAvatarValue(),
				})
				if err != nil {
					return &document.File{}, nil
				}
				return resp.File, err
			},
		},
	})

	util.BindOutput(&account.Account{}, "", util.EnrichmentMap{
		"id": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				acc, ok := extractAccount(p.Source)
				if !ok {
					return "", nil
				}

				return acc.GetIdValue(), nil
			},
		},
		"verification_info": &util.Enrichment{
			OutputNullable: false,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.GetVerificationResponse{}, "status && veriff sessions", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				acc, ok := extractAccount(p.Source)
				if !ok || acc.GetIdValue() == "" {
					return &account.GetVerificationResponse{}, nil
				}

				params := &account.GetVerificationRequest{
					AccountId: acc.GetIdValue(),
					IsAdmin:   acc.GetAdminValue(),
				}

				verificationInfo, err := connections.AccountService.GetVerificationInfo(p.Context, params)
				if err != nil {
					return nil, err
				}

				result := &account.GetVerificationResponse{
					IsVerified: verificationInfo.GetIsVerified(),
					Sessions:   verificationInfo.GetSessions(),
				}

				return result, nil
			},
		},
		"documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				acc, ok := extractAccount(p.Source)
				if !ok || acc.GetIdValue() == "" {
					return []*document.Document{}, nil
				}
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return []*document.Document{}, nil
				}
				switch {
				case acc.GetIdValue() == user.GetAccount().GetIdValue():
				default:
					return []*document.Document{}, nil
				}

				resp, err := connections.DocumentService.List(p.Context, &document.ListRequest{
					User: acc.GetIdValue(),
				})
				if err != nil {
					return []*document.Document{}, nil
				}
				var docs []*document.Document
				for _, d := range resp.Documents {
					if d.TypeCode == document.TypeIdentity {
						docs = append(docs, d)
					}
				}
				return docs, err
			},
		},
		"resources": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.AccountResources{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				acc, ok := extractAccount(p.Source)
				if !ok || acc.GetIdValue() == "" {
					return nil, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || (!user.GetAccount().GetAdminValue() && user.GetAccount().GetIdValue() != acc.GetIdValue()) {
					return nil, nil
				}

				enrichment.ResolveAccountResources(p.Context, acc)

				return acc.GetResourcesValue(), nil
			},
		},
	})
}

func resolveTransactionScheduledRecords(
	ctx context.Context,
	t *invoice.Transaction,
) ([]*expenses.ScheduledRecord, error) {
	if t.GetReconciliation() != invoice.TransactionReconciliation_TRANSACTION_RECONCILIATION_FULL {
		return nil, nil
	}

	resp, err := connections.ExpensesService.ScheduledRecordList(ctx, &expenses.ScheduledRecordListRequest{
		PayoutOptions: &expenses.ScheduledRecordListPayoutOptions{
			TransactionId: t.GetId(),
		},
		Type: expenses.RecordType_RECORD_TYPE_INCOME,
	})
	if err != nil {
		return nil, err
	}

	return resp.GetScheduledRecords(), nil
}

func initTransactionEnrichment() {
	util.BindOutputName(&invoice.TransactionPaymentMatch{}, "", util.EnrichmentMap{
		"payment": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&invoice.Payment{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*invoice.TransactionPaymentMatch)
				if !ok {
					return nil, nil
				}

				resp, err := connections.InvoiceService.PaymentGet(p.Context, &invoice.PaymentGetRequest{
					PaymentId: req.GetPaymentId(),
				})
				if err != nil {
					return nil, nil
				}

				return resp.GetPayment(), nil
			},
		},
		"transaction": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&invoice.Transaction{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*invoice.TransactionPaymentMatch)
				if !ok {
					return nil, nil
				}

				resp, err := connections.InvoiceService.TransactionGet(p.Context, &invoice.TransactionGetRequest{
					TransactionId: req.GetTransactionId(),
				})
				if err != nil {
					return nil, nil
				}

				return resp.GetTransaction(), nil
			},
		},
	}, "invoice_api_TransactionPaymentMatch")

	util.BindOutputName(&invoice.Transaction{}, "", util.EnrichmentMap{
		"offer_id": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				t, ok := p.Source.(*invoice.Transaction)
				if !ok {
					return "", nil
				}

				scheduledRecords, err := resolveTransactionScheduledRecords(p.Context, t)
				if err != nil {
					return "", nil
				}

				for _, sr := range scheduledRecords {
					if sr.GetRecord().GetCorrelationType() == expenses.CorrelationType_CORRELATION_TYPE_OFFER {
						return sr.GetRecord().GetCorrelationId(), nil
					}
				}

				return "", nil
			},
		},
		"scheduled_records": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&scheduledRecordsTransaction{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				t, ok := p.Source.(*invoice.Transaction)
				if !ok {
					return nil, nil
				}

				scheduledRecords, err := resolveTransactionScheduledRecords(p.Context, t)
				if err != nil {
					return nil, nil
				}

				resp := &scheduledRecordsTransaction{
					ScheduledRecords: scheduledRecords,
				}

				return resp, nil
			},
		},
	}, "invoice_api_Transaction")
}

//nolint:gocognit,gocyclo // enrichment initialization. maybe extract common enrichment handlers.
func initOrderEnrichment() {
	util.BindOutputName(invoice.ProductDocument{}, "", util.EnrichmentMap{
		"file": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&document.File{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				doc, ok := p.Source.(*invoice.ProductDocument)
				if !ok {
					return nil, nil
				}

				if doc.GetFileid() == "" {
					return nil, nil
				}

				resp, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{
					Id: doc.GetFileid(),
				})
				if err != nil {
					return nil, nil
				}

				if resp.GetFile().GetId() == "" {
					return nil, nil
				}

				return resp.GetFile(), nil
			},
		},
	}, "invoice_api_ProductDocument")

	util.BindOutputName(invoice.ProductDetailsPerformance{}, "", util.EnrichmentMap{
		"desired_performance_date": &util.Enrichment{
			DeprecationReason: "Use time.start",
			OutputNullable:    false,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				acc, ok := p.Source.(*invoice.ProductDetailsPerformance)
				if !ok {
					return nil, nil
				}

				return acc.GetDesiredPerformanceDate().AsTime().Format(time.RFC1123), nil
			},
		},
	}, "invoice_api_ProductDetailsPerformance")

	util.BindOutputName(invoice.ProductDetails{}, "", util.EnrichmentMap{
		"performance": &util.Enrichment{
			OutputNullable: false,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&invoice.ProductDetailsPerformance{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				acc, ok := p.Source.(*invoice.ProductDetails)
				if !ok {
					return nil, nil
				}

				if acc.Performance == nil {
					acc.Performance = &invoice.ProductDetailsPerformance{}
				}

				acc.Performance.DesiredPerformanceDate = acc.GetTime().GetStart()

				return acc.GetPerformance(), nil
			},
		},
	}, "invoice_api_ProductDetails")

	util.BindOutputName(invoice.Payment{}, "", util.EnrichmentMap{
		"invoice": &util.Enrichment{
			Fields: nil,
			Input:  nil,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(invoice.Invoice{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ps, ok := p.Source.(*invoice.Payment)
				if !ok {
					return nil, nil
				}

				resp, err := connections.InvoiceService.InvoiceGet(p.Context, &invoice.InvoiceGetRequest{
					InvoiceId: ps.GetInvoiceId(),
				})
				if err != nil {
					return nil, err
				}

				return resp.GetInvoice(), nil
			},
			OutputNullable: true,
		},
		"property": &util.Enrichment{
			Fields: nil,
			Input:  nil,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(property.Property{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ps, ok := p.Source.(*invoice.Payment)
				if !ok {
					return nil, nil
				}

				if ps.GetProperty() == "" {
					return nil, nil
				}

				resp, err := connections.PropertyService.Get(p.Context, &property.GetRequest{
					Id: ps.GetProperty(),
				})
				if err != nil {
					return nil, err
				}

				return resp.GetProperty(), nil
			},
			OutputNullable: true,
		},
		"overdue_payment_notice_count": &util.Enrichment{
			Fields: nil,
			Input:  nil,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.Int
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				// Will be finished in other task - for now we don't have a logic and notifications which should be
				// counted. It's required to have got this field now, just not to block FE.

				return 0, nil
			},
			OutputNullable: true,
		},
		"user_data": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ps, ok := p.Source.(*invoice.Payment)
				if !ok {
					return &account.Account{}, nil
				}

				acc, err := connections.AccountService.Get(
					p.Context,
					&account.GetRequest{
						Id: ps.GetUser(),
					},
				)
				if err != nil {
					log.WithError(err).Error()
					return &account.Account{}, nil
				}

				return CensoredAccount(acc.GetAccount()), nil
			},
			OutputNullable: true,
		},
	}, "invoice_api_Payment")

	util.BindOutputName(invoice.Product{}, "", util.EnrichmentMap{
		"service": &util.Enrichment{
			Fields: nil,
			Input:  nil,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(extras.ServiceSnapshotStatus{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ps, ok := oldvalue.(*extras.ServiceSnapshotStatus)
				if !ok {
					return nil, nil
				}

				return ps, nil
			},
			OutputNullable: true,
		},
	}, "invoice_api_Product")

	util.BindOutputName(orderSaveRequestProduct{}, "", util.EnrichmentMap{}, "invoice_api_Product")

	util.BindOutputName(invoice.Order{}, "", util.EnrichmentMap{
		"invoices": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(
					util.BindOutput(invoice.Invoice{}, "", nil),
				))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				var invoices []*invoice.Invoice

				order, ok := p.Source.(*invoice.Order)
				if !ok {
					return invoices, nil
				}

				resp, err := connections.InvoiceService.InvoiceList(p.Context, &invoice.InvoiceListRequest{
					OrderId: order.GetId(),
				})
				if err != nil {
					return invoices, nil
				}

				return resp.GetInvoices(), nil
			},
		},
		"user_data": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ps, ok := p.Source.(*invoice.Order)
				if !ok {
					return &account.Account{}, nil
				}

				acc, err := connections.AccountService.Get(
					p.Context,
					&account.GetRequest{
						Id: ps.GetUserid(),
					},
				)
				if err != nil {
					log.WithError(err).Error()
					return &account.Account{}, nil
				}

				return CensoredAccount(acc.GetAccount()), nil
			},
			OutputNullable: true,
		},
	}, "invoice_api_Order")
}

func containsTenant(tenants []*property.Tenant, id string) bool {
	for _, t := range tenants {
		if t.GetUser().GetValue() == id {
			return true
		}
	}

	return false
}

func propertyRelation(people *property.PropertyPeople, id string) (landlord, related bool) {
	landlord = people.GetLandlord().GetValue() == id

	related = landlord ||
		people.GetSubmitter().GetValue() == id ||
		people.GetManager().GetValue() == id ||
		containsTenant(people.GetTenants().GetValue(), id)

	return
}

var relatedDocumentTypes = []string{
	document.TypeContract,
	document.TypeInventory,
	document.TypeCertificate,
}

func propertyDocumentTypes(landlord, related bool, reqtypes []string) (types []string) {
	switch {
	case landlord:
		types = reqtypes
	case related:
		for _, r := range reqtypes {
			for _, d := range relatedDocumentTypes {
				if r == d {
					types = append(types, r)

					break
				}
			}
		}
	}

	return
}

type propertyFinances struct {
	Invoices  []*invoice.Invoice                `json:"invoices"`
	Income    []*invoice.Invoice                `json:"income"`
	Scheduled []*invoice.Invoice                `json:"scheduled"`
	Mortgage  *property.MortgageEntryArrayValue `json:"mortgage"`
}

//nolint:gocyclo, gocognit // this func can have high cyclomatic and cognitive complexity
func initViewingEnrichment() {
	util.BindOutputName(&viewing.ViewingDetails{}, "Property-related details", util.EnrichmentMap{
		"flatmates": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&account.Account{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*viewing.ViewingDetails)
				if !ok {
					return []*account.Account{}, nil
				}

				if len(req.GetFlatmates()) == 0 {
					return []*account.Account{}, nil
				}

				var list []*account.Account

				for _, f := range req.GetFlatmates() {
					acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: f})
					if err != nil {
						log.WithError(err).Error()
						continue
					}

					list = append(list, acc.GetAccount())
				}

				return list, nil
			},
		},
	}, "viewing_api_ViewingDetails")

	util.BindOutputName(&viewing.Viewing{}, "Viewing request", util.EnrichmentMap{
		"property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return &property.Property{}, nil
				}

				req, ok := p.Source.(*viewing.Viewing)
				if !ok {
					return &property.Property{}, nil
				}

				if req.GetId() == "" {
					return &property.Property{}, nil
				}

				test := user.GetAccount().GetTestValue()

				resp, err := connections.PropertyService.Get(
					p.Context,
					&property.GetRequest{Id: req.GetProperty(), User: user.GetAccount().GetIdValue(), Test: test},
				)
				if err != nil {
					return nil, err
				}

				return resp.GetProperty(), nil
			},
		},
		"flatmates": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&account.Account{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*viewing.ViewingDetails)
				if !ok {
					return []*account.Account{}, nil
				}

				if len(req.GetFlatmates()) == 0 {
					return []*account.Account{}, nil
				}

				var list []*account.Account

				for _, f := range req.GetFlatmates() {
					acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: f})
					if err != nil {
						log.WithError(err).Error()
						continue
					}

					list = append(list, acc.GetAccount())
				}

				return list, nil
			},
		},
		"offer": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&offer.Offer{}, "offer", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*viewing.Viewing)
				if !ok {
					return &offer.Offer{}, nil
				}

				if req.GetSender() == "" {
					return &offer.Offer{}, nil
				}

				resp, err := connections.OfferService.Get(p.Context, &offer.GetRequest{Id: req.GetOffer()})
				if err != nil {
					return &offer.Offer{}, nil
				}

				return resp.GetOffer(), err
			},
		},
		"manager": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "manager", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*viewing.Viewing)
				if !ok {
					return &account.Account{}, nil
				}

				if req.GetManager() == "" {
					return &account.Account{}, nil
				}

				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetManager()})
				if err != nil {
					return &account.Account{}, nil
				}

				return resp.GetAccount(), err
			},
		},
	}, "viewing_api_Viewing")

	delete(util.DecomposedTypes, reflect.TypeOf(&offer.Offer{}))
	delete(util.DecomposedTypes, reflect.TypeOf(&offer.OfferConditions{}))
	delete(util.DecomposedTypes, reflect.TypeOf(&offer.Landlord{}))
	delete(util.DecomposedTypes, reflect.TypeOf(&offer.OfferDetails{}))
}

//nolint:gocyclo, gocognit // this func can have high cyclomatic and cognitive complexity
func initPropertyEnrichment() {
	resolveFileList := func(
		resolver func(desc *property.PropertyDescription) *generic.StringArrayValue,
	) *util.Enrichment {
		return &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.File{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyDescription)
				if !ok {
					return []*document.File{}, nil
				}
				arr := resolver(prop)
				if arr == nil || len(arr.GetValue()) == 0 {
					return []*document.File{}, nil
				}
				var ids []*document.File
				for _, id := range arr.GetValue() {
					resp, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{
						Id: id,
					})
					if err != nil {
						continue
					}
					if resp.GetFile().GetId() == "" {
						continue
					}

					ids = append(ids, resp.GetFile())
				}
				return ids, nil
			},
		}
	}

	util.BindOutput(&property.PropertyDescription{}, "Property-related descriptions", util.EnrichmentMap{
		"photos": resolveFileList(func(desc *property.PropertyDescription) *generic.StringArrayValue {
			return desc.GetPhotos()
		}),
		"videos": resolveFileList(func(desc *property.PropertyDescription) *generic.StringArrayValue {
			return desc.GetVideos()
		}),
		"plans": resolveFileList(func(desc *property.PropertyDescription) *generic.StringArrayValue {
			return desc.GetPlans()
		}),
	})

	util.BindOutput(&property.Tenant{}, "Property-related tenant", util.EnrichmentMap{
		"user": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.Tenant)
				if !ok || prop.GetUser().GetValue() == "" {
					return &account.Account{}, nil
				}

				acc, err := connections.AccountService.Get(
					p.Context,
					&account.GetRequest{
						Id: prop.GetUser().GetValue(),
					},
				)
				if err != nil {
					log.WithError(err).Error()
					return &account.Account{}, nil
				}
				return CensoredAccount(acc.GetAccount()), nil
			},
		},
	})

	util.BindOutput(&property.InvitationStatus{}, "", util.EnrichmentMap{
		"inviter": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				stat, ok := p.Source.(*property.InvitationStatus)
				if !ok || stat.GetInviterId() == "" {
					return nil, nil
				}

				acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: stat.GetInviterId()})
				if err != nil {
					log.WithError(err).Error()
					return nil, nil
				}

				return CensoredAccount(acc.GetAccount()), err
			},
		},
		"invitee": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				stat, ok := p.Source.(*property.InvitationStatus)
				if !ok || stat.GetInviteeId() == "" {
					return nil, nil
				}

				acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: stat.GetInviteeId()})
				if err != nil {
					log.WithError(err).Error()
					return nil, nil
				}

				return CensoredAccount(acc.GetAccount()), err
			},
		},
		"email": &util.Enrichment{
			OutputNullable:    true,
			DeprecationReason: "use invitee_email",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewNonNull(graphql.String)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				stat, ok := p.Source.(*property.InvitationStatus)
				if !ok {
					return nil, nil
				}

				return stat.GetInviteeEmail(), nil
			},
		},
		"invitee_email": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewNonNull(graphql.String)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				stat, ok := p.Source.(*property.InvitationStatus)
				if !ok {
					return nil, nil
				}

				return stat.GetInviteeEmail(), nil
			},
		},
	})

	util.BindOutput(&property.PropertyPeople{}, "Property-related people", util.EnrichmentMap{
		"landlord": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok || prop.GetLandlord().GetValue() == "" {
					return nil, nil
				}

				acc, err := connections.AccountService.Get(
					p.Context,
					&account.GetRequest{
						Id: prop.GetLandlord().GetValue(),
					},
				)
				if err != nil {
					log.WithError(err).Error()
					return nil, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil || ForeignProperty(user, prop) {
					return CensoredAccount(acc.GetAccount()), nil
				}

				return acc.GetAccount(), nil
			},
		},
		"landlord_invitation": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.InvitationStatus{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok {
					return nil, nil
				}

				inv := prop.GetLandlordInvitation()

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if inv != nil && (!ok || user == nil || ForeignInvite(user, prop, inv)) {
					inv.InviteePhone = ""
					inv.InviteeEmail = ""
				}

				return inv, nil
			},
		},
		"manager": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok || prop.GetManager().GetValue() == "" {
					return nil, nil
				}

				acc, err := connections.AccountService.Get(
					p.Context,
					&account.GetRequest{
						Id: prop.GetManager().GetValue(),
					},
				)
				if err != nil {
					log.WithError(err).Error()
					return nil, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil || ForeignProperty(user, prop) {
					return CensoredAccount(acc.GetAccount()), nil
				}

				return acc.GetAccount(), nil
			},
		},
		"manager_invitation": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.InvitationStatus{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok {
					return nil, nil
				}

				inv := prop.GetManagerInvitation()

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if inv != nil && (!ok || user == nil || ForeignInvite(user, prop, inv)) {
					inv.InviteePhone = ""
					inv.InviteeEmail = ""
				}

				return inv, nil
			},
		},
		"submitter": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok || prop.GetSubmitter().GetValue() == "" {
					return nil, nil
				}
				acc, err := connections.AccountService.Get(
					p.Context,
					&account.GetRequest{
						Id: prop.GetSubmitter().GetValue(),
					},
				)
				if err != nil {
					log.WithError(err).Error()
					return nil, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil || ForeignProperty(user, prop) {
					return CensoredAccount(acc.GetAccount()), nil
				}

				return acc.GetAccount(), nil
			},
		},
		"tenants": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&property.Tenant{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok {
					return []*property.Tenant{}, nil
				}

				return prop.GetTenants().GetValue(), nil
			},
		},
		"administrator": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok || prop.GetAdministrator().GetValue() == "" {
					return nil, nil
				}
				acc, err := connections.AccountService.Get(
					p.Context,
					&account.GetRequest{
						Id:    prop.GetAdministrator().GetValue(),
						Admin: true,
					},
				)
				if err != nil {
					log.WithError(err).Error()
					return nil, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil || ForeignProperty(user, prop) {
					return CensoredAccount(acc.GetAccount()), nil
				}

				return acc.GetAccount(), nil
			},
		},
	})

	util.BindOutput(&property.Viewing{}, "Property-related people", util.EnrichmentMap{
		"manager": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "manager", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.Viewing)
				if !ok {
					return &account.Account{}, nil
				}

				if req.GetManager() == "" {
					return &account.Account{}, nil
				}

				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetManager()})
				if err != nil {
					log.WithError(err).Error()
					return &account.Account{}, nil
				}

				return resp.GetAccount(), err
			},
		},
	})

	util.BindOutput(&property.Property{}, "Property struct", util.EnrichmentMap{
		"id": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := extractProperty(p.Source)
				if !ok {
					return "", nil
				}

				return prop.GetId(), nil
			},
		},
		"favorited": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.Boolean
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return false, nil
				}
				prop, ok := extractProperty(p.Source)
				if !ok {
					return false, nil
				}

				for _, p := range user.GetAccount().GetFavoritesValue().GetProperties().GetValue() {
					if p == prop.GetId() {
						return true, nil
					}
				}
				return false, nil
			},
		},
		"documents": &util.Enrichment{
			OutputNullable: true,
			Args:           util.BindArguments(&document.ListRequest{}),
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := extractProperty(p.Source)
				if !ok || prop.GetId() == "" {
					return []*document.Document{}, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return []*document.Document{}, nil
				}

				req := &document.ListRequest{}
				if err := util.UnmarshalArguments(p.Args, req); err != nil {
					return nil, err
				}

				landlord, related := propertyRelation(prop.GetPeople(), user.GetAccount().GetIdValue())

				switch {
				case landlord:
				case related:
					req.TypeCodes = relatedDocumentTypes
				default:
					return []*document.Document{}, nil
				}

				req.Property = prop.GetId()

				resp, err := connections.DocumentService.List(p.Context, req)
				if err != nil {
					return []*document.Document{}, nil
				}

				return resp.GetDocuments(), err
			},
		},
		"pending_extra_services": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(graphql.String))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := extractProperty(p.Source)
				if !ok || prop.GetId() == "" {
					return []string{}, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return []string{}, nil
				}

				resp, err := connections.InvoiceService.ProductList(p.Context, &invoice.ProductListRequest{
					User:     user.GetAccount().GetIdValue(),
					Property: prop.GetId(),
					OrderStatuses: []invoice.OrderStatus{
						invoice.OrderStatus_ORDER_PENDING,
						invoice.OrderStatus_ORDER_ACTIVE,
					},
				})
				if err != nil {
					return []string{}, nil
				}

				prods := make([]string, 0)

				for _, pr := range resp.GetProduct() {
					if len(pr.GetServiceId()) != 24 {
						continue
					}

					prods = append(prods, pr.GetServiceId())
				}

				return prods, nil
			},
			OutputNullable: false,
		},
		"offers_list": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&offer.Offer{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return []*offer.Offer{}, nil
				}

				prop, ok := extractProperty(p.Source)
				if !ok || prop.GetId() == "" {
					return []*offer.Offer{}, nil
				}

				if user.GetAccount().GetIdValue() == "" {
					return []*offer.Offer{}, nil
				}

				var err error

				var resp *offer.ListbyPropertyResponse

				switch user.GetAccount().GetIdValue() {
				case prop.GetPeople().GetLandlord().GetValue(),
					prop.GetPeople().GetManager().GetValue():
					resp, err = connections.OfferService.ListByPropertyLandlord(p.Context, &offer.ListRequest{
						Property: &offer.ListRequest_PropertyValue{PropertyValue: prop.GetId()},
						User:     user.GetAccount().GetIdValue(),
					})
				default:
					resp, err = connections.OfferService.ListByPropertyTenant(p.Context, &offer.ListRequest{
						Property: &offer.ListRequest_PropertyValue{PropertyValue: prop.GetId()},
						User:     user.GetAccount().GetIdValue(),
						Listing:  prop.GetListing().GetId().GetValue(),
					})
				}

				if err != nil {
					return []*offer.Offer{}, nil
				}

				return resp.Offers, nil
			},
		},
		"viewings_list": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&viewing.Viewing{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return []*viewing.Viewing{}, nil
				}

				prop, ok := extractProperty(p.Source)
				if !ok || prop.GetId() == "" {
					return []*viewing.Viewing{}, nil
				}

				req := &viewing.ViewingListRequest{
					Spec: &viewing.ViewingSpec{
						UserId: user.GetAccount().GetIdValue(),
						Select: &viewing.ViewingSpecSelect{
							PropertyId: prop.GetId(),
						},
					},
				}

				resp, err := connections.ViewingService.ViewingList(p.Context, req)
				if err != nil {
					return []*viewing.Viewing{}, nil
				}

				return resp.GetViewings(), err
			},
		},
		"status_history": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&property.PropertyStatusData{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := extractProperty(p.Source)
				if !ok || prop.GetId() == "" {
					return []*property.PropertyStatusData{}, nil
				}

				resp, err := connections.PropertyService.GetPropertyStatusHistory(
					p.Context,
					&property.PropertyStatusHistoryRequest{
						Property: prop.GetId(),
					},
				)
				if err != nil {
					return []*property.PropertyStatusData{}, nil
				}
				return resp.History, err
			},
		},
		"incoming_invitations": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewNonNull(graphql.NewList(util.BindOutput(&invitation.InvitationInfo{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := extractProperty(p.Source)
				if !ok {
					return []*invitation.InvitationInfo{}, nil
				}

				people := prop.GetPeople()

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil || (ForeignInvite(user, people, people.GetLandlordInvitation()) &&
					ForeignInvite(user, people, people.GetManagerInvitation())) {
					return []*invitation.InvitationInfo{}, nil
				}

				request := &invitation.InvitationListIncomingRequest{
					Email:      user.GetAccount().GetContactValue().GetEmailPrimaryValue(),
					PropertyId: prop.GetId(),
				}

				invs, err := connections.InvitationService.InvitationListIncoming(p.Context, request)
				if err != nil {
					return []*invitation.InvitationInfo{}, err
				}

				if invs.Invitations == nil {
					return []*invitation.InvitationInfo{}, nil
				}

				return invs.Invitations, nil
			},
		},
		"products": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(
					util.BindOutput(invoice.Product{}, "product", nil),
				))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				var products []*invoice.Product

				prop, ok := extractProperty(p.Source)
				if !ok {
					return products, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return products, nil
				}

				resp, err := connections.InvoiceService.ProductList(p.Context, &invoice.ProductListRequest{
					User:     user.GetAccount().GetIdValue(),
					Property: prop.GetId(),
					OrderStatuses: []invoice.OrderStatus{
						invoice.OrderStatus_ORDER_ACTIVE,
						invoice.OrderStatus_ORDER_PAID,
					},
				})
				if err != nil {
					return products, nil
				}

				return resp.GetProduct(), nil
			},
		},
		"maintenance": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(
					util.BindOutput(maintenance.Maintenance{}, "maintenance", nil),
				))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				var ms []*maintenance.Maintenance

				prop, ok := extractProperty(p.Source)
				if !ok {
					return ms, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return ms, nil
				}

				resp, err := connections.MaintenanceService.List(p.Context, &maintenance.ListRequest{
					User:     user.GetAccount().GetIdValue(),
					Property: prop.GetId(),
				})
				if err != nil || resp.GetMaintenance() == nil {
					return ms, nil
				}

				return resp.GetMaintenance(), nil
			},
		},
		"finances": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewNonNull(
					util.BindOutputName(
						&propertyFinances{},
						"Property-related finances",
						nil,
						"PropertyFinances",
					),
				)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				finances := propertyFinances{}

				prop, ok := extractProperty(p.Source)
				if !ok {
					return finances, nil
				}

				var userID string

				if !grpcctx.ContextIsAdminUser(p.Context) {
					var user *auth.AuthResponse

					user, ok = p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
					if !ok || user == nil {
						return finances, nil
					}

					userID = user.GetAccount().GetIdValue()
				}

				invs, err := connections.InvoiceService.InvoiceList(p.Context, &invoice.InvoiceListRequest{
					Property: prop.GetId(),
					User:     userID,
					Statuses: []invoice.InvoiceStatus{
						invoice.InvoiceStatus_INVOICE_OPEN,
						invoice.InvoiceStatus_INVOICE_FULFILLED,
					},
				})
				if err != nil {
					log.WithError(err).Errorf("fetching expense invoices for user %s", userID)
				}

				for _, inv := range invs.GetInvoices() {
					switch {
					case inv.GetType() == invoice.InvoiceType_INVOICE_TYPE_PAYOUT:
						finances.Income = append(finances.Income, inv)
					case inv.GetType() == invoice.InvoiceType_INVOICE_TYPE_PAYMENT && inv.GetPeriodic():
						finances.Scheduled = append(finances.Scheduled, inv)
					default:
						finances.Invoices = append(finances.Invoices, inv)
					}
				}

				var rent, holding *invoice.Invoice

				corr := 0

				for idx, inv := range finances.Invoices {
					switch inv.GetCorrelationTarget() {
					case "tenant-services":
						for _, s := range inv.GetServices() {
							if s.GetService().GetCode() == "first_month_rent" {
								rent = inv
								break
							}
						}
					case "holding-deposit":
						for _, s := range inv.GetServices() {
							if s.GetService().GetCode() == "holding_deposit" {
								holding = inv
								finances.Invoices = append(
									finances.Invoices[:idx-corr],
									finances.Invoices[idx+1-corr:]...,
								)
								corr++

								break
							}
						}
					}
				}

				if rent != nil {
					rent.Amount += holding.GetAmount()

					for _, s := range rent.GetServices() {
						if s.GetService().GetCode() == "first_month_rent" {
							s.GetService().Price += holding.GetAmount()
							break
						}
					}
				}

				if prop.GetPeople().GetLandlord().GetValue() == userID || userID == "" {
					finances.Mortgage = prop.GetFinances().GetMortgage()
				}

				return finances, nil
			},
		},
	})

	util.BindOutput(&viewing.ViewingFeedbackEntry{}, "Property-related details entry", util.EnrichmentMap{
		"initiator": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*viewing.ViewingFeedbackEntry)
				if !ok {
					return &account.Account{}, nil
				}

				if req.GetInitiator() == "" {
					return &account.Account{}, nil
				}

				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetInitiator()})
				if err != nil {
					return &account.Account{}, nil
				}

				return resp.GetAccount(), err
			},
		},
	})

	util.BindOutputName(&property.ViewRequestDetails{}, "Property-related details", util.EnrichmentMap{
		"flatmates": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&account.Account{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.ViewRequestDetails)
				if !ok {
					return []*account.Account{}, nil
				}

				if len(req.GetFlatmates().GetValue()) == 0 {
					return []*account.Account{}, nil
				}

				var list []*account.Account

				for _, f := range req.GetFlatmates().GetValue() {
					acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: f})
					if err != nil {
						log.WithError(err).Error()
						continue
					}

					list = append(list, acc.GetAccount())
				}

				return list, nil
			},
		},
	}, "property_api_ViewRequestDetails")

	util.BindOutput(offer.Guarantor{}, "offerGuarantor", util.EnrichmentMap{
		"identity_documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "document", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Guarantor)
				if !ok {
					return []*document.Document{}, nil
				}

				var documents []*document.Document

				if req.GetIdentityDocuments() == nil {
					req.IdentityDocuments = &generic.StringArrayValue{}
				}
				for _, id := range req.GetIdentityDocuments().GetValue() {
					doc, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{
						Id: id,
					})
					if err != nil {
						log.WithError(err)
						continue
					}
					if doc.GetFile().GetId() != "" {
						d := &document.Document{
							Id:       id,
							TypeCode: document.TypeIdentity,
							File:     doc.GetFile(),
						}
						documents = append(documents, d)
					}
				}
				return documents, nil
			},
		},
		"bill_documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "document", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Guarantor)
				if !ok {
					return []*document.Document{}, nil
				}

				var documents []*document.Document
				if req.GetBillDocuments() == nil {
					req.BillDocuments = &generic.StringArrayValue{}
				}
				for _, id := range req.GetBillDocuments().GetValue() {
					doc, err := connections.DocumentService.Get(p.Context, &document.GetRequest{Id: id})
					if err != nil {
						log.WithError(err)
						continue
					}
					documents = append(documents, doc.GetDocument())
				}

				return documents, nil
			},
		},
	})

	util.BindOutput(offer.Tenant{}, "offerTenant", util.EnrichmentMap{
		"guarantor": &util.Enrichment{
			DeprecationReason: "use offer.guarantors",
			OutputNullable:    true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&offer.Guarantor{}, "offerGuarantor", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				return nil, nil
			},
		},
		"identity_documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "document", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Tenant)
				if !ok {
					return []*document.Document{}, nil
				}

				var documents []*document.Document
				if req.GetIdentityDocuments() == nil {
					req.IdentityDocuments = &generic.StringArrayValue{}
				}
				for _, id := range req.GetIdentityDocuments().GetValue() {
					doc, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{Id: id})
					if err != nil {
						log.WithError(err)
						continue
					}
					if doc.GetFile().GetId() != "" {
						d := &document.Document{
							Id:       id,
							TypeCode: document.TypeIdentity,
							File:     doc.GetFile(),
						}
						documents = append(documents, d)
					}
				}
				return documents, nil
			},
		},
		"bill_documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "document", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Tenant)
				if !ok {
					return []*document.Document{}, nil
				}

				var documents []*document.Document
				if req.GetBillDocuments() == nil {
					req.BillDocuments = &generic.StringArrayValue{}
				}
				for _, id := range req.GetBillDocuments().GetValue() {
					doc, err := connections.DocumentService.Get(p.Context, &document.GetRequest{Id: id})
					if err != nil {
						log.WithError(err)
						continue
					}
					documents = append(documents, doc.GetDocument())
				}

				return documents, nil
			},
		},
		"account": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Tenant)
				if !ok {
					return &account.Account{}, nil
				}
				if req.GetUser() == "" {
					return &account.Account{}, nil
				}
				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetUser()})
				if err != nil {
					return &account.Account{}, nil
				}
				return resp.GetAccount(), err
			},
		},
	})

	util.BindOutput(&property.ViewRequest{}, "Property-related people", util.EnrichmentMap{
		"property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return &property.Property{}, nil
				}

				req, ok := p.Source.(*property.ViewRequest)
				if !ok {
					return &property.Property{}, nil
				}

				if req.GetId() == "" {
					return &property.Property{}, nil
				}

				test := user.GetAccount().GetTestValue()

				resp, err := connections.PropertyService.Get(
					p.Context,
					&property.GetRequest{Id: req.GetProperty(), User: user.GetAccount().GetIdValue(), Test: test},
				)
				if err != nil {
					return nil, err
				}

				return resp.GetProperty(), nil
			},
		},
		"sender": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.ViewRequest)
				if !ok {
					return &account.Account{}, nil
				}

				if req.GetSender() == "" {
					return &account.Account{}, nil
				}

				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetSender()})
				if err != nil {
					return &account.Account{}, nil
				}

				return resp.GetAccount(), err
			},
		},
		"offer": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&offer.Offer{}, "offer", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.ViewRequest)
				if !ok {
					return &offer.Offer{}, nil
				}

				if req.GetSender() == "" {
					return &offer.Offer{}, nil
				}

				resp, err := connections.OfferService.Get(p.Context, &offer.GetRequest{Id: req.GetOffer()})
				if err != nil {
					return &offer.Offer{}, nil
				}

				return resp.GetOffer(), err
			},
		},
		"manager": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "manager", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.ViewRequest)
				if !ok {
					return &account.Account{}, nil
				}

				if req.GetManager() == "" {
					return &account.Account{}, nil
				}

				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetManager()})
				if err != nil {
					return &account.Account{}, nil
				}

				return resp.GetAccount(), err
			},
		},
	})
}

type OfferProvider interface {
	ProvideOffer() *offer.Offer
}

func extractOffer(v interface{}) (*offer.Offer, bool) {
	o, ok := v.(*offer.Offer)
	if ok {
		return o, true
	}

	p, ok := v.(OfferProvider)
	if ok {
		return p.ProvideOffer(), true
	}

	return nil, false
}

type PropertyProvider interface {
	ProvideProperty() *property.Property
}

func extractProperty(v interface{}) (*property.Property, bool) {
	o, ok := v.(*property.Property)
	if ok {
		return o, true
	}

	p, ok := v.(PropertyProvider)
	if ok {
		return p.ProvideProperty(), true
	}

	return nil, false
}

type AccountProvider interface {
	ProvideAccount() *account.Account
}

func extractAccount(v interface{}) (*account.Account, bool) {
	o, ok := v.(*account.Account)
	if ok {
		return o, true
	}

	p, ok := v.(AccountProvider)
	if ok {
		return p.ProvideAccount(), true
	}

	return nil, false
}

func renderOfferPayment(inv *invoice.Invoice) *offer.OfferPayment {
	var services []*extras.ServiceSnapshot

	for _, s := range inv.GetServices() {
		services = append(services, s.GetService())
	}

	var status offer.PaymentStatus

	switch inv.GetStatus() {
	case invoice.InvoiceStatus_INVOICE_FULFILLED:
		status = offer.PaymentStatus_PAYMENT_SUCCESS
	case invoice.InvoiceStatus_INVOICE_OPEN:
		status = offer.PaymentStatus_PAYMENT_PENDING
		for _, p := range inv.GetPayments() {
			if p.GetStatus() == invoice.PaymentStatus_PAYMENT_CONFIRM {
				status = offer.PaymentStatus_PAYMENT_CONFIRM
			}
		}
	case invoice.InvoiceStatus_INVOICE_CANCELED:
		status = offer.PaymentStatus_PAYMENT_FAILED
	}

	var typ offer.PaymentType

	var csrf string

	var validTyp, validCsrf bool

	for _, p := range inv.GetPayments() {
		if p.GetSystem().GetStripe() != nil && !validTyp {
			typ = offer.PaymentType(p.GetSystem().GetStripe().GetRequest().GetType())
			validTyp = p.GetStatus() == invoice.PaymentStatus_PAYMENT_PENDING ||
				p.GetStatus() == invoice.PaymentStatus_PAYMENT_SUCCESS
		}

		if p.GetSystem().GetTokenio() != nil && !validCsrf {
			csrf = p.GetSystem().GetTokenio().GetResponse().GetCsrf()
			validCsrf = p.GetStatus() == invoice.PaymentStatus_PAYMENT_PENDING ||
				p.GetStatus() == invoice.PaymentStatus_PAYMENT_SUCCESS
		}
	}

	return &offer.OfferPayment{
		ReceiptId:     inv.GetInvoiceId(),
		Amount:        inv.GetAmount(),
		Status:        status,
		Type:          typ,
		CreatedAt:     inv.GetCreatedAt(),
		UpdatedAt:     inv.GetUpdatedAt(),
		Services:      services,
		InvoiceNumber: inv.GetReference(),
		CsrfToken:     csrf,
	}
}

//nolint:gocyclo,gocognit // this func can have high cyclomatic and cognitive complexity
func initOfferEnrichment() {
	util.BindOutputName(offer.Landlord{}, "", util.EnrichmentMap{
		"bank_details": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.AccountBankDetails{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				n, ok := p.Source.(*offer.Landlord)
				if !ok || n.GetUser() == "" {
					return nil, nil
				}

				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{
					Id: n.GetUser(),
				})
				if err != nil {
					return nil, nil
				}

				return resp.GetAccount().GetBankDetailsValue(), nil
			},
		},
	}, "offer_api_Landlord")

	util.BindOutputName(offer.PayoutResponse{}, "", util.EnrichmentMap{
		"success": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.Boolean
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return true, nil
			},
			OutputNullable:    false,
			DeprecationReason: "do not use",
		},
	}, "offer_api_PayoutResponse")

	util.BindOutputName(offer.Payout{}, "", util.EnrichmentMap{
		"invoice": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&invoice.Invoice{}, "invoice", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*offer.Payout)
				if !ok || req == nil || req.InvoiceId == "" {
					return &invoice.Invoice{}, nil
				}
				inv, err := connections.InvoiceService.InvoiceGet(p.Context, &invoice.InvoiceGetRequest{
					InvoiceId: req.InvoiceId,
				})
				if err != nil || inv == nil || inv.GetInvoice() == nil {
					return &invoice.Invoice{}, nil
				}
				return inv.GetInvoice(), err
			},
		},
	}, "offer_api_Payout")

	util.BindOutputName(offer.ScheduledPaymentPart{}, "", util.EnrichmentMap{
		"invoice": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&invoice.Invoice{}, "invoice", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*offer.ScheduledPaymentPart)
				if !ok || req == nil || req.InvoiceId == "" {
					return nil, nil
				}

				inv, err := connections.InvoiceService.InvoiceGet(p.Context, &invoice.InvoiceGetRequest{
					InvoiceId: req.GetInvoiceId(),
				})
				if err != nil || inv == nil || inv.GetInvoice() == nil {
					return nil, nil
				}

				return inv.GetInvoice(), err
			},
			OutputNullable: true,
		},
		"payment": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&invoice.Payment{}, "payment", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*offer.ScheduledPaymentPart)
				if !ok || req == nil || req.PaymentId == "" {
					return nil, nil
				}

				pmt, err := connections.InvoiceService.PaymentGet(p.Context, &invoice.PaymentGetRequest{
					PaymentId: req.GetPaymentId(),
				})
				if err != nil || pmt == nil || pmt.GetPayment() == nil {
					return nil, nil
				}

				return pmt.GetPayment(), err
			},
			OutputNullable: true,
		},
	}, "offer_api_ScheduledPaymentPart")

	util.BindOutputName(offer.ScheduledPayment{}, "", util.EnrichmentMap{
		"due": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return "", nil
			},
			OutputNullable:    false,
			DeprecationReason: "use parts.payment_due",
		},
		"invoice_id": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return "", nil
			},
			OutputNullable:    false,
			DeprecationReason: "use payout.invoice_id / parts.invoice_id",
		},
		"services": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(extras.ServiceSnapshot{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return []*extras.ServiceSnapshot{}, nil
			},
			OutputNullable:    false,
			DeprecationReason: "use parts.services",
		},
		"offer": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&offer.Offer{}, "offer", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*offer.ScheduledPayment)
				if !ok || req.Offer == "" {
					return &offer.Offer{}, nil
				}
				off, err := connections.OfferService.ServiceGet(p.Context, &offer.GetRequest{
					Id: req.GetOffer(),
				})
				if err != nil || off == nil || off.GetOffer() == nil {
					return &offer.Offer{}, nil
				}
				return off.GetOffer(), err
			},
		},
	}, "offer_api_ScheduledPayment")

	util.BindOutputName(offer.OfferConditions{}, "", util.EnrichmentMap{
		"termination_allowed": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.Boolean
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*offer.OfferConditions)
				if !ok {
					return false, nil
				}

				if req.GetMoveinDate().AsTime().Unix() <= 0 || req.GetBreakClause().GetValue() <= 0 {
					return false, nil
				}

				tEndBreakClause := req.GetMoveinDate().AsTime().AddDate(0, int(req.GetBreakClause().GetValue()), 0)

				tEndTenancy := req.GetMoveinDate().AsTime().AddDate(0, int(req.GetTermMonths().GetValue()), 0)

				now := time.Now()

				if tEndBreakClause.Before(now) && now.Before(tEndTenancy.AddDate(0, -3, 0)) {
					return true, nil
				}

				return false, nil
			},
		},
	}, "offer_api_OfferConditions")

	util.BindOutput(offer.Offer{}, "offer", util.EnrichmentMap{
		"viewing": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&viewing.Viewing{}, "viewing", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := extractOffer(p.Source)

				if req.GetTenants() == nil {
					req.Tenants = &offer.TenantArrayValue{}
				}

				if !ok || len(req.GetTenants().GetValue()) < 1 {
					return &viewing.Viewing{}, nil
				}

				view, err := connections.ViewingService.ViewingGet(p.Context, &viewing.ViewingGetRequest{
					Spec: &viewing.ViewingSpec{
						UserId: req.GetTenants().GetValue()[0].GetUser(),
						Select: &viewing.ViewingSpecSelect{
							PropertyId: req.GetProperty(),
							ViewingId:  req.GetViewing(),
						},
					},
				})
				if err != nil || view == nil || view.GetViewing() == nil {
					return &viewing.Viewing{}, nil
				}

				return view.GetViewing(), nil
			},
		},
		"property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "property", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := extractOffer(p.Source)
				if !ok {
					return &property.Property{}, nil
				}

				if req.GetProperty() == "" {
					return &property.Property{}, nil
				}

				var test bool

				user, _ := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if user != nil {
					test = user.GetAccount().GetTestValue()
				}

				resp, err := connections.PropertyService.Get(p.Context, &property.GetRequest{
					Id:   req.GetProperty(),
					User: user.GetAccount().GetIdValue(),
					Test: test,
				})
				if err != nil || resp == nil || resp.GetProperty() == nil {
					return &property.Property{}, nil
				}
				return resp.GetProperty(), nil
			},
		},
		"payment": &util.Enrichment{
			DeprecationReason: "use invoices field",
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&offer.OfferPayment{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				off, ok := extractOffer(p.Source)
				if !ok {
					return nil, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok {
					return nil, nil
				}

				inv, err := connections.InvoiceService.InvoiceGet(p.Context, &invoice.InvoiceGetRequest{
					CorrelationType: correlationTypeOffer,
					CorrelationId:   off.GetId(),
					User:            user.GetAccount().GetIdValue(),
				})
				if err != nil {
					return nil, nil
				}

				return renderOfferPayment(inv.GetInvoice()), nil
			},
			OutputNullable: true,
		},
		"payment_history": &util.Enrichment{
			DeprecationReason: "use invoices field",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&offer.OfferPayment{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				payments := make([]*offer.OfferPayment, 0)

				off, ok := extractOffer(p.Source)
				if !ok {
					return payments, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok {
					return payments, nil
				}

				filters := []*query.QueryParamsFilter{
					{
						Fields:   []string{"correlation_type"},
						Value:    []string{correlationTypeOffer},
						Operator: query.QueryFilterOperator_EQUALS,
					},
					{
						Fields:   []string{"correlation_id"},
						Value:    []string{off.GetId()},
						Operator: query.QueryFilterOperator_EQUALS,
					},
					{
						Fields:   []string{"status"},
						Value:    []string{"INVOICE_OPEN", "INVOICE_FULFILLED"},
						Operator: query.QueryFilterOperator_ONEOF,
					},
				}

				userid := user.GetAccount().GetIdValue()

				if !HasRole(user, AdminRole) && off.GetLandlord().GetUser() != userid {
					filters = append(filters, &query.QueryParamsFilter{
						Fields:   []string{"userid"},
						Value:    []string{user.GetAccount().GetIdValue()},
						Operator: query.QueryFilterOperator_EQUALS,
					})
				}

				invs, err := connections.InvoiceService.InvoiceFilterList(p.Context, &invoice.InvoiceFilterListRequest{
					Params: &query.QueryParams{
						Pagination: &query.QueryParamsPagination{
							Page:    1,
							PerPage: 1000,
						},
						Sort: []*query.QueryParamsSort{
							{
								Field: "created",
								Order: 1,
							},
						},
						Filter: filters,
					},
				})
				if err != nil {
					return payments, nil
				}

				for _, inv := range invs.GetInvoices() {
					if off.GetLandlord().GetUser() == userid && inv.GetUser() != userid {
						found := false

					innerloop:
						for _, s := range inv.GetServices() {
							switch s.GetService().GetType() {
							case extras.ServiceType_RENT, extras.ServiceType_DEPOSIT:
								found = true
								break innerloop
							}

							if !found {
								continue
							}
						}
					}

					payments = append(payments, renderOfferPayment(inv))
				}

				return payments, nil
			},
		},
		"invoices": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&invoice.Invoice{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				off, ok := extractOffer(p.Source)
				if !ok {
					return []*invoice.Invoice{}, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok {
					return []*invoice.Invoice{}, nil
				}

				var userid string

				if !HasRole(user, AdminRole) && user.GetAccount().GetIdValue() != off.GetLandlord().GetUser() {
					userid = user.GetAccount().GetIdValue()
				}

				invs, err := connections.InvoiceService.InvoiceList(p.Context, &invoice.InvoiceListRequest{
					CorrelationType: correlationTypeOffer,
					CorrelationId:   off.GetId(),
					User:            userid,
				})
				if err != nil {
					return []*invoice.Invoice{}, nil
				}

				return invs.GetInvoices(), nil
			},
		},
		"administrator": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				off, ok := extractOffer(p.Source)
				if !ok || off.GetAdministrator() == "" {
					return nil, nil
				}
				acc, err := connections.AccountService.Get(
					p.Context,
					&account.GetRequest{
						Id:    off.GetAdministrator(),
						Admin: true,
					},
				)
				if err != nil {
					log.WithError(err).Error()
					return nil, nil
				}

				return acc.GetAccount(), nil
			},
		},
	})
}

func initMaintenanceEnrichment() {
	util.BindOutputName(maintenance.SpecialistPrivateContractor{}, "Private contractor", util.EnrichmentMap{
		"first_name": &util.Enrichment{
			DeprecationReason: "Use name",
			OutputNullable:    false,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return "", nil
			},
		},
		"last_name": &util.Enrichment{
			DeprecationReason: "Use name",
			OutputNullable:    false,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return "", nil
			},
		},
	}, "maintenance_api_SpecialistPrivateContractor")

	util.BindOutput(maintenance.Maintenance{}, "maintenance", util.EnrichmentMap{
		"specialist_someone_else": &util.Enrichment{
			DeprecationReason: "Use specialist",
			OutputNullable:    true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return &maintenance.Specialist{}, nil
			},
		},
		"specialist_local_heroes": &util.Enrichment{
			DeprecationReason: "Use specialist",
			OutputNullable:    true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return &maintenance.SpecialistLocalHeroes{}, nil
			},
		},
		"specialist_type": &util.Enrichment{
			DeprecationReason: "Use specialist",
			OutputNullable:    true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				m, ok := p.Source.(*maintenance.Maintenance)
				if !ok {
					return maintenance.MaintenanceSpecialistType_SPECIALIST_TYPE_UNKNOWN, nil
				}

				switch {
				case m.GetLocalHeroes() != nil:
					return maintenance.MaintenanceSpecialistType_LOCAL_HEROES, nil
				case m.GetSomeoneElse() != nil:
					return maintenance.MaintenanceSpecialistType_SOMEONE_ELSE, nil
				default:
					return maintenance.MaintenanceSpecialistType_SPECIALIST_TYPE_UNKNOWN, nil
				}
			},
		},
		"occurring_date": &util.Enrichment{
			DeprecationReason: "Use desired_repair_date",
			OutputNullable:    true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				m, ok := p.Source.(*maintenance.Maintenance)
				if !ok {
					return "", nil
				}

				return m.GetDesiredRepairDate().AsTime().Format(time.RFC1123), nil
			},
		},
		"category": &util.Enrichment{
			DeprecationReason: "Use taxonomy_id",
			OutputNullable:    true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return maintenance.LocalHeroesCategory_UNKNOWN_CATEGORY, nil
			},
		},
		"subcategory": &util.Enrichment{
			DeprecationReason: "Use taxonomy_id",
			OutputNullable:    true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return maintenance.LocalHeroesSubcategory_UNKNOWN_SUBCATEGORY, nil
			},
		},
		"job_description": &util.Enrichment{
			DeprecationReason: "Use title/description",
			OutputNullable:    true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				m, ok := p.Source.(*maintenance.Maintenance)
				if !ok {
					return "", nil
				}

				return m.GetDescription(), nil
			},
		},
		"documents": &util.Enrichment{
			DeprecationReason: "Use files",
			OutputNullable:    true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "document", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*maintenance.Maintenance)
				if !ok {
					return []*document.Document{}, nil
				}

				var documents []*document.Document

				for _, id := range req.GetDocuments() {
					doc, err := connections.DocumentService.Get(p.Context, &document.GetRequest{Id: id})
					if err != nil {
						log.WithError(err)
						continue
					}
					documents = append(documents, doc.GetDocument())
				}

				return documents, nil
			},
		},
		"files": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.File{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*maintenance.Maintenance)
				if !ok {
					return []*document.File{}, nil
				}

				var files []*document.File

				for _, id := range req.GetFiles() {
					f, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{Id: id})
					if err != nil {
						log.WithError(err)
						continue
					}

					files = append(files, f.GetFile())
				}

				return files, nil
			},
		},
	})
}

func initMortgageEnrichment() {
	util.BindOutput(mortgage.Lender{}, "Lender description", util.EnrichmentMap{
		"small_image": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&document.File{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				acc, ok := p.Source.(*mortgage.Lender)
				if !ok || acc.GetSmallImage() == "" {
					return &document.File{}, nil
				}

				resp, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{
					Id: acc.GetSmallImage(),
				})
				if err != nil {
					return &document.File{}, nil
				}

				return resp.GetFile(), err
			},
		},
	})

	util.BindOutput(mortgage.SourceResult{}, "Lender description", util.EnrichmentMap{
		"lender_small_image": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&document.File{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				acc, ok := p.Source.(*mortgage.SourceResult)
				if !ok || acc.GetLenderSmallImage() == "" {
					return &document.File{}, nil
				}

				resp, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{
					Id: acc.GetLenderSmallImage(),
				})
				if err != nil {
					return &document.File{}, nil
				}

				return resp.GetFile(), err
			},
		},
	})
}

func initNoteEnrichment() {
	util.BindOutputName(notes.Note{}, "Note details", util.EnrichmentMap{
		"user": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				n, ok := p.Source.(*notes.Note)
				if !ok || n.GetUser() == "" {
					return nil, nil
				}

				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{
					Id:    n.GetUser(),
					Admin: true,
				})
				if err != nil {
					return nil, nil
				}

				return resp.GetAccount(), nil
			},
		},
		"correlated_user": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				n, ok := p.Source.(*notes.Note)
				if !ok || n.GetCorrelationId() == "" ||
					n.GetCorrelationType() != notes.CorrelationType_CORRELATION_TYPE_USER {
					return nil, nil
				}

				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{
					Id: n.GetCorrelationId(),
				})
				if err != nil {
					return nil, nil
				}

				return resp.GetAccount(), nil
			},
		},
		"correlated_property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				n, ok := p.Source.(*notes.Note)
				if !ok || n.GetCorrelationId() == "" ||
					n.GetCorrelationType() != notes.CorrelationType_CORRELATION_TYPE_PROPERTY {
					return nil, nil
				}

				var test bool

				user, _ := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if user != nil {
					test = user.GetAccount().GetTestValue()
				}

				resp, err := connections.PropertyService.Get(p.Context, &property.GetRequest{
					Id:   n.GetCorrelationId(),
					Test: test,
				})
				if err != nil {
					return nil, nil
				}

				return resp.GetProperty(), nil
			},
		},
	}, "notes_api_Note")
}

func initCommentEnrichment() {
	util.BindOutputName(comments.Comment{}, "Comment details", util.EnrichmentMap{
		"user": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				n, ok := p.Source.(*comments.Comment)
				if !ok || n.GetUser() == "" {
					return nil, nil
				}

				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{
					Id: n.GetUser(),
				})
				if err != nil {
					return nil, nil
				}

				return resp.GetAccount(), nil
			},
		},
		"nested": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(util.BindOutput(&comments.Comment{}, "", nil))
			},
			Args: util.BindArguments(commentsNestedPagination{}),
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				n, ok := p.Source.(*comments.Comment)
				if !ok || n.GetUser() == "" {
					return []*comments.Comment{}, nil
				}

				if n.GetReplies() == 0 {
					return []*comments.Comment{}, nil
				}

				request := &commentsNestedPagination{}
				if err := util.UnmarshalArguments(p.Args, request); err != nil {
					return []*comments.Comment{}, nil
				}

				if request.Limit == 0 {
					request.Limit = 10
				}

				if len(request.Sort) == 0 {
					request.Sort = append(request.Sort, &query.QueryParamsSort{
						Field: "created",
						Order: 1,
					})
				}

				request.Filter = append(request.Filter, &query.QueryParamsFilter{
					Fields:   []string{"parent_id"},
					Value:    []string{n.GetId()},
					Operator: query.QueryFilterOperator_EQUALS,
				})

				resp, err := connections.CommentsService.CommentFilter(p.Context, &comments.CommentFilterRequest{
					Params: &query.QueryParams{
						Pagination: &query.QueryParamsPagination{
							Page:    0,
							PerPage: request.Limit,
						},
						Filter: append([]*query.QueryParamsFilter{
							{
								Fields:   []string{"correlation_type"},
								Value:    []string{n.GetCorrelationType().String()},
								Operator: query.QueryFilterOperator_EQUALS,
							},
							{
								Fields:   []string{"correlation_id"},
								Value:    []string{n.GetCorrelationId()},
								Operator: query.QueryFilterOperator_EQUALS,
							},
						}, request.Filter...),
						Sort: request.Sort,
					},
				})
				if err != nil {
					return []*comments.Comment{}, nil
				}

				return resp.GetComments(), nil
			},
		},
	}, "comments_api_Comment")
}

func initDocumentEnrichment() {
	util.BindOutputName(document.Document{}, "", util.EnrichmentMap{
		"property": &util.Enrichment{
			DeprecationReason: "use property_id",
			OutputNullable:    false,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, _ := p.Source.(*document.Document)

				return req.GetProperty(), nil
			},
		},
		"property_id": &util.Enrichment{
			OutputNullable: false,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, _ := p.Source.(*document.Document)

				return req.GetProperty(), nil
			},
		},
		"property_data": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*document.Document)
				if !ok || req.GetProperty() == "" {
					return nil, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return nil, nil
				}

				test := user.GetAccount().GetTestValue()

				resp, _ := connections.PropertyService.Get(
					p.Context,
					&property.GetRequest{Id: req.GetProperty(), User: user.GetAccount().GetIdValue(), Test: test},
				)

				return resp.GetProperty(), nil
			},
		},
		"product": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&invoice.Product{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*document.Document)
				if !ok || req.ProductId == "" {
					return &invoice.Product{}, nil
				}
				product, err := connections.InvoiceService.ProductGet(p.Context, &invoice.ProductGetRequest{
					Id: req.GetProductId(),
				})
				if err != nil || product == nil || product.GetProduct() == nil {
					return &invoice.Product{}, nil
				}
				return product.GetProduct(), nil
			},
		},
	}, "document_api_Document")

	util.BindOutputName(document.FileAttribute{}, "", util.EnrichmentMap{
		"file": &util.Enrichment{
			OutputNullable: false,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&document.File{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				s, _ := p.Source.(*document.FileAttribute)

				resp, _ := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{
					Id: s.GetFile(),
				})

				if resp.GetFile() == nil {
					return &document.File{}, nil
				}

				return resp.GetFile(), nil
			},
		},
	}, "document_api_FileAttribute")
}

func initDocumentTypeEnrichment() {
	util.BindOutputName(document.DocumentTypeDesc{}, "", util.EnrichmentMap{
		"services": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(
					util.BindOutput(&extras.Service{}, "", nil),
				))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*document.DocumentTypeDesc)
				if !ok {
					return []extras.Service{}, nil
				}

				resp, err := connections.ExtrasService.List(p.Context, &extras.ListRequest{
					Attribute: &extras.Attribute{
						Key:   string(extras.AttributeDocumentType),
						Value: req.Code,
					},
				})
				if err != nil {
					return []extras.Service{}, nil
				}

				return resp.GetExtraServices(), nil
			},
		},
	}, "document_api_DocumentTypeDesc")
}

func initExpensesEnrichment() {
	util.BindOutputName(expenses.Record{}, "", util.EnrichmentMap{
		"offer": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&offer.Offer{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				n, ok := p.Source.(*expenses.Record)
				if !ok || n.GetCorrelationType() != expenses.CorrelationType_CORRELATION_TYPE_OFFER {
					return nil, nil
				}

				resp, err := connections.OfferService.ServiceGet(p.Context, &offer.GetRequest{
					Id: n.GetCorrelationId(),
				})
				if err != nil {
					return nil, nil
				}

				return resp.GetOffer(), nil
			},
		},
	}, "expenses_api_Record")

	util.BindOutputName(expenses.TaxVAT{}, "", util.EnrichmentMap{
		"stub": &util.Enrichment{
			DeprecationReason: "stub, do not use",
			OutputNullable:    true,
		},
	}, "expenses_api_TaxVAT")

	util.BindOutputName(expenses.IncomeRent{}, "", util.EnrichmentMap{
		"overdue_payment_notice_count": &util.Enrichment{
			Fields: nil,
			Input:  nil,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.Int
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				// Will be finished in other task - for now we don't have a logic and notifications which should be
				// counted. It's required to have got this field now, just not to block FE.

				return 0, nil
			},
			OutputNullable: true,
		},
	}, "expenses_api_IncomeRent")

	util.BindOutputName(expenses.ExpenseRent{}, "", util.EnrichmentMap{
		"overdue_payment_notice_count": &util.Enrichment{
			Fields: nil,
			Input:  nil,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.Int
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				// Will be finished in other task - for now we don't have a logic and notifications which should be
				// counted. It's required to have got this field now, just not to block FE.

				return 0, nil
			},
			OutputNullable: true,
		},
	}, "expenses_api_ExpenseRent")

	util.BindOutputName(expenses.ExpenseDrp{}, "", util.EnrichmentMap{
		"overdue_payment_notice_count": &util.Enrichment{
			Fields: nil,
			Input:  nil,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.Int
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				// Will be finished in other task - for now we don't have a logic and notifications which should be
				// counted. It's required to have got this field now, just not to block FE.

				return 0, nil
			},
			OutputNullable: true,
		},
	}, "expenses_api_ExpenseDrp")
}

func initInvoiceEnrichment() {
	util.BindOutputName(invoice.Invoice{}, "invoice", util.EnrichmentMap{
		"administrator": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				inv, ok := p.Source.(*invoice.Invoice)
				if !ok || inv.GetAdmin() == "" {
					return nil, nil
				}
				acc, err := connections.AccountService.Get(
					p.Context,
					&account.GetRequest{
						Id:    inv.GetAdmin(),
						Admin: true,
					},
				)
				if err != nil {
					log.WithError(err).Error()
					return nil, nil
				}

				return acc.GetAccount(), nil
			},
		},
		"admin": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				inv, _ := p.Source.(*invoice.Invoice)

				return inv.GetAdmin(), nil
			},
			OutputNullable:    false,
			DeprecationReason: "use administrator",
		},
		"order": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&invoice.Order{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				n, ok := p.Source.(*invoice.Invoice)
				if !ok || n.GetOrderId() == "" {
					return nil, nil
				}

				ord, _ := connections.InvoiceService.OrderLoad(p.Context, &invoice.OrderLoadRequest{
					Id: n.GetOrderId(),
				})

				return ord.GetOrder(), nil
			},
		},
		"user_data": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ps, ok := p.Source.(*invoice.Invoice)
				if !ok {
					return &account.Account{}, nil
				}

				acc, err := connections.AccountService.Get(
					p.Context,
					&account.GetRequest{
						Id: ps.GetUser(),
					},
				)
				if err != nil {
					log.WithError(err).Error()
					return &account.Account{}, nil
				}

				return CensoredAccount(acc.GetAccount()), nil
			},
			OutputNullable: true,
		},
		"property_data": &util.Enrichment{
			Fields: nil,
			Input:  nil,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(property.Property{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ps, ok := p.Source.(*invoice.Invoice)
				if !ok {
					return nil, nil
				}

				if ps.GetProperty() == "" {
					return nil, nil
				}

				resp, err := connections.PropertyService.Get(p.Context, &property.GetRequest{
					Id: ps.GetProperty(),
				})
				if err != nil {
					return nil, err
				}

				return resp.GetProperty(), nil
			},
			OutputNullable: true,
		},
	}, "invoice_api_Invoice")
}

func initInsuranceEnrichment() {
	util.BindOutputName(insurance.Payment{}, "Payment details", util.EnrichmentMap{
		"order": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&invoice.Order{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				n, ok := p.Source.(*insurance.Payment)
				if !ok || n.GetOrderId() == "" {
					return nil, nil
				}

				user, _ := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)

				ord, _ := connections.InvoiceService.OrderLoad(p.Context, &invoice.OrderLoadRequest{
					Id:        n.GetOrderId(),
					Userid:    user.GetAccount().GetIdValue(),
					OrderType: invoice.OrderType_CONTENTS_INSURANCE_ORDER,
				})

				return ord.GetOrder(), nil
			},
		},
	}, "insurance_api_Payment")
}

func initRtoEnrichment() {
	util.BindOutputName(rto.Rto{}, "", util.EnrichmentMap{
		"invoice": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&invoice.Invoice{}, "invoice", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*rto.Rto)
				if !ok || req == nil || req.GetInvoiceId() == "" {
					return &invoice.Invoice{}, nil
				}

				invResp, err := connections.InvoiceService.InvoiceGet(p.Context, &invoice.InvoiceGetRequest{
					InvoiceId: req.GetInvoiceId(),
				})
				if err != nil || invResp == nil {
					return nil, nil
				}

				return invResp.GetInvoice(), err
			},
		},
	}, "offer_api_rto_Rto")
}

func InitSchema(rootQueries, rootMutations, rootSubscriptions *graphql.Fields) {
	initTransactionEnrichment()
	initOrderEnrichment()
	initDocumentEnrichment()
	initProductEnrichment()
	initAccountEnrichment()
	initPropertyEnrichment()
	initViewingEnrichment()
	initOfferEnrichment()
	initMaintenanceEnrichment()
	initMortgageEnrichment()
	initNoteEnrichment()
	initExpensesEnrichment()
	initInvoiceEnrichment()
	initInsuranceEnrichment()
	initCommentEnrichment()
	initDocumentTypeEnrichment()
	initRtoEnrichment()

	initAccount(rootQueries, rootMutations)
	initAddress(rootQueries, rootMutations)
	initAuth(rootQueries, rootMutations)
	initCardPayment(rootQueries, rootMutations)
	initContact(rootQueries, rootMutations)
	initContract(rootQueries, rootMutations)
	initCloseCom(rootQueries, rootMutations)
	initDocument(rootQueries, rootMutations)
	initNotification(rootQueries, rootMutations)
	initGeo(rootQueries, rootMutations)
	initInvitation(rootQueries, rootMutations)
	initInvoice(rootQueries, rootMutations)
	initMaintenance(rootQueries, rootMutations)
	initMessaging(rootQueries, rootMutations, rootSubscriptions)
	initOAuth(rootQueries, rootMutations)
	initOffer(rootQueries, rootMutations)
	initRto(rootQueries, rootMutations)
	initProperty(rootQueries, rootMutations)
	initViewing(rootQueries, rootMutations)
	initListing(rootQueries, rootMutations)
	initService(rootQueries, rootMutations)
	initDocumentType(rootQueries)
	initSignature(rootQueries, rootMutations)
	initOrder(rootQueries, rootMutations)
	initMortgage(rootQueries, rootMutations)
	initInsurance(rootQueries, rootMutations)
	initPayment(rootQueries, rootMutations)
	initMixpanel(rootQueries, rootMutations)
	initGrabdata(rootQueries, rootMutations)
	initExpenses(rootQueries, rootMutations)
	initComments(rootQueries, rootMutations)
	initOwnership(rootQueries, rootMutations)
}
