package admin

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	q "gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/generic/env"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	tokenio "gitlab.mashplace.com/mashplace/backend/tokenio/api"
	viewing "gitlab.mashplace.com/mashplace/backend/viewing/api"
)

const (
	MediaCacheDirEnv       = "MEDIA_CACHE_DIR"
	MediaCacheDirDefault   = "/media"
	SitemapCacheDirEnv     = "SITEMAP_CACHE_DIR"
	SitemapCacheDirDefault = "/sitemaps"
	correlationTypeOffer   = "offer"
)

var (
	MediaCacheDir   = env.ResolveString(MediaCacheDirEnv, MediaCacheDirDefault)
	SitemapCacheDir = env.ResolveString(SitemapCacheDirEnv, SitemapCacheDirDefault)
)

type allParamsPagination struct {
	Page    int32 `json:"page,omitempty"`
	PerPage int32 `json:"perPage,omitempty"`
}

type allParams struct {
	Pagination allParamsPagination    `json:"pagination,omitempty"`
	Sort       []*q.QueryParamsSort   `json:"sort,omitempty"`
	Filter     []*q.QueryParamsFilter `json:"filter,omitempty"`
}

type getParams struct {
	ID string `json:"id"`
}

func InitSchema(rootQueries, rootMutations *graphql.Fields) {
	queryFields := graphql.Fields{}
	mutationFields := graphql.Fields{}

	initTokenioEnrichment()

	initUser(&queryFields, &mutationFields)
	initAdmin(&queryFields, &mutationFields)
	initProperty(&queryFields, &mutationFields)
	initRto(&queryFields, &mutationFields)
	initOffer(&queryFields, &mutationFields)
	initDocument(&queryFields, &mutationFields)
	initRightmove(&queryFields, &mutationFields)
	initBoomin(&queryFields, &mutationFields)
	initListing(&queryFields, &mutationFields)
	initZoopla(&queryFields, &mutationFields)
	initContract(&queryFields, &mutationFields)
	initCardPayment(&queryFields, &mutationFields)
	initInvoice(&queryFields, &mutationFields)
	initSignature(&queryFields, &mutationFields)
	initMaintenance(&queryFields, &mutationFields)
	initMessaging(&queryFields, &mutationFields)
	initInvitation(&queryFields, &mutationFields)
	initPayment(&queryFields, &mutationFields)
	initGeo(&queryFields, &mutationFields)
	initServices(&queryFields, &mutationFields)
	initDocumentType(&queryFields, &mutationFields)
	initStats(&queryFields, &mutationFields)
	initOrder(&queryFields, &mutationFields)
	initMortgage(&queryFields, &mutationFields)
	initNotes(&queryFields, &mutationFields)
	initComments(&queryFields, &mutationFields)
	initNotification(&queryFields, &mutationFields)
	initExpenses(&queryFields, &mutationFields)
	initVerification(&queryFields, &mutationFields)
	initOwnership(&queryFields, &mutationFields)

	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "AdminQuery",
		Description: "Admin query root",
		Fields:      queryFields,
	}))

	(*rootQueries)["admin"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "AdminMutation",
		Description: "Admin mutation root",
		Fields:      mutationFields,
	}))

	(*rootMutations)["admin"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}

	initOfferEnrichment()
}

func initTokenioEnrichment() {
	util.BindOutputName(PaymentUpdate{}, "", util.EnrichmentMap{}, "tokenio_api_PaymentUpdate")
	util.BindOutputName(UpdateSchedule{}, "", util.EnrichmentMap{}, "tokenio_api_UpdateSchedule")

	util.BindOutputName(tokenio.Payment{}, "", util.EnrichmentMap{
		"offer": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "use correlation_id / correlation_type",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*tokenio.Payment)
				if !ok || req.CorrelationType != correlationTypeOffer {
					return "", nil
				}

				return req.CorrelationId, nil
			},
		},
		"receipt_id": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "use payment_id",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*tokenio.Payment)
				if !ok {
					return "", nil
				}

				return req.PaymentId, nil
			},
		},
	}, "tokenio_api_Payment")

	util.BindOutputName(tokenio.Payment{}, "", util.EnrichmentMap{
		"offer": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "use correlation_id / correlation_type",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*tokenio.Payment)
				if !ok || req.CorrelationType != correlationTypeOffer {
					return "", nil
				}

				return req.CorrelationId, nil
			},
		},
		"receipt_id": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "use payment_id",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*tokenio.Payment)
				if !ok {
					return "", nil
				}

				return req.PaymentId, nil
			},
		},
	}, "tokenio_api_Payment")

	util.BindOutputName(tokenio.Schedule{}, "", util.EnrichmentMap{
		"offer": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "use correlation_id / correlation_type",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*tokenio.Schedule)
				if !ok || req.CorrelationType != correlationTypeOffer {
					return "", nil
				}

				return req.CorrelationId, nil
			},
		},
	}, "tokenio_api_Schedule")
}

func initOfferEnrichment() {
	util.BindOutput(offerData{}, "offer", util.EnrichmentMap{
		"viewing": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&viewing.Viewing{}, "viewing", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Offer)

				if req.Tenants == nil {
					req.Tenants = &offer.TenantArrayValue{}
				}

				if !ok || len(req.GetTenants().GetValue()) < 1 {
					return &viewing.Viewing{}, nil
				}

				view, err := connections.ViewingService.ViewingGet(p.Context, &viewing.ViewingGetRequest{
					Spec: &viewing.ViewingSpec{
						UserId: req.GetTenants().GetValue()[0].GetUser(),
						Select: &viewing.ViewingSpecSelect{
							PropertyId: req.GetProperty(),
							ViewingId:  req.GetViewing(),
						},
					},
				})
				if err != nil || view == nil || view.GetViewing() == nil {
					return &viewing.Viewing{}, nil
				}

				return view.GetViewing(), nil
			},
		},
		"property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "property", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Offer)
				if !ok {
					return &property.Property{}, nil
				}

				if req.GetProperty() == "" {
					return &property.Property{}, nil
				}

				var test bool

				user, _ := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if user != nil {
					test = user.GetAccount().GetTestValue()
				}

				resp, err := connections.PropertyService.Get(p.Context, &property.GetRequest{
					Id:   req.GetProperty(),
					Test: test,
				})
				if err != nil || resp == nil || resp.GetProperty() == nil {
					return &property.Property{}, nil
				}

				return resp.GetProperty(), nil
			},
		},
	})
}
