package admin

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	boomin "gitlab.mashplace.com/mashplace/backend/boomin/api"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
)

func initBoomin(query, mutation *graphql.Fields) {
	(*mutation)["publishBoomin"] = &graphql.Field{
		Description:       "Publish property on Boomin",
		DeprecationReason: "use publishVersion",
		Args:              util.BindArguments(getParams{}),
		Resolve:           BoominPublishResolver,
		Type:              util.BindOutput(boomin.SendPropertyResponse{}, "Property ids", nil),
	}
	(*mutation)["deleteBoomin"] = &graphql.Field{
		Description:       "Delete published property from Boomin",
		DeprecationReason: "use unpublishVersion",
		Args:              util.BindArguments(getParams{}),
		Resolve:           BoominDeleteResolver,
		Type:              util.BindOutput(boomin.DeletePropertyResponse{}, "Delete result", nil),
	}
}

func BoominPublishResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &getParams{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	idv, err := connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
		Spec: listing.VersionSpecPropertyID("", req.ID, listing.VersionStatus_VERSION_VERIFIED),
	})
	if err != nil {
		return nil, err
	}

	resp, err := connections.ListingService.VersionPublish(params.Context, &listing.VersionPublishRequest{
		UserId:    user.GetAccount().GetIdValue(),
		VersionId: idv.GetVersion().GetId(),
		Boomin:    &listing.VersionPublicationChangeItem{},
	})
	if err != nil {
		return nil, err
	}

	return &boomin.SendPropertyResponse{
		Success:   true,
		Errors:    nil,
		ErrorMsg:  "",
		Property:  req.ID,
		BoominId:  resp.GetVersion().GetPublication().GetBoomin().GetId(),
		BoominUrl: resp.GetVersion().GetPublication().GetBoomin().GetUrl(),
	}, nil
}

func BoominDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &getParams{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	idv, err := connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
		Spec: listing.VersionSpecPropertyID(
			"",
			req.ID,
			listing.VersionStatus_VERSION_VERIFIED,
		),
	})
	if err != nil {
		idv, err = connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
			Spec: listing.VersionSpecPropertyID(
				"",
				req.ID,
				listing.VersionStatus_VERSION_VERIFIED,
				listing.VersionStatus_VERSION_VERIFYING,
				listing.VersionStatus_VERSION_DRAFT,
				listing.VersionStatus_VERSION_LET,
			),
		})
		if err != nil {
			return nil, err
		}
	}

	_, err = connections.ListingService.VersionUnpublish(params.Context, &listing.VersionUnpublishRequest{
		UserId:    user.GetAccount().GetIdValue(),
		VersionId: idv.GetVersion().GetId(),
		Boomin:    true,
	})
	if err != nil {
		return nil, err
	}

	return &boomin.DeletePropertyResponse{
		PropertyId: req.ID,
		Success:    true,
	}, nil
}
