package admin

import (
	"github.com/graphql-go/graphql"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	document "gitlab.mashplace.com/mashplace/backend/document/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
)

func initDocumentType(query, mutation *graphql.Fields) {
	(*query)["DocumentType"] = &graphql.Field{
		Description: "Get document type by spec",
		Args:        util.BindArguments(document.DocumentTypeGetRequest{}),
		Resolve:     adminQueryGetDocumentTypeResolver,
		Type:        util.BindOutput(document.DocumentTypeDesc{}, "Get document type", nil),
	}
	(*query)["allDocumentTypes"] = &graphql.Field{
		Description: "Get documents by filters",
		Args:        util.BindArguments(document.DocumentTypeFilterRequest{}),
		Resolve:     adminQueryAllDocumentTypesResolver,
		Type:        util.BindOutput(document.DocumentTypeFilterResponse{}, "List document types", nil),
	}

	(*mutation)["createDocumentType"] = &graphql.Field{
		Description: "Create document type",
		Args:        util.BindArguments(document.DocumentTypeCreateRequest{}),
		Resolve:     adminMutationCreateDocumentTypeResolver,
		Type:        util.BindOutput(document.DocumentTypeCreateResponse{}, "Document type create response", nil),
	}
	(*mutation)["updateDocumentType"] = &graphql.Field{
		Description: "Update document type",
		Args:        util.BindArguments(document.DocumentTypeUpdateRequest{}),
		Resolve:     adminMutationUpdateDocumentTypeResolver,
		Type:        util.BindOutput(document.DocumentTypeUpdateResponse{}, "Document type update response", nil),
	}
	(*mutation)["deleteDocumentType"] = &graphql.Field{
		Description: "Delete document type",
		Args:        util.BindArguments(document.DocumentTypeRemoveRequest{}),
		Resolve:     adminMutationDeleteDocumentTypeResolver,
		Type:        graphql.Boolean,
	}
}

func adminQueryGetDocumentTypeResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &document.DocumentTypeGetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	doc, err := connections.DocumentService.DocumentTypeGet(params.Context, request)
	if err != nil {
		return nil, err
	}

	return doc.GetDocumentTypeDesc(), err
}

func adminQueryAllDocumentTypesResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &document.DocumentTypeFilterRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.DocumentService.DocumentTypeFilter(params.Context, request)
}

func adminMutationCreateDocumentTypeResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &document.DocumentTypeCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.DocumentService.DocumentTypeCreate(params.Context, request)
}

func adminMutationUpdateDocumentTypeResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &document.DocumentTypeUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.DocumentService.DocumentTypeUpdate(params.Context, request)
}

func adminMutationDeleteDocumentTypeResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &document.DocumentTypeRemoveRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	_, err := connections.DocumentService.DocumentTypeRemove(params.Context, request)
	if err != nil {
		return nil, err
	}

	return true, nil
}
