package admin

import (
	"github.com/graphql-go/graphql"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	expenses "gitlab.mashplace.com/mashplace/backend/expenses/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
)

func initExpenses(query, mutation *graphql.Fields) {
	(*query)["records"] = &graphql.Field{
		Description: "List records by filters",
		Args:        util.BindArguments(expenses.RecordFilterListRequest{}),
		Resolve:     adminQueryRecordsResolver,
		Type:        util.BindOutput(expenses.RecordFilterListResponse{}, "Records list", nil),
	}

	(*query)["listScheduledRecords"] = &graphql.Field{
		Description: "List scheduled records including matched records by time and correlation_id",
		Args:        util.BindArguments(expenses.ScheduledRecordListRequest{}),
		Resolve:     adminQueryListScheduledRecords,
		Type:        util.BindOutput(expenses.ScheduledRecordListResponse{}, "Scheduled records list", nil),
	}

	(*mutation)["updateAdvanceStatus"] = &graphql.Field{
		Description: "Updates advance status",
		Args:        util.BindArguments(expenses.ServiceRecordAdvanceUpdateRequest{}),
		Resolve:     adminMutationUpdateAdvanceStatusResolver,
		Type:        util.BindOutput(expenses.ServiceRecordAdvanceUpdateResponse{}, "Advance status update", nil),
	}

	(*mutation)["updateRentPlan"] = &graphql.Field{
		Description: "Updates rent plan",
		Args:        util.BindArguments(expenses.ServiceRentPlanUpdateRequest{}),
		Resolve:     adminMutationUpdateRentPlanResolver,
		Type:        graphql.Boolean,
	}
}

func adminQueryRecordsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &expenses.RecordFilterListRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.ExpensesService.RecordFilterList(params.Context, request)
}

func adminQueryListScheduledRecords(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &expenses.ScheduledRecordListRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.ExpensesService.ScheduledRecordList(params.Context, request)
}

func adminMutationUpdateAdvanceStatusResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &expenses.ServiceRecordAdvanceUpdateRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.ExpensesService.ServiceRecordAdvanceUpdate(params.Context, req)
}

func adminMutationUpdateRentPlanResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &expenses.ServiceRentPlanUpdateRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.ExpensesService.ServiceRentPlanUpdate(params.Context, req)
}
