package admin

import (
	"context"
	"fmt"

	account "gitlab.mashplace.com/mashplace/backend/account/api"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	q "gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	"gitlab.mashplace.com/mashplace/backend/offer/api/rto"

	"github.com/graphql-go/graphql"
)

type AllTransactionsResponse struct {
	Transactions []*invoice.Transaction `json:"transactions,omitempty"`
	Total        int32                  `json:"total"`
}

func initInvoice(query, mutation *graphql.Fields) {
	(*query)["Invoice"] = &graphql.Field{
		Description: "Retrieves specified invoice",
		Args:        util.BindArguments(invoice.InvoiceGetRequest{}),
		Resolve:     adminQueryGetInvoiceResolver,
		Type:        util.BindOutput(invoice.Invoice{}, "Invoice information", nil),
	}
	(*query)["allInvoices"] = &graphql.Field{
		Args:        util.BindArguments(allParams{}),
		Description: "Get all invoices",
		Resolve:     adminQueryAllInvoicesResolver,
		Type:        util.BindOutput(invoice.InvoiceFilterListResponse{}, "Invoices list", nil),
	}
	(*query)["allPayments"] = &graphql.Field{
		Args:        util.BindArguments(allParams{}),
		Description: "Get all payments",
		Resolve:     adminQueryAllPaymentsResolver,
		Type:        util.BindOutput(invoice.PaymentFilterListResponse{}, "Payments list", nil),
	}
	(*query)["downloadInvoicesCsv"] = &graphql.Field{
		DeprecationReason: "use downloadReport",
		Args:              util.BindArguments(invoice.DownloadReportRequest{}),
		Description:       "Download invoices csv file",
		Resolve:           adminQueryDownloadReportResolver,
		Type:              util.BindOutput(invoice.DownloadReportResponse{}, "Invoices csv file", nil),
	}
	(*query)["downloadReport"] = &graphql.Field{
		Args:        util.BindArguments(invoice.DownloadReportRequest{}),
		Description: "Download report",
		Resolve:     adminQueryDownloadReportResolver,
		Type:        util.BindOutput(invoice.DownloadReportResponse{}, "Invoices report file", nil),
	}

	(*query)["Transaction"] = &graphql.Field{
		Description: "Fetches transaction by ID",
		Args:        util.BindArguments(invoice.TransactionGetRequest{}),
		Resolve:     adminQueryGetTransactionResolver,
		Type:        util.BindOutput(invoice.Transaction{}, "Transaction details", nil),
	}

	(*query)["allTransactions"] = &graphql.Field{
		Description: "Fetches transaction list",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllTransactionsResolver,
		Type:        util.BindOutput(AllTransactionsResponse{}, "Transaction list", nil),
	}

	(*mutation)["updateInvoiceItems"] = &graphql.Field{
		Description: "Confirm an invoice item",
		Args:        util.BindArguments(invoice.ServiceItemStatusRequest{}),
		Resolve:     adminMutationUpdateInvoiceItemsResolver,
		Type:        util.BindOutput(invoice.Invoice{}, "Update invoice item status", nil),
	}

	(*mutation)["pullTransactions"] = &graphql.Field{
		Description: "Pulls and matches transactions in given time range",
		Args:        util.BindArguments(invoice.TransactionPullRequest{}),
		Resolve:     adminMutationPullTransactionsResolver,
		Type:        util.BindOutput(invoice.TransactionPullResponse{}, "Transaction pull/match stats", nil),
	}

	(*mutation)["matchTransactions"] = &graphql.Field{
		Description: "Forces repeated match on pulled transaction in given time range",
		Args:        util.BindArguments(invoice.TransactionMatchRequest{}),
		Resolve:     adminMutationMatchTransactionsResolver,
		Type:        util.BindOutput(invoice.TransactionMatchResponse{}, "Transaction match stats", nil),
	}

	(*mutation)["createPayment"] = &graphql.Field{
		Description: "Creates a payment",
		Args:        util.BindArguments(invoice.ServicePaymentCreateRequest{}),
		Resolve:     adminMutationCreatePaymentResolver,
		Type:        util.BindOutput(invoice.ServicePaymentCreateResponse{}, "Payment create response", nil),
	}

	(*mutation)["createInvoice"] = &graphql.Field{
		Description: "Creates invoice",
		Args:        util.BindArguments(invoice.InvoiceCreateRequest{}),
		Resolve:     adminMutationCreateInvoiceResolver,
		Type:        util.BindOutput(invoice.InvoiceCreateResponse{}, "Invoice create response", nil),
	}

	(*mutation)["updateInvoice"] = &graphql.Field{
		Description: "Updates invoice",
		Args:        util.BindArguments(invoice.ServiceInvoiceUpdateRequest{}),
		Resolve:     adminMutationUpdateInvoiceResolver,
		Type:        util.BindOutput(invoice.ServiceInvoiceUpdateResponse{}, "Invoice update response", nil),
	}

	(*mutation)["updatePaymentStatus"] = &graphql.Field{
		Description: "Updates payment status, optionally providing description and manually correlated transaction id",
		Args:        util.BindArguments(invoice.ServicePaymentStatusRequest{}),
		Resolve:     adminMutationUpdatePaymentStatusResolver,
		Type:        util.BindOutput(invoice.ServicePaymentStatusResponse{}, "Payment status update response", nil),
	}

	(*mutation)["addPaymentTransactionMatch"] = &graphql.Field{
		Description: "Creates manually correlated payment<->transaction match",
		Args:        util.BindArguments(invoice.ServicePaymentTransactionAddRequest{}),
		Resolve:     adminMutationAddPaymentTransactionMatchResolver,
		Type: util.BindOutput(
			invoice.ServicePaymentTransactionAddResponse{},
			"Payment-transaction match response",
			nil,
		),
	}

	(*mutation)["subscribeTransactions"] = &graphql.Field{
		Description: "Subscribe admin to new transactions",
		Resolve:     adminMutationSubscribeTransactionsResolver,
		Type:        graphql.Boolean,
	}

	(*mutation)["unsubscribeTransactions"] = &graphql.Field{
		Description: "Unsubscribe admin from new transactions",
		Resolve:     adminMutationUnsubscribeTransactionsResolver,
		Type:        graphql.Boolean,
	}

	if connections.SiteURL != "mashroom.com" {
		(*mutation)["createTransaction"] = &graphql.Field{
			Description: "Create test transaction",
			Args:        util.BindArguments(invoice.ServiceTransactionCreateRequest{}),
			Resolve:     adminMutationCreateTransactionResolver,
			Type: util.BindOutput(
				invoice.ServiceTransactionCreateResponse{},
				"Create transaction response",
				nil,
			),
		}
	}

	(*mutation)["addProductDocuments"] = &graphql.Field{
		Description: "Updates duration of the product, list of documents",
		Args:        util.BindArguments(invoice.ProductAddDocumentsRequest{}),
		Resolve:     adminMutationAddProductDocuments,
		Type:        util.BindOutput(invoice.ProductAddDocumentsResponse{}, "Product response", nil),
	}

	(*mutation)["setAdminInvoice"] = &graphql.Field{
		DeprecationReason: "use updateInvoice",
		Description:       "Add an admin user to an invoice",
		Args:              util.BindArguments(invoice.ServiceInvoiceUpdateRequest{}),
		Resolve:           adminMutationSetAdminInvoiceResolver,
		Type:              graphql.Boolean,
	}
}
func adminQueryDownloadReportResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &invoice.DownloadReportRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.Report == nil && request.ReportType == 0 {
		request.Report = &invoice.DownloadReportRequest_AddonProducts{
			AddonProducts: &invoice.ReportAddonProducts{
				Format: invoice.ReportFormat_CSV,
			},
		}
	}

	return connections.InvoiceService.DownloadReport(params.Context, request)
}

func adminQueryGetInvoiceResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &invoice.InvoiceGetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.InvoiceService.InvoiceGet(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp.GetInvoice(), nil
}

func adminQueryAllInvoicesResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &allParams{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var pagination *query.QueryParamsPagination

	if request.Pagination.Page != 0 && request.Pagination.PerPage != 0 {
		pagination = &query.QueryParamsPagination{
			Page:    request.Pagination.Page,
			PerPage: request.Pagination.PerPage,
		}
	}

	handleAllInvoicesFilter(params.Context, request.Filter)

	resp, err := connections.InvoiceService.InvoiceFilterList(params.Context, &invoice.InvoiceFilterListRequest{
		Params: &query.QueryParams{
			Pagination: pagination,
			Sort:       request.Sort,
			Filter:     request.Filter,
		},
	})
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func handleAllInvoicesFilter(ctx context.Context, filter []*q.QueryParamsFilter) {
	for _, v := range filter {
		if v == nil {
			continue
		}

		for _, f := range v.Fields {
			if f == "userid" {
				ids, err := handleAllInvoicesFindUsers(ctx, v.Value[0])
				if err == nil {
					v.Value = ids
				}
			}
		}
	}
}

func handleAllInvoicesFindUsers(ctx context.Context, username string) (ids []string, err error) {
	accResp, err := connections.AccountService.Get(ctx, &account.GetRequest{Id: username})
	if err == nil && accResp.GetAccount() != nil {
		ids = append(ids, accResp.GetAccount().GetIdValue())
		return
	}

	resp, err := connections.AccountService.ListFilter(ctx, &account.AccountListFilterRequest{
		Params: &account.AccountListFilterRequest_ParamsValue{
			ParamsValue: &q.QueryParams{
				Filter: []*q.QueryParamsFilter{
					{
						Fields:   []string{"details.first_name", "details.last_name", "details.full_name"},
						Operator: q.QueryFilterOperator_CONTAINS_ANY,
						Value:    []string{username},
					},
				},
			}},
	})
	if err != nil {
		return nil, err
	}

	for _, a := range resp.GetAccounts() {
		ids = append(ids, a.GetIdValue())
	}

	return
}

func adminQueryAllPaymentsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &allParams{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var pagination *query.QueryParamsPagination

	if request.Pagination.Page != 0 && request.Pagination.PerPage != 0 {
		pagination = &query.QueryParamsPagination{
			Page:    request.Pagination.Page,
			PerPage: request.Pagination.PerPage,
		}
	}

	resp, err := connections.InvoiceService.PaymentFilterList(params.Context, &invoice.PaymentFilterListRequest{
		Params: &query.QueryParams{
			Pagination: pagination,
			Sort:       request.Sort,
			Filter:     request.Filter,
		},
	})
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func adminQueryGetTransactionResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &invoice.TransactionGetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.InvoiceService.TransactionGet(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp.GetTransaction(), nil
}

func adminQueryAllTransactionsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &allParams{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var pagination *query.QueryParamsPagination

	if request.Pagination.Page != 0 && request.Pagination.PerPage != 0 {
		pagination = &query.QueryParamsPagination{
			Page:    request.Pagination.Page,
			PerPage: request.Pagination.PerPage,
		}
	}

	resp, err := connections.InvoiceService.TransactionFilterList(params.Context, &invoice.TransactionFilterListRequest{
		Params: &query.QueryParams{
			Pagination: pagination,
			Sort:       request.Sort,
			Filter:     request.Filter,
		},
	})
	if err != nil {
		return nil, err
	}

	return &AllTransactionsResponse{
		Transactions: resp.GetTransactions(),
		Total:        resp.GetTotal(),
	}, nil
}

func adminMutationUpdateInvoiceItemsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &invoice.ServiceItemStatusRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.AdminId = params.Context.Value(ws.CtxUser).(*auth.AuthResponse).GetAccount().GetIdValue()

	resp, err := connections.InvoiceService.ServiceItemStatus(params.Context, req)
	if err != nil {
		return nil, err
	}

	return resp.GetInvoice(), nil
}

func adminMutationCreatePaymentResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &invoice.ServicePaymentCreateRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	resp, err := connections.InvoiceService.ServicePaymentCreate(params.Context, req)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func adminMutationUpdatePaymentStatusResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &invoice.ServicePaymentStatusRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.AdminId = params.Context.Value(ws.CtxUser).(*auth.AuthResponse).GetAccount().GetIdValue()

	resp, err := connections.InvoiceService.ServicePaymentStatus(params.Context, req)
	if err != nil {
		return nil, err
	}

	if resp.Payment.CorrelationTarget == invoice.CorrelationTargetStandingOrder {
		_, err = connections.RtoServiceClient.UpdateStateToActive(params.Context,
			&rto.UpdateStateToActiveRequest{Id: resp.Payment.CorrelationId})
		fmt.Printf("Rto.UpdateStateToActive resp.Payment.CorrelationId: %s\r\n", resp.Payment.CorrelationId)
		if err != nil {
			fmt.Printf("Rto.UpdateStateToActive error: %s\r\n", err.Error())
		}
	}

	return resp, nil
}

func adminMutationAddPaymentTransactionMatchResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &invoice.ServicePaymentTransactionAddRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.AdminId = params.Context.Value(ws.CtxUser).(*auth.AuthResponse).GetAccount().GetIdValue()

	resp, err := connections.InvoiceService.ServicePaymentTransactionAdd(params.Context, req)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func adminMutationPullTransactionsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &invoice.TransactionPullRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	resp, err := connections.InvoiceService.TransactionPull(params.Context, req)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func adminMutationMatchTransactionsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &invoice.TransactionMatchRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	resp, err := connections.InvoiceService.TransactionMatch(params.Context, req)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func adminMutationSubscribeTransactionsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user.GetAccount() == nil {
		return nil, errs.ErrNoUser
	}

	_, err := connections.InvoiceService.TransactionSubscribe(params.Context, &invoice.TransactionSubscribeRequest{
		User: user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return true, nil
}

func adminMutationUnsubscribeTransactionsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user.GetAccount() == nil {
		return nil, errs.ErrNoUser
	}

	_, err := connections.InvoiceService.TransactionUnsubscribe(params.Context, &invoice.TransactionUnsubscribeRequest{
		User: user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return true, nil
}

func adminMutationCreateTransactionResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user.GetAccount() == nil {
		return nil, errs.ErrNoUser
	}

	req := &invoice.ServiceTransactionCreateRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.InvoiceService.ServiceTransactionCreate(params.Context, req)
}

func adminMutationAddProductDocuments(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &invoice.ProductAddDocumentsRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.InvoiceService.ProductAddDocuments(params.Context, request)
}

func adminMutationSetAdminInvoiceResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &invoice.ServiceInvoiceUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	_, err := connections.InvoiceService.ServiceInvoiceUpdate(params.Context, req)

	return err == nil, err
}

func adminMutationCreateInvoiceResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &invoice.InvoiceCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.InvoiceService.InvoiceCreate(params.Context, req)
}

func adminMutationUpdateInvoiceResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &invoice.ServiceInvoiceUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.InvoiceService.ServiceInvoiceUpdate(params.Context, req)
}
