package admin

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
)

func initListing(query, mutation *graphql.Fields) {
	(*query)["ListingVersion"] = &graphql.Field{
		Description: "Retrieves property/listing versions list",
		Args:        util.BindArguments(listing.VersionGetRequest{}),
		Resolve:     adminQueryListingVersionResolver,
		Type:        util.BindOutput(listing.Version{}, "Version", nil),
	}
	(*query)["allListingVersions"] = &graphql.Field{
		Description: "Retrieves property/listing versions list",
		Args:        util.BindArguments(listing.VersionFilterRequest{}),
		Resolve:     adminQueryAllListingVersionsResolver,
		Type:        util.BindOutput(listing.VersionFilterResponse{}, "Version list", nil),
	}
	(*mutation)["createVersion"] = &graphql.Field{
		Description: "Create property",
		Args:        util.BindArguments(listing.VersionCreateRequest{}),
		Resolve:     adminMutationListingCreateResolver,
		Type:        util.BindOutput(listing.VersionCreateResponse{}, "Property id", nil),
	}
	(*mutation)["publishVersion"] = &graphql.Field{
		Description: "Publish property",
		Args:        util.BindArguments(listing.VersionPublishRequest{}),
		Resolve:     adminMutationListingPublishResolver,
		Type:        util.BindOutput(listing.VersionPublishResponse{}, "Property ids", nil),
	}
	(*mutation)["unpublishVersion"] = &graphql.Field{
		Description: "Delete published property",
		Args:        util.BindArguments(listing.VersionUnpublishRequest{}),
		Resolve:     adminMutationListingUnpublishResolver,
		Type:        util.BindOutput(listing.VersionUnpublishResponse{}, "Delete result", nil),
	}
	(*mutation)["flagUpdate"] = &graphql.Field{
		Description: "Versions flag update",
		Args:        util.BindArguments(listing.VersionFlagsRequest{}),
		Resolve:     ListingFlagUpdateResolver,
		Type:        util.BindOutput(listing.VersionFlagsResponse{}, "Flag update result", nil),
	}
}

func adminQueryListingVersionResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &listing.VersionGetRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.ListingService.VersionGet(params.Context, req)
}

func adminQueryAllListingVersionsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &listing.VersionFilterRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.ListingService.VersionFilter(params.Context, req)
}

func adminMutationListingCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	req := &listing.VersionCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.Spec = &listing.VersionSpec{
		UserId: user.GetAccount().GetIdValue(),
	}

	return connections.ListingService.VersionCreate(params.Context, req)
}

func adminMutationListingPublishResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	req := &listing.VersionPublishRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ListingService.VersionPublish(params.Context, req)
}

func adminMutationListingUnpublishResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	req := &listing.VersionUnpublishRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ListingService.VersionUnpublish(params.Context, req)
}

func ListingFlagUpdateResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &listing.VersionFlagsRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.ListingService.VersionFlags(params.Context, req)
}
