package admin

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	q "gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
	notes "gitlab.mashplace.com/mashplace/backend/notes/api"
)

func initNotes(query, mutation *graphql.Fields) {
	(*query)["allNotes"] = &graphql.Field{
		Description: "List all notes",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllNotesResolver,
		Type:        util.BindOutput(notes.NoteFilterResponse{}, "Notes filtered list", nil),
	}

	(*mutation)["createNote"] = &graphql.Field{
		Description: "Create note",
		Args:        util.BindArguments(notes.NoteCreateRequest{}),
		Resolve:     adminMutationCreateNoteResolver,
		Type:        util.BindOutput(notes.NoteCreateResponse{}, "Create note response", nil),
	}
	(*mutation)["updateNote"] = &graphql.Field{
		Description: "Update note",
		Args:        util.BindArguments(notes.NoteUpdateRequest{}),
		Resolve:     adminMutationUpdateNoteResolver,
		Type:        util.BindOutput(notes.NoteUpdateResponse{}, "Update note response", nil),
	}
	(*mutation)["deleteNote"] = &graphql.Field{
		Description: "Delete note",
		Args:        util.BindArguments(notes.NoteDeleteRequest{}),
		Resolve:     adminMutationDeleteNoteResolver,
		Type:        graphql.Boolean,
	}
}

func adminQueryAllNotesResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	input := &allParams{}

	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}

	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "created", Order: -1}}
	}

	return connections.NotesService.NoteFilter(params.Context, &notes.NoteFilterRequest{
		Params: req,
	})
}

func adminMutationCreateNoteResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &notes.NoteCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	req.User = user.GetAccount().GetIdValue()

	return connections.NotesService.NoteCreate(params.Context, req)
}

func adminMutationUpdateNoteResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &notes.NoteUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.NotesService.NoteUpdate(params.Context, req)
}

func adminMutationDeleteNoteResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &notes.NoteDeleteRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	_, err := connections.NotesService.NoteDelete(params.Context, req)
	if err != nil {
		return nil, err
	}

	return true, nil
}
