package admin

import (
	"github.com/graphql-go/graphql"
	q "gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
)

func initNotification(query, mutation *graphql.Fields) {
	(*query)["allSendgridListsSubscriptions"] = &graphql.Field{
		Description: "List all sendgrid list subscriptions",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllSendgridListsSubscriptions,
		Type: util.BindOutput(
			notification.SendgridListSubscriptionListResponse{},
			"Subscriptions list",
			nil,
		),
	}
	(*query)["notificationDefinitionList"] = &graphql.Field{
		Description: "List notification definitions",
		Resolve:     adminNotificationDefinitionListResolver,
		Args:        util.BindArguments(notification.NotificationDefinitionListRequest{}),
		Type: util.BindOutput(
			notification.NotificationDefinitionListResponse{},
			"All Notification Definitions matching query",
			nil,
		),
	}

	(*mutation)["sendgridListsSubscriptionsFetch"] = &graphql.Field{
		Description: "Fetch subscription details for given emails",
		Args:        util.BindArguments(notification.SendgridListSubscriptionFetchRequest{}),
		Resolve:     adminMutationSendgridListsSubscriptionsFetch,
		Type: util.BindOutput(
			notification.SendgridListSubscriptionFetchResponse{},
			"Fetch result",
			nil,
		),
	}
	(*mutation)["sendgridListsSubscriptionsForceRebuild"] = &graphql.Field{
		Description: "Force rebuild subscriptions cache for entire platform",
		Resolve:     adminMutationSendgridListsSubscriptionsForceRebuild,
		Type:        graphql.Boolean,
	}
	(*mutation)["notificationDefinitionEdit"] = &graphql.Field{
		Description: "Create or update a notification definition",
		Resolve:     adminNotificationDefinitionEditResolver,
		Args:        util.BindArguments(notification.NotificationDefinitionEditRequest{}),
		Type: util.BindOutput(
			notification.NotificationDefinitionEditResponse{},
			"Notification Definition after edit",
			nil,
		),
	}
}

func adminQueryAllSendgridListsSubscriptions(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	input := &allParams{}

	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}

	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "created", Order: -1}}
	}

	return connections.NotificationService.SendgridListSubscriptionList(
		params.Context,
		&notification.SendgridListSubscriptionListRequest{
			Params: req,
		},
	)
}

func adminNotificationDefinitionListResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &notification.NotificationDefinitionListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.Params == nil {
		request.Params = &q.QueryParams{}
	}

	return connections.NotificationService.NotificationDefinitionList(params.Context, request)
}

func adminMutationSendgridListsSubscriptionsFetch(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &notification.SendgridListSubscriptionFetchRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.NotificationService.SendgridListSubscriptionFetch(
		params.Context,
		req,
	)
}

func adminMutationSendgridListsSubscriptionsForceRebuild(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	_, err := connections.NotificationService.SendgridListSubscriptionForceRebuild(
		params.Context,
		&notification.SendgridListSubscriptionForceRebuildRequest{},
	)
	if err != nil {
		return nil, err
	}

	return true, nil
}

func adminNotificationDefinitionEditResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &notification.NotificationDefinitionEditRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	response, err := connections.NotificationService.NotificationDefinitionEdit(params.Context, request)
	if err != nil {
		return nil, err
	}

	return response, nil
}
