package admin

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
	rightmove "gitlab.mashplace.com/mashplace/backend/rightmove/api"
)

func initRightmove(query, mutation *graphql.Fields) {
	(*mutation)["publishRightmove"] = &graphql.Field{
		Description:       "Publish property on Rightmove",
		DeprecationReason: "use publishVersion",
		Args:              util.BindArguments(getParams{}),
		Resolve:           RightmovePublishResolver,
		Type:              util.BindOutput(rightmove.SendPropertyResponse{}, "Property ids", nil),
	}
	(*mutation)["deleteRightmove"] = &graphql.Field{
		Description:       "Delete published property from Rightmove",
		DeprecationReason: "use unpublishVersion",
		Args:              util.BindArguments(getParams{}),
		Resolve:           RightmoveDeleteResolver,
		Type:              util.BindOutput(rightmove.DeletePropertyResponse{}, "Delete result", nil),
	}
}

func RightmovePublishResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &getParams{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	idv, err := connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
		Spec: listing.VersionSpecPropertyID("", req.ID, listing.VersionStatus_VERSION_VERIFIED),
	})
	if err != nil {
		return nil, err
	}

	resp, err := connections.ListingService.VersionPublish(params.Context, &listing.VersionPublishRequest{
		UserId:    user.GetAccount().GetIdValue(),
		VersionId: idv.GetVersion().GetId(),
		Rightmove: &listing.VersionPublicationChangeItem{},
	})
	if err != nil {
		return nil, err
	}

	return &rightmove.SendPropertyResponse{
		Success:      true,
		Errors:       nil,
		ErrorMsg:     "",
		Property:     req.ID,
		RightmoveId:  resp.GetVersion().GetPublication().GetRightmove().GetId(),
		RightmoveUrl: resp.GetVersion().GetPublication().GetRightmove().GetUrl(),
	}, nil
}

func RightmoveDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &getParams{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	idv, err := connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
		Spec: listing.VersionSpecPropertyID(
			"",
			req.ID,
			listing.VersionStatus_VERSION_VERIFIED,
		),
	})
	if err != nil {
		idv, err = connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
			Spec: listing.VersionSpecPropertyID(
				"",
				req.ID,
				listing.VersionStatus_VERSION_VERIFIED,
				listing.VersionStatus_VERSION_VERIFYING,
				listing.VersionStatus_VERSION_DRAFT,
				listing.VersionStatus_VERSION_LET,
			),
		})
		if err != nil {
			return nil, err
		}
	}

	_, err = connections.ListingService.VersionUnpublish(params.Context, &listing.VersionUnpublishRequest{
		UserId:    user.GetAccount().GetIdValue(),
		VersionId: idv.GetVersion().GetId(),
		Rightmove: true,
	})
	if err != nil {
		return nil, err
	}

	return &rightmove.DeletePropertyResponse{
		PropertyId: req.ID,
		Success:    true,
	}, nil
}
