package admin

import (
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
	"gitlab.mashplace.com/mashplace/backend/offer/api/rto"

	"github.com/graphql-go/graphql"
)

func initRto(query, mutation *graphql.Fields) {
	(*query)["rtosByLandlordId"] = &graphql.Field{
		Description: "Retrieves rto by landlord id",
		Args:        util.BindArguments(rto.GetByLandlordIDRequest{}),
		Resolve:     adminQueryRtoGetByLandlordIDResolver,
		Type:        util.BindOutput(rto.RtosResponse{}, "Requested rtos by property ids", nil),
	}

	(*mutation)["rtoCreate"] = &graphql.Field{
		Description: "Create rto",
		Args:        util.BindArguments(rto.CreateRequest{}),
		Resolve:     adminMutationRtoCreateResolver,
		Type:        util.BindOutput(rto.RtoResponse{}, "Create rto", nil),
	}

	(*mutation)["rtoCreateTenancy"] = &graphql.Field{
		Description: "Create rto tenancy",
		Args:        util.BindArguments(rto.CreateTenancyRequest{}),
		Resolve:     adminMutationRtoCreateTenancyResolver,
		Type:        util.BindOutput(rto.RtoResponse{}, "Create rto tenancy", nil),
	}

	(*mutation)["rtoCancel"] = &graphql.Field{
		Description: "Cancel rto",
		Args:        util.BindArguments(rto.CancelRequest{}),
		Resolve:     adminMutationRtoCancelResolver,
		Type:        util.BindOutput(rto.RtoResponse{}, "Cancel rto", nil),
	}
}

func adminQueryRtoGetByLandlordIDResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &rto.GetByLandlordIDRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.RtoServiceClient.GetByLandlordID(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func adminMutationRtoCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &rto.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.RtoServiceClient.Create(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func adminMutationRtoCreateTenancyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &rto.CreateTenancyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.RtoServiceClient.CreateTenancy(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func adminMutationRtoCancelResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &rto.CancelRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.RtoServiceClient.Cancel(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}
