package admin

import (
	"github.com/graphql-go/graphql"
	q "gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
)

type allServicesResult struct {
	Data  []*extras.Service `json:"data,omitempty"`
	Total int32             `json:"total,omitempty"`
}

func initServices(query, mutation *graphql.Fields) {
	createMethods := func(query, muration *graphql.Fields, title, name, depr string) {
		(*query)[title] = &graphql.Field{
			DeprecationReason: depr,
			Description:       "Retrieves specified " + name,
			Args:              util.BindArguments(extras.GetRequest{}),
			Resolve:           adminQueryGetServiceResolver,
			Type:              util.BindOutput(extras.Service{}, title+" information", nil),
		}
		(*query)["all"+title+"s"] = &graphql.Field{
			DeprecationReason: depr,
			Args:              util.BindArguments(allParams{}),
			Description:       "Get all " + title + "s",
			Resolve:           adminQueryAllServicesResolver,
			Type:              util.BindOutput(allServicesResult{}, title+"s list", nil),
		}
		(*query)["PriceCalculate"] = &graphql.Field{
			DeprecationReason: depr,
			Description:       "Calculates service price",
			Args:              util.BindArguments(extras.PriceCalculateRequest{}),
			Resolve:           adminQueryServicePriceCalculateResolver,
			Type:              util.BindOutput(extras.PriceCalculateResponse{}, "Calculation response", nil),
		}

		(*mutation)["add"+title] = &graphql.Field{
			DeprecationReason: depr,
			Description:       "Adds a " + name,
			Args:              util.BindArguments(extras.Service{}),
			Resolve:           adminMutationAddServiceResolver,
			Type:              util.BindOutput(extras.Service{}, title+" information", nil),
		}
		(*mutation)["delete"+title] = &graphql.Field{
			DeprecationReason: depr,
			Description:       "Deletes a " + name,
			Args:              util.BindArguments(extras.DeleteRequest{}),
			Resolve:           adminMutationDeleteServiceResolver,
			Type:              graphql.Boolean,
		}
		(*mutation)["update"+title] = &graphql.Field{
			DeprecationReason: depr,
			Description:       "Updates a " + name,
			Args:              util.BindArguments(extras.Service{}),
			Resolve:           adminMutationUpdateServiceResolver,
			Type:              graphql.Boolean,
		}
	}

	createMethods(query, mutation, "Service", "service", "use Products")
	createMethods(query, mutation, "Product", "product", "")
}

func adminQueryGetServiceResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &extras.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.GetProperty() == nil {
		request.Property = &extras.Property{}
	}

	if request.GetProperty().GetId() == "" {
		request.Property.Id = request.GetPropertyId()
	}

	svc, err := connections.ExtrasService.Get(params.Context, request)
	if err != nil {
		return nil, err
	}

	return svc.GetExtraService(), err
}

func adminQueryAllServicesResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &allParams{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	req := &q.QueryParams{
		Filter: request.Filter,
		Sort:   request.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    request.Pagination.Page,
			PerPage: request.Pagination.PerPage,
		},
	}

	if req.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "created", Order: -1}}
	}

	svc, err := connections.ExtrasService.Search(params.Context, &extras.SearchRequest{
		Params: req,
	})
	if err != nil {
		return nil, err
	}

	return &allServicesResult{
		Data:  svc.GetServices(),
		Total: svc.GetTotal(),
	}, err
}

func adminQueryServicePriceCalculateResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &extras.PriceCalculateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.ExtrasService.PriceCalculate(params.Context, request)
}

func adminMutationAddServiceResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &extras.Service{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.ExtrasService.Add(params.Context, &extras.AddRequest{
		ExtraService: request,
	})
	if err != nil {
		return nil, err
	}

	svc, err := connections.ExtrasService.Get(params.Context, &extras.GetRequest{
		Id: resp.GetId(),
	})
	if err != nil {
		return nil, err
	}

	return svc.GetExtraService(), nil
}

func adminMutationDeleteServiceResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &extras.DeleteRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.ExtrasService.Delete(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp.GetSuccess(), nil
}

func adminMutationUpdateServiceResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &extras.Service{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.ExtrasService.Update(params.Context, &extras.UpdateRequest{
		ExtraService: request,
	})
	if err != nil {
		return nil, err
	}

	return resp.GetSuccess(), nil
}
