package admin

import (
	"github.com/graphql-go/graphql"
	account "gitlab.mashplace.com/mashplace/backend/account/api"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	q "gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/enrichment"
)

type allUsersResult struct {
	Data  []*account.Account `json:"data,omitempty"`
	Total int32              `json:"total,omitempty"`
}

func initUser(query, mutation *graphql.Fields) {
	(*query)["User"] = &graphql.Field{
		Description: "Get user by id",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminQueryGetUserResolver,
		Type:        util.BindOutput(account.Account{}, "User get response", nil),
	}
	(*query)["UserByEmail"] = &graphql.Field{
		Description: "Get user by email",
		Args:        util.BindArguments(account.GetByEmailRequest{}),
		Resolve:     adminQueryGetUserByEmailResolver,
		Type:        util.BindOutput(account.Account{}, "User get by email response", nil),
	}
	(*query)["allUsers"] = &graphql.Field{
		Description: "Returns users list",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllUsersResolver,
		Type:        util.BindOutput(allUsersResult{}, "User list", nil),
	}
	(*query)["downloadAllUsersCsv"] = &graphql.Field{
		Description: "Returns users list",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllUsersCsvResolver,
		Type:        util.BindOutput(account.DownloadUsersCsvResponse{}, "Download users list csv response", nil),
	}

	(*mutation)["createUser"] = &graphql.Field{
		Description: "Create user account",
		Args:        util.BindArguments(auth.SignupRequest{}),
		Resolve:     adminMutationCreateUserResolver,
		Type:        util.BindOutput(auth.CreateResponse{}, "User create response", nil),
	}
	(*mutation)["deleteUser"] = &graphql.Field{
		Description: "Delete user account",
		Args:        util.BindArguments(account.ServiceDeleteRequest{}),
		Resolve:     adminMutationDeleteUserResolver,
		Type:        util.BindOutput(account.ServiceDeleteResponse{}, "Account delete response", nil),
	}
	(*mutation)["updateUser"] = &graphql.Field{
		Description: "Update user account",
		Args:        util.BindArguments(account.ServiceUpdateRequest{}),
		Resolve:     adminMutationUpdateUserResolver,
		Type:        util.BindOutput(account.Account{}, "Account update response", nil),
	}
}

func adminQueryGetUserResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &account.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.Admin = false
	request.Soft = true

	resp, err := connections.AccountService.Get(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp.GetAccount(), nil
}

func adminQueryGetUserByEmailResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &account.GetByEmailRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.Admin = false

	resp, err := connections.AccountService.GetByEmail(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp.GetAccount(), nil
}

func adminQueryAllUsersResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	input := &allParams{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}
	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}
	request := &account.AccountListFilterRequest{
		Params: &account.AccountListFilterRequest_ParamsValue{ParamsValue: req},
	}
	request.Admin = false
	resp, err := connections.AccountService.ListFilter(params.Context, request)
	if err != nil {
		return nil, err
	}

	accounts := resp.GetAccounts()
	total := resp.GetTotal()

	enrichment.ResolveAccountResources(params.Context, accounts...)

	return allUsersResult{
		Data:  accounts,
		Total: total,
	}, nil
}

func adminQueryAllUsersCsvResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	input := &allParams{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}
	req := &q.QueryParams{
		Filter:     input.Filter,
		Sort:       input.Sort,
		Pagination: nil,
	}
	request := &account.DownloadUsersCsvRequest{
		Params: &account.DownloadUsersCsvRequest_ParamsValue{ParamsValue: req},
	}
	request.Admin = false
	resp, err := connections.AccountService.DownloadUsersCsv(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func adminMutationCreateUserResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &auth.SignupRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.Account == nil {
		request.AdminSite = &auth.SignupRequest_AdminSiteValue{AdminSiteValue: false}
	}

	if request.Account.Admin == nil {
		request.Account.Admin = &account.Account_AdminValue{AdminValue: false}
	}

	return connections.AuthService.Create(params.Context, request)
}

func adminMutationUpdateUserResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &account.ServiceUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.Account.Admin = nil
	resp, err := connections.AccountService.ServiceUpdate(params.Context, request)
	if err != nil {
		return nil, err
	}

	if request.GetNewPassword() != nil {
		resp, err := connections.AuthService.ServiceUpdatePassword(params.Context, &auth.ServiceUpdatePasswordRequest{
			UserId:      request.GetAccount().GetIdValue(),
			NewPassword: request.GetNewPasswordValue(),
		})

		if err != nil || resp == nil || !resp.Success || len(resp.Errors) > 0 {
			return resp, err
		}
	}

	return resp.GetAccount(), nil
}

func adminMutationDeleteUserResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &account.ServiceDeleteRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	resp, err := connections.AccountService.ServiceDelete(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}
