package admin

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
	zoopla "gitlab.mashplace.com/mashplace/backend/zoopla/api"
)

type getParamsOpt struct {
	Ids []string `json:"ids"`
}

type deleteZooplaResponse struct {
	Result []*zoopla.DeletePropertyResponse `json:"result"`
}

func initZoopla(query, mutation *graphql.Fields) {
	(*mutation)["publishZoopla"] = &graphql.Field{
		Description:       "Publish property on Zoopla",
		DeprecationReason: "use publishVersion",
		Args:              util.BindArguments(getParams{}),
		Resolve:           ZooplaPublishResolver,
		Type:              util.BindOutput(zoopla.SendPropertyResponse{}, "Property ids", nil),
	}
	(*mutation)["deleteZoopla"] = &graphql.Field{
		Description:       "Delete published property from Zoopla",
		DeprecationReason: "use unpublishVersion",
		Args:              util.BindArguments(getParams{}),
		Resolve:           ZooplaDeleteResolver,
		Type:              util.BindOutput(zoopla.DeletePropertyResponse{}, "Delete result", nil),
	}
	(*mutation)["deleteZooplaList"] = &graphql.Field{
		Description:       "Delete multiple published property from Zoopla",
		DeprecationReason: "use unpublishVersion",
		Args:              util.BindArguments(getParamsOpt{}),
		Resolve:           ZooplaDeleteListResolver,
		Type:              util.BindOutput(deleteZooplaResponse{}, "Delete result", nil),
	}
}

func ZooplaPublishResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	idv, err := connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
		Spec: listing.VersionSpecPropertyID("", req.ID, listing.VersionStatus_VERSION_VERIFIED),
	})
	if err != nil {
		return nil, err
	}

	resp, err := connections.ListingService.VersionPublish(params.Context, &listing.VersionPublishRequest{
		UserId:    user.GetAccount().GetIdValue(),
		VersionId: idv.GetVersion().GetId(),
		Zoopla:    &listing.VersionPublicationChangeItem{},
	})
	if err != nil {
		return nil, err
	}

	return &zoopla.SendPropertyResponse{
		Success:   true,
		Errors:    nil,
		ErrorMsg:  "",
		Property:  req.ID,
		ZooplaId:  resp.GetVersion().GetPublication().GetZoopla().GetId(),
		ZooplaUrl: resp.GetVersion().GetPublication().GetZoopla().GetUrl(),
	}, nil
}

func ZooplaDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &getParams{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	idv, err := connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
		Spec: listing.VersionSpecPropertyID(
			"",
			req.ID,
			listing.VersionStatus_VERSION_VERIFIED,
		),
	})
	if err != nil {
		idv, err = connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
			Spec: listing.VersionSpecPropertyID(
				"",
				req.ID,
				listing.VersionStatus_VERSION_VERIFIED,
				listing.VersionStatus_VERSION_VERIFYING,
				listing.VersionStatus_VERSION_DRAFT,
				listing.VersionStatus_VERSION_LET,
			),
		})
		if err != nil {
			return nil, err
		}
	}

	_, err = connections.ListingService.VersionUnpublish(params.Context, &listing.VersionUnpublishRequest{
		UserId:    user.GetAccount().GetIdValue(),
		VersionId: idv.GetVersion().GetId(),
		Zoopla:    true,
	})
	if err != nil {
		return nil, err
	}

	return &zoopla.DeletePropertyResponse{
		PropertyId: req.ID,
		Success:    true,
	}, nil
}

func ZooplaDeleteListResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &getParamsOpt{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	acc := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	var res []*zoopla.DeletePropertyResponse

	for _, id := range req.Ids {
		idv, err := connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
			Spec: listing.VersionSpecPropertyID(
				"",
				id,
				listing.VersionStatus_VERSION_VERIFIED,
			),
		})
		if err != nil {
			idv, err = connections.ListingService.VersionGet(params.Context, &listing.VersionGetRequest{
				Spec: listing.VersionSpecPropertyID(
					"",
					id,
					listing.VersionStatus_VERSION_VERIFIED,
					listing.VersionStatus_VERSION_VERIFYING,
					listing.VersionStatus_VERSION_DRAFT,
					listing.VersionStatus_VERSION_LET,
				),
			})
			if err != nil {
				return nil, err
			}
		}

		_, err = connections.ListingService.VersionUnpublish(params.Context, &listing.VersionUnpublishRequest{
			UserId:    acc.GetAccount().GetIdValue(),
			VersionId: idv.GetVersion().GetId(),
			Zoopla:    true,
		})
		if err != nil {
			return nil, err
		}

		if err == nil {
			res = append(res, &zoopla.DeletePropertyResponse{
				PropertyId: id,
				Success:    true,
			})
		}
	}

	return deleteZooplaResponse{Result: res}, nil
}
