package resolvers

import (
	"fmt"

	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	cardpayment "gitlab.mashplace.com/mashplace/backend/cardpayment/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
)

type MarkReceiptPaidRequest struct {
	ID string `json:"id,omitempty"`
}

type CardpaymentFormArguments struct {
	cardpayment.FormReceiptRequest
	Items []string `json:"items"`
}

func initCardPayment(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "CardPaymentQuery",
		Description: "Card payment query root",
		Fields: graphql.Fields{
			"user": &graphql.Field{
				DeprecationReason: "use invoice list",
				Description:       "Get current user receipts",
				Resolve:           CardPaymentQueryUserResolver,
				Type:              util.BindOutput(cardpayment.GetUserResponse{}, "Receipt list", nil),
			},
			"property": &graphql.Field{
				DeprecationReason: "Use property { products } enrichment or invoice list",
				Description:       "Get services list",
				Args:              util.BindArguments(cardpayment.GetPropertyRequest{}),
				Resolve:           CardPaymentQueryPropertyResolver,
				Type:              util.BindOutput(cardpayment.GetPropertyResponse{}, "Receipt list", nil),
			},
		},
	}))

	(*rootQueries)["cardpayment"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "CardPaymentMutation",
		Description: "Card payment mutation root",
		Fields: graphql.Fields{
			"form": &graphql.Field{
				DeprecationReason: "use orders",
				Description:       "Form a receipt",
				Args:              util.BindArguments(CardpaymentFormArguments{}),
				Resolve:           CardPaymentMutationFormResolver,
				Type: util.BindOutput(
					cardpayment.FormReceiptResponse{},
					"Payment intent id and secret",
					nil,
				),
			},
			"cancel": &graphql.Field{
				DeprecationReason: "use orders",
				Description:       "Cancel a receipt",
				Args:              util.BindArguments(cardpayment.CancelReceiptRequest{}),
				Resolve:           CardPaymentMutationCancelResolver,
				Type:              util.BindOutput(cardpayment.CancelReceiptResponse{}, "Cancellation status", nil),
			},
			"check": &graphql.Field{
				Description: "Check receipt status",
				Args:        util.BindArguments(cardpayment.CheckReceiptRequest{}),
				Resolve:     CardPaymentMutationCheckResolver,
				Type:        util.BindOutput(cardpayment.CheckReceiptResponse{}, "Actual state of receipt", nil),
			},
			"mark": &graphql.Field{
				DeprecationReason: "use check instead",
				Description:       "Mark as paid",
				Args:              util.BindArguments(cardpayment.MarkReceiptPaidRequest{}),
				Resolve:           CardPaymentMutationMarkResolver,
				Type:              util.BindOutput(cardpayment.MarkReceiptPaidResponse{}, "", nil),
			},
			"markbank": &graphql.Field{
				DeprecationReason: "use orders/check",
				Description:       "Mark as paid via bank",
				Args:              util.BindArguments(cardpayment.MarkReceiptPaidRequest{}),
				Resolve:           CardPaymentMutationMarkBankResolver,
				Type:              util.BindOutput(cardpayment.MarkReceiptPaidResponse{}, "", nil),
			},
		},
	}))

	(*rootMutations)["cardpayment"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func CardPaymentQueryUserResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &cardpayment.GetUserRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var products []*invoice.Product

	resp, err := connections.InvoiceService.InvoiceList(params.Context, &invoice.InvoiceListRequest{
		User: user.GetAccount().GetIdValue(),
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
	})
	if err != nil {
		return products, nil
	}

	return &cardpayment.GetUserResponse{
		Payments: RenderCardpaymentReceipts(resp.GetInvoices()),
	}, nil
}

func RenderCardpaymentReceipts(invoices []*invoice.Invoice) (receipts []*cardpayment.Receipt) {
	for _, inv := range invoices {
		for _, p := range inv.GetPayments() {
			if p.GetSystem().GetStripe() == nil ||
				(p.GetStatus() != invoice.PaymentStatus_PAYMENT_SUCCESS &&
					p.GetStatus() != invoice.PaymentStatus_PAYMENT_CONFIRM) {
				continue
			}

			var services []*extras.ServiceSnapshot

			for _, s := range inv.GetServices() {
				if s.GetService().GetServiceId() == "" {
					continue
				}

				services = append(services, s.GetService())
			}

			if len(services) == 0 {
				continue
			}

			var status cardpayment.PaymentStatus

			switch p.GetStatus() {
			case invoice.PaymentStatus_PAYMENT_SUCCESS:
				status = cardpayment.PaymentStatus_SUCCEEDED
			case invoice.PaymentStatus_PAYMENT_PENDING:
				status = cardpayment.PaymentStatus_UNPAID
			case invoice.PaymentStatus_PAYMENT_FAILED:
				status = cardpayment.PaymentStatus_FAILED
			case invoice.PaymentStatus_PAYMENT_CONFIRM:
				status = cardpayment.PaymentStatus_PROCESSING
			case invoice.PaymentStatus_PAYMENT_CANCELED:
				status = cardpayment.PaymentStatus_CANCELLED
			}

			receipts = append(receipts, &cardpayment.Receipt{
				Id:            p.GetPaymentId(),
				Type:          p.GetSystem().GetStripe().GetRequest().GetType(),
				ClientSecret:  p.GetSystem().GetStripe().GetResponse().GetClientSecret(),
				UserId:        inv.GetUser(),
				UserEmail:     "",
				Items:         services,
				Property:      inv.GetProperty(),
				Status:        status,
				CreatedAt:     p.GetCreatedAt(),
				UpdatedAt:     p.GetUpdatedAt(),
				InvoiceNumber: inv.GetInvoiceId(),
			})
		}
	}

	return
}

func CardPaymentQueryPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &cardpayment.GetPropertyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var products []*invoice.Product

	resp, err := connections.InvoiceService.InvoiceList(params.Context, &invoice.InvoiceListRequest{
		User:     user.GetAccount().GetIdValue(),
		Property: request.GetProperty(),
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
	})
	if err != nil {
		return products, nil
	}

	return &cardpayment.GetPropertyResponse{
		Payments: RenderCardpaymentReceipts(resp.GetInvoices()),
	}, nil
}

func CardPaymentMutationCancelResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &cardpayment.CancelReceiptRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.UserId = user.GetAccount().GetIdValue()
	return connections.CardPaymentService.CancelReceipt(params.Context, request)
}

func CardPaymentMutationCheckResolver(params graphql.ResolveParams) (interface{}, error) {
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)

	request := &cardpayment.CheckReceiptRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	switch {
	case user != nil:
		request.UserId = user.GetAccount().GetIdValue()
	case tuser != nil:
		request.UserId = tuser.Session
	default:
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	return connections.CardPaymentService.CheckReceipt(params.Context, request)
}

func CardPaymentMutationFormResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &CardpaymentFormArguments{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.UserId = user.GetAccount().GetIdValue()
	request.UserEmail = user.GetAccount().GetContactValue().GetEmailPrimaryValue()

	resp, err := connections.InvoiceService.OrderList(params.Context, &invoice.OrderListRequest{
		Userid: request.UserId,
		Params: &query.QueryParams{
			Filter: []*query.QueryParamsFilter{
				{
					Fields:   []string{"correlation_type"},
					Value:    []string{invoice.CorrelationTypeProperty},
					Operator: query.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"correlation_id"},
					Value:    []string{request.GetProperty()},
					Operator: query.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"correlation_target"},
					Value:    []string{invoice.CorrelationTargetListingServices},
					Operator: query.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"status"},
					Value:    []string{invoice.OrderStatus_ORDER_PENDING.String()},
					Operator: query.QueryFilterOperator_EQUALS,
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	if len(resp.GetOrders()) == 0 {
		return nil, grpcerr.Newf(grpcerr.InvalidState, "no pending orders for property %s", request.GetProperty())
	}

	ord := resp.GetOrders()[0]

	comm, err := connections.InvoiceService.OrderCommit(params.Context, &invoice.OrderCommitRequest{
		Userid: request.UserId,
		Id:     ord.GetId(),
	})
	if err != nil {
		return nil, err
	}

	pay, err := connections.InvoiceService.PaymentCreate(params.Context, &invoice.PaymentCreateRequest{
		InvoiceId: comm.GetInvoice().GetInvoiceId(),
		User:      user.GetAccount().GetIdValue(),
		System: &invoice.PaymentSystemDataRequest{
			Data: &invoice.PaymentSystemDataRequest_Stripe{
				Stripe: &invoice.PaymentSystemStripeRequest{
					Type: cardpayment.PaymentType_CARD,
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	return &cardpayment.FormReceiptResponse{
		Id:           pay.GetPayment().GetSystem().GetStripe().GetResponse().GetStripeId(),
		ClientSecret: pay.GetPayment().GetSystem().GetStripe().GetResponse().GetClientSecret(),
	}, nil
}

func CardPaymentMutationMarkResolver(params graphql.ResolveParams) (interface{}, error) {
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)

	request := &cardpayment.MarkReceiptPaidRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	switch {
	case user != nil:
		request.UserId = user.GetAccount().GetIdValue()
	case tuser != nil:
		request.UserId = tuser.Session
	default:
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	_, err := connections.CardPaymentService.CheckReceipt(params.Context, &cardpayment.CheckReceiptRequest{
		Id:     request.UserId,
		UserId: user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return &cardpayment.MarkReceiptPaidResponse{
		Success: true,
	}, nil
}

func CardPaymentMutationMarkBankResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	req := &MarkReceiptPaidRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	_, err := connections.CardPaymentService.CheckReceipt(params.Context, &cardpayment.CheckReceiptRequest{
		Id:     req.ID,
		UserId: user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return &cardpayment.MarkReceiptPaidResponse{
		Success: true,
	}, nil
}
