package resolvers

import (
	"github.com/graphql-go/graphql"
	comments "gitlab.mashplace.com/mashplace/backend/comments/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
)

type commentsPagination struct {
	CorrelationType comments.CorrelationType   `json:"correlation_type,omitempty"`
	CorrelationID   string                     `json:"correlation_id,omitempty"`
	ParentID        string                     `json:"parent_id,omitempty"`
	Page            int32                      `json:"page,omitempty"`
	PerPage         int32                      `json:"per_page,omitempty"`
	Filter          []*query.QueryParamsFilter `json:"filter,omitempty"`
	Sort            []*query.QueryParamsSort   `json:"sort,omitempty"`
}

type commentsNestedPagination struct {
	Limit  int32                      `json:"limit,omitempty"`
	Filter []*query.QueryParamsFilter `json:"filter,omitempty"`
	Sort   []*query.QueryParamsSort   `json:"sort,omitempty"`
}

func initComments(rootQueries, rootMutations *graphql.Fields) {
	q := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "CommentsQuery",
		Description: "Comments query root",
		Fields: graphql.Fields{
			"list": &graphql.Field{
				Description: "Returns comments list",
				Args:        util.BindArguments(commentsPagination{}),
				Resolve:     CommentsQueryListResolver,
				Type:        util.BindOutput(comments.CommentFilterResponse{}, "Comments list", nil),
			},
		},
	}))

	(*rootQueries)["comments"] = &graphql.Field{
		Type: q,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return q, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "CommentsMutation",
		Description: "Comments mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Creates comment",
				Args:        util.BindArguments(comments.CommentCreateRequest{}),
				Resolve:     CommentsMutationCreateResolver,
				Type:        util.BindOutput(comments.CommentCreateResponse{}, "Created comment", nil),
			},
			"update": &graphql.Field{
				Description: "Updates comment",
				Args:        util.BindArguments(comments.CommentUpdateRequest{}),
				Resolve:     CommentsMutationUpdateResolver,
				Type:        util.BindOutput(comments.CommentUpdateResponse{}, "Updated comment", nil),
			},
			"delete": &graphql.Field{
				Description: "Deletes comment",
				Args:        util.BindArguments(comments.CommentDeleteRequest{}),
				Resolve:     CommentsMutationDeleteResolver,
				Type:        graphql.Boolean,
			},
		},
	}))

	(*rootMutations)["comments"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func CommentsQueryListResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &commentsPagination{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if len(request.Sort) == 0 {
		request.Sort = append(request.Sort, &query.QueryParamsSort{
			Field: "created",
			Order: 1,
		})
	}

	if request.ParentID == "" {
		request.ParentID = "0"
	}

	request.Filter = append(request.Filter, &query.QueryParamsFilter{
		Fields:   []string{"parent_id"},
		Value:    []string{request.ParentID},
		Operator: query.QueryFilterOperator_EQUALS,
	})

	return connections.CommentsService.CommentFilter(params.Context, &comments.CommentFilterRequest{
		Params: &query.QueryParams{
			Pagination: &query.QueryParamsPagination{
				Page:    request.Page,
				PerPage: request.PerPage,
			},
			Filter: append([]*query.QueryParamsFilter{
				{
					Fields:   []string{"correlation_type"},
					Value:    []string{request.CorrelationType.String()},
					Operator: query.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"correlation_id"},
					Value:    []string{request.CorrelationID},
					Operator: query.QueryFilterOperator_EQUALS,
				},
			}, request.Filter...),
			Sort: request.Sort,
		},
	})
}

func CommentsMutationCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &comments.CommentCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.CommentsService.CommentCreate(params.Context, request)
}

func CommentsMutationUpdateResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &comments.CommentUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.CommentsService.CommentUpdate(params.Context, request)
}

func CommentsMutationDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &comments.CommentDeleteRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	_, err := connections.CommentsService.CommentDelete(params.Context, request)

	return true, err
}
