package resolvers

import (
	"github.com/graphql-go/graphql"

	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	insurance "gitlab.mashplace.com/mashplace/backend/insurance/api"
)

func initInsurance(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ContentsInsuranceQuery",
		Description: "ContentsInsurance query root",
		Fields: graphql.Fields{
			"premiumGet": &graphql.Field{
				Description: "Returns contents insurance premium policy.",
				Args: func() graphql.FieldConfigArgument {
					input := util.BindArguments(&insurance.ContentsInsurancePremium_GetRequest{})
					delete(input, "user")
					delete(input, "session")

					return input
				}(),
				Type:    util.BindOutput(&insurance.ContentsInsurancePremium_Policy{}, "", nil),
				Resolve: resolveContentsInsurancePremiumGet,
			},
			"professionsLookup": &graphql.Field{
				Description: "Returns profession dictionary data.",
				Args:        util.BindArguments(&insurance.ProfessionsLookupRequest{}),
				Type:        util.BindOutput(&insurance.ProfessionsLookupResponse{}, "", nil),
				Resolve:     resolveContentsInsuranceProfessionsLookup,
			},
		},
	}))

	(*rootQueries)["contentsInsurance"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ContentsInsuranceMutation",
		Description: "ContentsInsurance mutation root",
		Fields: graphql.Fields{
			"premiumRegisterUser": &graphql.Field{
				Description: "Attach current logged-in user to the contents insurance premium policy" +
					" if previously it did not contain him",
				Args: func() graphql.FieldConfigArgument {
					input := util.BindArguments(&insurance.ContentsInsurancePremium_RegisterUserRequest{})
					delete(input, "user")

					return input
				}(),
				Type:    util.BindOutput(&insurance.ContentsInsurancePremium_RegisterUserResponse{}, "", nil),
				Resolve: resolveContentsInsurancePremiumRegisterUser,
			},
			"premiumSave": &graphql.Field{
				Description: "Save contents insurance premium policy.",
				Args: func() graphql.FieldConfigArgument {
					input := util.BindArguments(&insurance.ContentsInsurancePremium_SaveRequest{})
					delete(input, "user")
					delete(input, "session")

					input["recaptcha_token"] = &graphql.ArgumentConfig{
						Type: graphql.String,
					}

					return input
				}(),
				Type:    util.BindOutput(&insurance.ContentsInsurancePremium_Policy{}, "", nil),
				Resolve: resolveContentsInsurancePremiumSave,
			},
			"premiumPay": &graphql.Field{
				Description: "Create invoice to pay contents insurance premium policy.",
				Args: func() graphql.FieldConfigArgument {
					input := util.BindArguments(&insurance.ContentsInsurancePremium_PayRequest{})
					delete(input, "user")

					input["recaptcha_token"] = &graphql.ArgumentConfig{
						Type: graphql.String,
					}

					return input
				}(),
				Type:    util.BindOutput(&insurance.ContentsInsurancePremium_PayResponse{}, "", nil),
				Resolve: resolveContentsInsurancePremiumPay,
			},
		},
	}))

	(*rootMutations)["contentsInsurance"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func resolveContentsInsurancePremiumGet(params graphql.ResolveParams) (interface{}, error) {
	input := &insurance.ContentsInsurancePremium_GetRequest{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	if user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse); user != nil {
		input.User = user.GetAccount().GetIdValue()
		input.Session = ""
	} else if session, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser); session != nil {
		input.Session = session.Session
		input.User = ""
	}

	res, err := connections.InsuranceService.PremiumGet(params.Context, input)
	if err != nil {
		return nil, err
	}

	return res.GetPolicy(), nil
}

func resolveContentsInsuranceProfessionsLookup(params graphql.ResolveParams) (interface{}, error) {
	input := &insurance.ProfessionsLookupRequest{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	return connections.InsuranceService.ProfessionsLookup(params.Context, input)
}

func resolveContentsInsurancePremiumSave(params graphql.ResolveParams) (interface{}, error) {
	input := &insurance.ContentsInsurancePremium_SaveRequest{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	/*v, ok := params.Args["recaptcha_token"]
	if !ok {
		return nil, grpcerr.InvalidData
	}

	err := recaptcha.CheckRecaptchaToken(params.Context, v.(string))
	if err != nil {
		return nil, err
	}*/

	if user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse); user != nil {
		input.User = user.GetAccount().GetIdValue()
		input.Session = ""
	} else if session, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser); session != nil {
		input.Session = session.Session
		input.User = ""
	}

	res, err := connections.InsuranceService.PremiumSave(params.Context, input)
	if err != nil {
		return nil, err
	}

	return res.GetPolicy(), nil
}

func resolveContentsInsurancePremiumPay(params graphql.ResolveParams) (interface{}, error) {
	input := &insurance.ContentsInsurancePremium_PayRequest{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	/*v, ok := params.Args["recaptcha_token"]
	if !ok {
		return nil, grpcerr.InvalidData
	}

	err := recaptcha.CheckRecaptchaToken(params.Context, v.(string))
	if err != nil {
		return nil, err
	}*/

	if user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse); user != nil {
		input.User = user.GetAccount().GetIdValue()
		input.Session = ""
	} else if session, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser); session != nil {
		input.Session = session.Session
		input.User = ""
	}

	return connections.InsuranceService.PremiumPay(params.Context, input)
}

func resolveContentsInsurancePremiumRegisterUser(params graphql.ResolveParams) (interface{}, error) {
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	input := &insurance.ContentsInsurancePremium_RegisterUserRequest{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	input.User = user.GetAccount().GetIdValue()

	return connections.InsuranceService.PremiumRegisterUser(params.Context, input)
}
