package resolvers

import (
	"github.com/graphql-go/graphql"
	account "gitlab.mashplace.com/mashplace/backend/account/api"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	document "gitlab.mashplace.com/mashplace/backend/document/api"
	expenses "gitlab.mashplace.com/mashplace/backend/expenses/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	"google.golang.org/protobuf/types/known/timestamppb"
)

type deleteRequest struct {
	ID string `json:"id,omitempty"`
}

func initDocument(rootQueries, rootMutations *graphql.Fields) {
	q := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "DocumentQuery",
		Description: "Document query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Get document by id",
				Args:        util.BindArguments(document.GetRequest{}),
				Resolve:     DocumentQueryGetResolver,
				Type:        util.BindOutput(document.GetResponse{}, "Document with given id", nil),
			},
			"list": &graphql.Field{
				Description: "Get documents by filters",
				Args:        util.BindArguments(document.ListRequest{}),
				Resolve:     DocumentQueryListResolver,
				Type:        util.BindOutput(document.ListResponse{}, "Document list", nil),
			},
		},
	}))

	(*rootQueries)["document"] = &graphql.Field{
		Type: q,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return q, nil
		},
	}

	fileQuery := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "FileQuery",
		Description: "File query root",
		Fields: graphql.Fields{
			"list": &graphql.Field{
				Description: "Get files by filters",
				Args:        util.BindArguments(document.FileAttributeFilterRequest{}),
				Resolve:     DocumentQueryFileAttributeFilterResolver,
				Type:        util.BindOutput(document.FileAttributeFilterResponse{}, "File lists", nil),
			},
		},
	}))

	(*rootQueries)["file"] = &graphql.Field{
		Type: fileQuery,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return fileQuery, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "DocumentMutation",
		Description: "Document mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Create document",
				Args:        util.BindArguments(document.CreateRequest{}),
				Resolve:     DocumentMutationCreateResolver,
				Type:        util.BindOutput(document.CreateResponse{}, "Document create response", nil),
			},
			"update": &graphql.Field{
				Description: "Update document",
				Args:        util.BindArguments(document.UpdateRequest{}),
				Resolve:     DocumentMutationUpdateResolver,
				Type:        util.BindOutput(document.UpdateResponse{}, "Document update response", nil),
			},
			"delete": &graphql.Field{
				Description: "Delete document",
				Args:        util.BindArguments(deleteRequest{}),
				Resolve:     DocumentMutationDeleteResolver,
				Type:        util.BindOutput(document.DeleteResponse{}, "Document delete response", nil),
			},
		},
	}))

	(*rootMutations)["document"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func DocumentQueryGetResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &document.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.DocumentService.Get(params.Context, request)
}

func DocumentQueryFileAttributeFilterResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &document.FileAttributeFilterRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.Params == nil {
		request.Params = &query.QueryParams{}
	}

	if !grpcctx.ContextIsAdminUser(params.Context) {
		request.Params.Filter = append(request.Params.Filter, &query.QueryParamsFilter{
			Fields:   []string{"user"},
			Value:    []string{user.GetAccount().GetIdValue()},
			Operator: query.QueryFilterOperator_EQUALS,
		})
	}

	return connections.DocumentService.FileAttributeFilter(params.Context, request)
}

func DocumentQueryListResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &document.ListRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.GetProperty() != "" {
		resp, err := connections.PropertyService.Get(params.Context, &property.GetRequest{
			Id:   request.GetProperty(),
			User: user.GetAccount().GetIdValue(),
			Test: user.GetAccount().GetTestValue(),
		})

		if err != nil {
			return nil, err
		}

		reqtypes := request.GetTypeCodes()

		for _, r := range request.GetTypes() {
			reqtypes = append(reqtypes, string(document.LegacyDocumentTypes[r]))
		}

		landlord, related := propertyRelation(resp.GetProperty().GetPeople(), user.GetAccount().GetIdValue())

		types := propertyDocumentTypes(landlord, related, reqtypes)

		request.TypeCodes = types
	}

	if !grpcctx.ContextIsAdminUser(params.Context) {
		request.User = user.GetAccount().GetIdValue()
	}

	return connections.DocumentService.List(params.Context, request)
}

func DocumentMutationCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &document.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()

	switch request.GetType() {
	case document.DocumentType_PHOTO_RAW:
		request.TypeCode = document.TypePhoto
	case document.DocumentType_PLAN_RAW:
		request.TypeCode = document.TypePlan
	}

	switch request.GetTypeCode() {
	case document.TypePhotoRaw:
		request.TypeCode = document.TypePhoto
	case document.TypePlanRaw:
		request.TypeCode = document.TypePlan
	}

	if request.GetAmount() > 0 {
		t := request.GetIssuedAt()
		if t == nil {
			t = request.GetExpiration()
		}

		if t == nil {
			t = timestamppb.Now()
		}

		_, _ = connections.ExpensesService.RecordCreate(params.Context, &expenses.RecordCreateRequest{
			UserId: request.GetUser(),
			Type: &expenses.RecordCreateRequest_Expense{
				Expense: &expenses.Expense{
					Expense: &expenses.Expense_Other{
						Other: &expenses.ExpenseOther{},
					},
				},
			},
			Config: &expenses.Config{
				UserId: request.GetUser(),
				Type: &expenses.Config_Property{
					Property: &expenses.ConfigProperty{
						PropertyId: request.GetProperty(),
					},
				},
			},
			Tax:         nil,
			Description: request.GetDescription(),
			Amount:      request.GetAmount(),
			Time:        t,
			PropertyId:  request.GetProductId(),
			ListingId:   "",
		})
	}

	return connections.DocumentService.Create(params.Context, request)
}

func DocumentMutationUpdateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &document.UpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()

	return connections.DocumentService.Update(params.Context, request)
}

func DocumentMutationDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	req := &deleteRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	userID := user.GetAccount().GetIdValue()

	// Delete document from account
	reqAcc := &account.DeleteDocumentRequest{
		Id:        req.ID,
		User:      userID,
		AdminSite: false,
	}
	if resp, err := connections.AccountService.DeleteDocument(
		params.Context,
		reqAcc,
	); (resp != nil) && resp.GetError() {
		return nil, err
	}

	// Delete document from property
	reqProp := &property.DeleteDocumentRequest{
		Id:        req.ID,
		User:      userID,
		AdminSite: false,
	}
	if resp, err := connections.PropertyService.DeleteDocument(
		params.Context,
		reqProp,
	); resp != nil && resp.GetError() {
		return nil, err
	}

	// Delete document from offer
	reqOffer := &offer.DeleteDocumentRequest{
		Id:        req.ID,
		User:      userID,
		AdminSite: false,
	}
	if resp, err := connections.OfferService.DeleteDocument(
		params.Context,
		reqOffer,
	); resp != nil && resp.GetError() {
		return nil, err
	}

	// Delete document itself
	reqDoc := &document.DeleteRequest{
		Id:        req.ID,
		User:      userID,
		AdminSite: false,
	}
	return connections.DocumentService.Delete(params.Context, reqDoc)
}
