package resolvers

import (
	"context"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"

	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic"
	genericQuery "gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	expenses "gitlab.mashplace.com/mashplace/backend/expenses/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	"google.golang.org/protobuf/types/known/timestamppb"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

type PropertyExpenseTaxTrackerSetupGetRequest struct {
	PropertyID string `json:"property_id"`
}

type PropertyExpenseTaxTrackerSetupGetResponse struct {
	Setup                   PropertyExpenseTaxTrackerSetup `json:"setup"`
	ExpenseMortgageSchedule *expenses.Schedule             `json:"expense_mortgage_schedule"`
	IncomeRentSchedule      *expenses.Schedule             `json:"income_rent_schedule"`
	IncomeSalarySchedule    *expenses.Schedule             `json:"income_salary_schedule"`
}

type PropertyExpenseTaxTrackerSetupSaveResponse struct {
	Setup                   PropertyExpenseTaxTrackerSetup `json:"setup"`
	ExpenseMortgageSchedule *expenses.Schedule             `json:"expense_mortgage_schedule"`
	IncomeRentSchedule      *expenses.Schedule             `json:"income_rent_schedule"`
	IncomeSalarySchedule    *expenses.Schedule             `json:"income_salary_schedule"`
}

type PropertyExpenseTaxTrackerSetup struct {
	PropertyID               string        `json:"property_id"`
	MonthlyRent              float64       `json:"monthly_rent,omitempty"`
	MonthlyMortgagePayment   float64       `json:"monthly_mortgage_payment,omitempty"`
	MonthlyMortgageInterest  float64       `json:"monthly_mortgage_interest,omitempty"`
	MonthlyAllowableExpenses float64       `json:"monthly_allowable_expenses,omitempty"`
	AnnualSalary             float64       `json:"annual_salary,omitempty"`
	TenancyPeriodStart       *generic.Date `json:"tenancy_period_start,omitempty"`
	TenancyPeriodEnd         *generic.Date `json:"tenancy_period_end,omitempty"`
	TaxYear                  *generic.Date `json:"tax_year,omitempty"`
}

func initExpenses(rootQueries, rootMutations *graphql.Fields) {
	queryRecords := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesRecordsQuery",
		Description: "Expenses records query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Get record",
				Args:        util.BindArguments(expenses.RecordGetRequest{}),
				Resolve:     ExpensesRecordsGetQueryResolver,
				Type:        util.BindOutput(expenses.RecordGetResponse{}, "Record details", nil),
			},
			"list": &graphql.Field{
				Description: "List records",
				Args:        util.BindArguments(expenses.RecordFilterListRequest{}),
				Resolve:     ExpensesRecordsListQueryResolver,
				Type:        util.BindOutput(expenses.RecordFilterListResponse{}, "Records list", nil),
			},
			"listScheduledRecords": &graphql.Field{
				Description: "List scheduled records including matched records by time and correlation_id",
				Args:        util.BindArguments(expenses.ScheduledRecordListRequest{}),
				Resolve:     ExpensesScheduledRecordsListQueryResolver,
				Type:        util.BindOutput(expenses.ScheduledRecordListResponse{}, "Scheduled records list", nil),
			},
			"range": &graphql.Field{
				Description: "Get records total in range",
				Args:        util.BindArguments(expenses.RecordRangeListRequest{}),
				Resolve:     ExpensesRecordsRangeQueryResolver,
				Type:        util.BindOutput(expenses.RecordRangeListResponse{}, "Records list with total", nil),
			},
			"csv": &graphql.Field{
				Description: "Get CSV report of expenses",
				Args:        util.BindArguments(expenses.RecordFilterListRequest{}),
				Resolve:     ExpensesRecordsCSVQueryResolver,
				Type:        util.BindOutput(expenses.ExpenseReportResponse{}, "CSV of expenses", nil),
			},
		},
	}))

	querySchedules := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesSchedulesQuery",
		Description: "Expenses schedules query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Get schedule",
				Args:        util.BindArguments(expenses.ScheduleGetRequest{}),
				Resolve:     ExpensesSchedulesGetQueryResolver,
				Type:        util.BindOutput(expenses.ScheduleGetResponse{}, "Schedule details", nil),
			},
			"list": &graphql.Field{
				Description: "List schedules",
				Args:        util.BindArguments(expenses.ScheduleFilterListRequest{}),
				Resolve:     ExpensesSchedulesListQueryResolver,
				Type:        util.BindOutput(expenses.ScheduleFilterListResponse{}, "Schedules list", nil),
			},
		},
	}))

	queryConfigs := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesConfigsQuery",
		Description: "Expenses Configs query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Get config",
				Args:        util.BindArguments(expenses.ConfigGetRequest{}),
				Resolve:     ExpensesConfigsGetQueryResolver,
				Type:        util.BindOutput(expenses.ConfigGetResponse{}, "Config details", nil),
			},
			"list": &graphql.Field{
				Description: "List Configs",
				Args:        util.BindArguments(expenses.ConfigFilterListRequest{}),
				Resolve:     ExpensesConfigsListQueryResolver,
				Type:        util.BindOutput(expenses.ConfigFilterListResponse{}, "Configs list", nil),
			},
		},
	}))

	queryAggregate := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesAggregateQuery",
		Description: "Expenses Aggregate query root",
		Fields: graphql.Fields{
			"propertyExpenseTaxTrackerSetupGet": &graphql.Field{
				Description: "Get property expense/tax tracker setup parameters",
				Args:        util.BindArguments(PropertyExpenseTaxTrackerSetupGetRequest{}),
				Resolve:     PropertyExpenseTaxTrackerSetupGetQueryResolver,
				Type: util.BindOutput(
					PropertyExpenseTaxTrackerSetupGetResponse{},
					"Property expense/tax tracker setup get response",
					nil,
				),
			},
		},
	}))

	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesQuery",
		Description: "Expenses query root",
		Fields: graphql.Fields{
			"records": &graphql.Field{
				Type: queryRecords,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return queryRecords, nil
				},
			},
			"schedules": &graphql.Field{
				Type: querySchedules,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return querySchedules, nil
				},
			},
			"configs": &graphql.Field{
				Type: queryConfigs,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return queryConfigs, nil
				},
			},
			"aggregate": &graphql.Field{
				Type: queryAggregate,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return queryAggregate, nil
				},
			},
		},
	}))

	(*rootQueries)["expenses"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutationRecords := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesRecordsMutation",
		Description: "Expenses records mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Create record",
				Resolve:     ExpensesRecordsCreateMutationResolver,
				Args:        util.BindArguments(expenses.RecordCreateRequest{}),
				Type:        util.BindOutput(expenses.RecordCreateResponse{}, "Create record", nil),
			},
			"update": &graphql.Field{
				Description: "Update record",
				Resolve:     ExpensesRecordsUpdateMutationResolver,
				Args:        util.BindArguments(expenses.RecordUpdateRequest{}),
				Type:        util.BindOutput(expenses.RecordUpdateResponse{}, "Update record", nil),
			},
			"delete": &graphql.Field{
				Description: "Delete record",
				Resolve:     ExpensesRecordsDeleteMutationResolver,
				Args:        util.BindArguments(expenses.RecordDeleteRequest{}),
				Type:        graphql.Boolean,
			},
		},
	}))

	mutationSchedules := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesSchedulesMutation",
		Description: "Expenses Schedules mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Create schedule",
				Resolve:     ExpensesSchedulesCreateMutationResolver,
				Args:        util.BindArguments(expenses.ScheduleCreateRequest{}),
				Type:        util.BindOutput(expenses.ScheduleCreateResponse{}, "Create schedule", nil),
			},
			"update": &graphql.Field{
				Description: "Update schedule",
				Resolve:     ExpensesSchedulesUpdateMutationResolver,
				Args:        util.BindArguments(expenses.ScheduleUpdateRequest{}),
				Type:        util.BindOutput(expenses.ScheduleUpdateResponse{}, "Update schedule", nil),
			},
			"delete": &graphql.Field{
				Description: "Delete schedule",
				Resolve:     ExpensesSchedulesDeleteMutationResolver,
				Args:        util.BindArguments(expenses.ScheduleDeleteRequest{}),
				Type:        graphql.Boolean,
			},
		},
	}))

	mutationConfigs := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesConfigsMutation",
		Description: "Expenses configs mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Create config",
				Resolve:     ExpensesConfigsCreateMutationResolver,
				Args:        util.BindArguments(expenses.ConfigCreateRequest{}),
				Type:        util.BindOutput(expenses.ConfigCreateResponse{}, "Create config", nil),
			},
			"update": &graphql.Field{
				Description: "Update config",
				Resolve:     ExpensesConfigsUpdateMutationResolver,
				Args:        util.BindArguments(expenses.ConfigUpdateRequest{}),
				Type:        util.BindOutput(expenses.ConfigUpdateResponse{}, "Update config", nil),
			},
		},
	}))

	mutationAggregate := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesAggregateMutation",
		Description: "Expenses Aggregate mutation root",
		Fields: graphql.Fields{
			"propertyExpenseTaxTrackerSetupSave": &graphql.Field{
				Description: "Save property expense/tax tracker setup parameters",
				Args:        util.BindArguments(PropertyExpenseTaxTrackerSetup{}),
				Resolve:     PropertyExpenseTaxTrackerSetupSaveMutationResolver,
				Type: util.BindOutput(
					PropertyExpenseTaxTrackerSetupSaveResponse{},
					"Property expense/tax tracker setup save response",
					nil,
				),
			},
		},
	}))

	mutationAdvance := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesAdvanceMutation",
		Description: "Expenses Advance mutation root",
		Fields: graphql.Fields{
			"request": &graphql.Field{
				Description: "Request an advance for rent payment",
				Args:        util.BindArguments(expenses.RecordAdvanceRequestRequest{}),
				Resolve:     ExpensesAdvanceRequestMutationResolver,
				Type: util.BindOutput(
					expenses.RecordAdvanceRequestResponse{},
					"Advanced requested records",
					nil,
				),
			},
			"cancel": &graphql.Field{
				Description: "Cancel an advance for rent payment",
				Args:        util.BindArguments(expenses.RecordAdvanceCancelRequest{}),
				Resolve:     ExpensesAdvanceCancelMutationResolver,
				Type: util.BindOutput(
					expenses.RecordAdvanceCancelResponse{},
					"Advanced cancelled records",
					nil,
				),
			},
		},
	}))

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ExpensesMutation",
		Description: "Expenses mutation root",
		Fields: graphql.Fields{
			"advance": &graphql.Field{
				Type: mutationAdvance,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return mutationAdvance, nil
				},
			},
			"records": &graphql.Field{
				Type: mutationRecords,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return mutationRecords, nil
				},
			},
			"schedules": &graphql.Field{
				Type: mutationSchedules,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return mutationSchedules, nil
				},
			},
			"configs": &graphql.Field{
				Type: mutationConfigs,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return mutationConfigs, nil
				},
			},
			"aggregate": &graphql.Field{
				Type: mutationAggregate,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return mutationAggregate, nil
				},
			},
		},
	}))

	(*rootMutations)["expenses"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func ExpensesRecordsGetQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.RecordGetRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.RecordGet(params.Context, req)
}

func ExpensesRecordsListQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.RecordFilterListRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	if req.Params == nil {
		req.Params = &genericQuery.QueryParams{}
	}

	req.GetParams().Filter = append(req.GetParams().Filter, &genericQuery.QueryParamsFilter{
		Fields:   []string{"user_id"},
		Value:    []string{user.GetAccount().GetIdValue()},
		Operator: genericQuery.QueryFilterOperator_EQUALS,
	})

	return connections.ExpensesService.RecordFilterList(params.Context, req)
}

func ExpensesScheduledRecordsListQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.ScheduledRecordListRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.ScheduledRecordList(params.Context, req)
}

func ExpensesRecordsRangeQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.RecordRangeListRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.RecordRangeList(params.Context, req)
}

func ExpensesRecordsCSVQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.RecordFilterListRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	if req.Params == nil {
		req.Params = &genericQuery.QueryParams{}
	}

	req.GetParams().Filter = append(req.GetParams().Filter, &genericQuery.QueryParamsFilter{
		Fields:   []string{"user_id"},
		Value:    []string{user.GetAccount().GetIdValue()},
		Operator: genericQuery.QueryFilterOperator_EQUALS,
	})

	return connections.ExpensesService.GenerateExpenseReport(params.Context, req)
}

func ExpensesRecordsCreateMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.RecordCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.RecordCreate(params.Context, req)
}

func ExpensesRecordsUpdateMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.RecordUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.RecordUpdate(params.Context, req)
}

func ExpensesRecordsDeleteMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.RecordDeleteRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	_, err := connections.ExpensesService.RecordDelete(params.Context, req)

	return true, err
}

func ExpensesSchedulesGetQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.ScheduleGetRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.ScheduleGet(params.Context, req)
}

func ExpensesSchedulesListQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.ScheduleFilterListRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	if req.Params == nil {
		req.Params = &genericQuery.QueryParams{}
	}

	req.GetParams().Filter = append(req.GetParams().Filter, &genericQuery.QueryParamsFilter{
		Fields:   []string{"user_id"},
		Value:    []string{user.GetAccount().GetIdValue()},
		Operator: genericQuery.QueryFilterOperator_EQUALS,
	})

	return connections.ExpensesService.ScheduleFilterList(params.Context, req)
}

func ExpensesSchedulesCreateMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.ScheduleCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.ScheduleCreate(params.Context, req)
}

func ExpensesSchedulesUpdateMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.ScheduleUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.ScheduleUpdate(params.Context, req)
}

func ExpensesSchedulesDeleteMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.ScheduleDeleteRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	_, err := connections.ExpensesService.ScheduleDelete(params.Context, req)

	return true, err
}

func ExpensesConfigsGetQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.ConfigGetRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.ConfigGet(params.Context, req)
}

func ExpensesConfigsListQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.ConfigFilterListRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	if req.Params == nil {
		req.Params = &genericQuery.QueryParams{}
	}

	req.GetParams().Filter = append(req.GetParams().Filter, &genericQuery.QueryParamsFilter{
		Fields:   []string{"user_id"},
		Value:    []string{user.GetAccount().GetIdValue()},
		Operator: genericQuery.QueryFilterOperator_EQUALS,
	})

	return connections.ExpensesService.ConfigFilterList(params.Context, req)
}

func ExpensesConfigsCreateMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.ConfigCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.ConfigCreate(params.Context, req)
}

func ExpensesConfigsUpdateMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.ConfigUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.ConfigUpdate(params.Context, req)
}

func ExpensesAdvanceRequestMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.RecordAdvanceRequestRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.RecordAdvanceRequest(params.Context, req)
}

func ExpensesAdvanceCancelMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &expenses.RecordAdvanceCancelRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.UserId = user.GetAccount().GetIdValue()

	return connections.ExpensesService.RecordAdvanceCancel(params.Context, req)
}

func propertyExpenseTaxTrackSetupGet(
	ctx context.Context,
	userID, propertyID string,
) (*PropertyExpenseTaxTrackerSetupGetResponse, error) {
	scheds, err := connections.ExpensesService.ScheduleFilterList(ctx, &expenses.ScheduleFilterListRequest{
		Params: &genericQuery.QueryParams{
			Pagination: &genericQuery.QueryParamsPagination{
				Page:    0,
				PerPage: 1000,
			},
			Sort: []*genericQuery.QueryParamsSort{
				{
					Field: "created",
					Order: -1,
				},
			},
			Filter: []*genericQuery.QueryParamsFilter{
				{
					Fields:   []string{"correlation_id"},
					Value:    []string{propertyID},
					Operator: genericQuery.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"user_id"},
					Value:    []string{userID},
					Operator: genericQuery.QueryFilterOperator_EQUALS,
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	resp := PropertyExpenseTaxTrackerSetupGetResponse{}

	for _, s := range scheds.GetSchedules() {
		switch {
		case s.GetExpense().GetMortgage() != nil:
			resp.ExpenseMortgageSchedule = s
			resp.Setup.MonthlyMortgagePayment = s.GetAmount()
			resp.Setup.MonthlyMortgageInterest = s.GetExpense().GetMortgage().GetMonthlyInterest()
		case s.GetIncome().GetRent() != nil:
			resp.IncomeRentSchedule = s
			resp.Setup.MonthlyRent = s.GetAmount()
			resp.Setup.TenancyPeriodStart = &generic.Date{
				Year:  int32(s.GetStart().AsTime().Year()),
				Month: int32(s.GetStart().AsTime().Month()),
				Day:   int32(s.GetStart().AsTime().Day()),
			}
			resp.Setup.TenancyPeriodEnd = &generic.Date{
				Year:  int32(s.GetEnd().AsTime().Year()),
				Month: int32(s.GetEnd().AsTime().Month()),
				Day:   int32(s.GetEnd().AsTime().Day()),
			}
		}

		if s.GetConfig().GetProperty() != nil {
			prop := s.GetConfig().GetProperty()
			resp.Setup.PropertyID = prop.GetPropertyId()
			resp.Setup.TaxYear = prop.GetTaxYear()
			resp.Setup.MonthlyAllowableExpenses = prop.GetMonthlyAllowableExpenses().GetValue()
		}
	}

	scheds, err = connections.ExpensesService.ScheduleFilterList(ctx, &expenses.ScheduleFilterListRequest{
		Params: &genericQuery.QueryParams{
			Pagination: &genericQuery.QueryParamsPagination{
				Page:    0,
				PerPage: 1000,
			},
			Sort: []*genericQuery.QueryParamsSort{
				{
					Field: "created",
					Order: -1,
				},
			},
			Filter: []*genericQuery.QueryParamsFilter{
				{
					Fields:   []string{"correlation_id"},
					Value:    []string{userID},
					Operator: genericQuery.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"income_type"},
					Value:    []string{expenses.IncomeType_INCOME_TYPE_SALARY.String()},
					Operator: genericQuery.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"user_id"},
					Value:    []string{userID},
					Operator: genericQuery.QueryFilterOperator_EQUALS,
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	if len(scheds.GetSchedules()) > 0 {
		s := scheds.GetSchedules()[0]

		resp.IncomeSalarySchedule = s
		resp.Setup.AnnualSalary = s.GetAmount() * 12.0
	}

	resp.Setup.PropertyID = propertyID

	return &resp, nil
}

func PropertyExpenseTaxTrackerSetupGetQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &PropertyExpenseTaxTrackerSetupGetRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return propertyExpenseTaxTrackSetupGet(params.Context, user.GetAccount().GetIdValue(), req.PropertyID)
}

func dateToTimestamp(d *generic.Date) *timestamppb.Timestamp {
	return timestamppb.New(time.Date(
		int(d.GetYear()),
		time.Month(d.GetMonth()),
		int(d.GetDay()),
		0,
		0,
		0,
		0,
		time.UTC,
	))
}

func PropertyExpenseTaxTrackerSetupSaveMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok {
		return nil, errs.ErrNoUser
	}

	req := &PropertyExpenseTaxTrackerSetup{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	if req.PropertyID == "" {
		return nil, grpcerr.New(grpcerr.InvalidData, "property_id required")
	}

	setup, err := propertyExpenseTaxTrackSetupGet(params.Context, user.GetAccount().GetIdValue(), req.PropertyID)
	if err != nil {
		return nil, err
	}

	setup.Setup = *req

	propconfig := &expenses.Config{
		UserId: user.GetAccount().GetIdValue(),
		Type: &expenses.Config_Property{
			Property: &expenses.ConfigProperty{
				PropertyId:               req.PropertyID,
				MonthlyAllowableExpenses: wrapperspb.Double(req.MonthlyAllowableExpenses),
				TaxYear:                  req.TaxYear,
			},
		},
	}

	err = propertyExpenseTaxTrackSetupSetupMortgage(
		params.Context,
		setup,
		propconfig,
		req,
		user.GetAccount().GetIdValue(),
	)
	if err != nil {
		return nil, err
	}

	err = propertyExpenseTaxTrackSetupSetupRent(
		params.Context,
		setup,
		propconfig,
		req,
		user.GetAccount().GetIdValue(),
	)
	if err != nil {
		return nil, err
	}

	err = propertyExpenseTaxTrackSetupSetupSalary(
		params.Context,
		setup,
		req,
		user.GetAccount().GetIdValue(),
	)
	if err != nil {
		return nil, err
	}

	return setup, nil
}

func propertyExpenseTaxTrackSetupSetupMortgage(
	ctx context.Context,
	setup *PropertyExpenseTaxTrackerSetupGetResponse,
	propconfig *expenses.Config,
	req *PropertyExpenseTaxTrackerSetup,
	userID string,
) error {
	if req.MonthlyMortgagePayment <= 0 {
		return nil
	}

	if setup.ExpenseMortgageSchedule == nil {
		now := time.Now()

		resp, err := connections.ExpensesService.ScheduleCreate(ctx, &expenses.ScheduleCreateRequest{
			UserId: userID,
			Period: &expenses.SchedulePeriod{
				Months: 1,
			},
			Type: &expenses.ScheduleCreateRequest_Expense{
				Expense: &expenses.Expense{
					Expense: &expenses.Expense_Mortgage{
						Mortgage: &expenses.ExpenseMortgage{
							MonthlyInterest: req.MonthlyMortgageInterest,
						},
					},
				},
			},
			Config:      propconfig,
			Tax:         nil,
			Description: "Monthly mortgage payment",
			Amount:      req.MonthlyMortgagePayment,
			Active:      true,
			Start: timestamppb.New(time.Date(
				now.Year(),
				now.Month(),
				1,
				0,
				0,
				0,
				0,
				time.UTC,
			)),
		})
		if err != nil {
			return err
		}

		setup.ExpenseMortgageSchedule = resp.GetSchedule()
	} else {
		resp, err := connections.ExpensesService.ScheduleUpdate(ctx, &expenses.ScheduleUpdateRequest{
			Id:     setup.ExpenseMortgageSchedule.Id,
			UserId: userID,
			Type: &expenses.ScheduleUpdateRequest_Expense{
				Expense: &expenses.Expense{
					Expense: &expenses.Expense_Mortgage{
						Mortgage: &expenses.ExpenseMortgage{
							MonthlyInterest: req.MonthlyMortgageInterest,
						},
					},
				},
			},
			Config: propconfig,
			Amount: wrapperspb.Double(req.MonthlyMortgagePayment),
			Active: wrapperspb.Bool(true),
		})
		if err != nil {
			return err
		}

		setup.ExpenseMortgageSchedule = resp.GetSchedule()
	}

	return nil
}

func propertyExpenseTaxTrackSetupSetupRent(
	ctx context.Context,
	setup *PropertyExpenseTaxTrackerSetupGetResponse,
	propconfig *expenses.Config,
	req *PropertyExpenseTaxTrackerSetup,
	userID string,
) error {
	if req.MonthlyRent == 0 {
		return grpcerr.New(grpcerr.InvalidData, "monthly_rent required")
	}

	if req.TenancyPeriodStart == nil || req.TenancyPeriodStart.Year <= 0 {
		return grpcerr.New(grpcerr.InvalidData, "tenancy_period_start required")
	}

	if req.TenancyPeriodStart == nil || req.TenancyPeriodEnd.Year <= 0 {
		return grpcerr.New(grpcerr.InvalidData, "tenancy_period_end required")
	}

	if setup.IncomeRentSchedule == nil {
		resp, err := connections.ExpensesService.ScheduleCreate(ctx, &expenses.ScheduleCreateRequest{
			UserId: userID,
			Period: &expenses.SchedulePeriod{
				Months: 1,
			},
			Type: &expenses.ScheduleCreateRequest_Income{
				Income: &expenses.Income{
					Income: &expenses.Income_Rent{
						Rent: &expenses.IncomeRent{},
					},
				},
			},
			Config:      propconfig,
			Tax:         nil,
			Description: "Monthly rent income",
			Amount:      req.MonthlyRent,
			Active:      true,
			Start:       dateToTimestamp(req.TenancyPeriodStart),
			End:         dateToTimestamp(req.TenancyPeriodEnd),
		})
		if err != nil {
			return err
		}

		setup.IncomeRentSchedule = resp.GetSchedule()
	} else {
		resp, err := connections.ExpensesService.ScheduleUpdate(ctx, &expenses.ScheduleUpdateRequest{
			Id:     setup.IncomeRentSchedule.Id,
			UserId: userID,
			Config: propconfig,
			Amount: wrapperspb.Double(req.MonthlyRent),
			Active: wrapperspb.Bool(true),
			Start:  dateToTimestamp(req.TenancyPeriodStart),
			End:    dateToTimestamp(req.TenancyPeriodEnd),
		})
		if err != nil {
			return err
		}

		setup.IncomeRentSchedule = resp.GetSchedule()
	}

	return nil
}

func propertyExpenseTaxTrackSetupSetupSalary(
	ctx context.Context,
	setup *PropertyExpenseTaxTrackerSetupGetResponse,
	req *PropertyExpenseTaxTrackerSetup,
	userID string,
) error {
	if req.AnnualSalary <= 0 {
		return nil
	}

	userconfig := &expenses.Config{
		UserId: userID,
		Type: &expenses.Config_User{
			User: &expenses.ConfigUser{
				UserId: userID,
			},
		},
	}

	if setup.IncomeSalarySchedule == nil {
		now := time.Now()

		resp, err := connections.ExpensesService.ScheduleCreate(ctx, &expenses.ScheduleCreateRequest{
			UserId: userID,
			Period: &expenses.SchedulePeriod{
				Months: 1,
			},
			Type: &expenses.ScheduleCreateRequest_Income{
				Income: &expenses.Income{
					Income: &expenses.Income_Salary{
						Salary: &expenses.IncomeSalary{},
					},
				},
			},
			Config:      userconfig,
			Tax:         nil,
			Description: "Monthly salary",
			Amount:      req.AnnualSalary / 12.0,
			Active:      true,
			Start: timestamppb.New(time.Date(
				now.Year(),
				now.Month(),
				1,
				0,
				0,
				0,
				0,
				time.UTC,
			)),
		})
		if err != nil {
			return err
		}

		setup.IncomeSalarySchedule = resp.GetSchedule()
	} else {
		resp, err := connections.ExpensesService.ScheduleUpdate(ctx, &expenses.ScheduleUpdateRequest{
			Id:     setup.IncomeSalarySchedule.Id,
			UserId: userID,
			Amount: wrapperspb.Double(req.AnnualSalary / 12.0),
			Active: wrapperspb.Bool(true),
		})
		if err != nil {
			return err
		}

		setup.IncomeSalarySchedule = resp.GetSchedule()
	}

	return nil
}
