package resolvers

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	genericUtil "gitlab.mashplace.com/mashplace/backend/common/generic/util"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/legacy"
	invitation "gitlab.mashplace.com/mashplace/backend/invitation/api"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
)

func initInvitation(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "InvitationQuery",
		Description: "Invitation query root",
		Fields: graphql.Fields{
			"incoming": &graphql.Field{
				Description: "Get incoming invitations",
				Resolve:     InvitationIncomingQueryResolver,
				Type:        util.BindOutput(invitation.InvitationListResponse{}, "Invitations list", nil),
			},
			"outgoing": &graphql.Field{
				Description: "Get outgoing invitations",
				Resolve:     InvitationOutgoingQueryResolver,
				Type:        util.BindOutput(invitation.InvitationListResponse{}, "Invitations list", nil),
			},
		},
	}))

	(*rootQueries)["invitation"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "InvitationMutation",
		Description: "Invitation mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Create invitation",
				Args:        util.BindArguments(invitation.InvitationInfo{}),
				Resolve:     InvitationCreateMutationResolver,
				Type:        util.BindOutput(invitation.CreateInvitationResponse{}, "Create status", nil),
			},
			"accept": &graphql.Field{
				Description: "Accept invitation",
				Args:        util.BindArguments(invitation.InvitationActionRequest{}),
				Resolve:     InvitationAcceptMutationResolver,
				Type:        util.BindOutput(invitation.InvitationActionResponse{}, "Accept status", nil),
			},
			"reject": &graphql.Field{
				Description: "Reject invitation",
				Args:        util.BindArguments(invitation.InvitationActionRequest{}),
				Resolve:     InvitationRejectMutationResolver,
				Type:        util.BindOutput(invitation.InvitationActionResponse{}, "Reject status", nil),
			},
			"decline": &graphql.Field{
				Description: "Decline invitation by id, no auth",
				Args:        util.BindArguments(invitation.InvitationActionRequest{}),
				Resolve:     InvitationDeclineByLinkMutationResolver,
				Type:        util.BindOutput(invitation.InvitationActionResponse{}, "Decline status", nil),
			},
			"cancel": &graphql.Field{
				Description: "Cancel invitation by id for inviter",
				Args:        util.BindArguments(invitation.CancelInvitationByIdRequest{}),
				Resolve:     InvitationCancelByIDMutationResolver,
				Type:        util.BindOutput(invitation.InvitationActionResponse{}, "Cancel status", nil),
			},
			"referlandlord": &graphql.Field{
				Description: "Save refer landlord data to google sheet",
				Args:        util.BindArguments(invitation.LandlordReferenceSaveRequest{}),
				Resolve:     InvitationReferLandlordMutationResolver,
				Type:        graphql.Boolean,
			},
			"cancelall": &graphql.Field{
				Description: "Cancel all property invitations by property id and invitation type",
				Args:        util.BindArguments(invitation.CancelInvitationsByPropertyRequest{}),
				Resolve:     InvitationCancelAllMutationResolver,
				Type:        graphql.Boolean,
			},
		},
	}))

	(*rootMutations)["invitation"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func InvitationIncomingQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &invitation.InvitationListIncomingRequest{
		Email: user.GetAccount().GetContactValue().GetEmailPrimaryValue(),
	}

	resp, err := connections.InvitationService.InvitationListIncoming(params.Context, request)
	if err != nil {
		return nil, err
	}

	return &invitation.InvitationListResponse{
		Invitations: legacy.ConvertInvitationsLegacy(resp.GetInvitations()),
	}, nil
}

func InvitationOutgoingQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &invitation.InvitationListOutgoingRequest{
		UserId: user.GetAccount().GetIdValue(),
	}

	resp, err := connections.InvitationService.InvitationListOutgoing(params.Context, request)
	if err != nil {
		return nil, err
	}

	return &invitation.InvitationListResponse{
		Invitations: legacy.ConvertInvitationsLegacy(resp.GetInvitations()),
	}, nil
}

func InvitationCreateMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &invitation.InvitationInfo{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.InviterId = user.GetAccount().GetIdValue()

	message := request.GetMessage()

	if message == "" {
		message = request.GetBody()
	}

	invreq := &listing.VersionInvitationCreateRequest{
		Spec:   listing.VersionSpecPropertyID(user.GetAccount().GetIdValue(), request.GetPropertyId()),
		HostId: request.GetInviterId(),
		Invitee: &listing.InvitationUser{
			UserId: request.GetInviteeId(),
			Email:  request.GetInviteeEmail(),
			Phone:  request.GetInviteePhone(),
			Name: genericUtil.JoinNonblankStrings(
				" ",
				request.GetInviteeFirstName(),
				request.GetInviteeLastName(),
			),
		},
		Message: message,
		SiteUrl: connections.SiteURL,
		Request: nil,
	}

	switch request.GetType() {
	case invitation.InvitationType_INVITATION_TYPE_LANDLORD_TO_MANAGER:
		invreq.Request = &listing.VersionInvitationCreateRequest_ManagerInvitation{
			ManagerInvitation: &listing.InvitationRequest{},
		}
	case invitation.InvitationType_INVITATION_TYPE_RELATED_TO_LANDLORD:
		invreq.Request = &listing.VersionInvitationCreateRequest_LandlordInvitation{
			LandlordInvitation: &listing.InvitationRequest{},
		}
	}

	resp, err := connections.ListingService.VersionInvitationCreate(params.Context, invreq)
	if err != nil {
		return nil, err
	}

	return &invitation.CreateInvitationResponse{
		InvitationId: resp.GetInvitationId(),
		Type:         request.GetType(),
	}, nil
}

func InvitationAcceptMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &invitation.InvitationActionRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.InvitationService.InvitationAccept(params.Context, &invitation.InvitationAcceptRequest{
		Id:     request.GetId(),
		UserId: user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return &invitation.InvitationActionResponse{
		Success: true,
		Error:   0,
		Info:    legacy.ConvertInvitationLegacy(resp.GetInvitation()),
	}, nil
}

func InvitationRejectMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &invitation.InvitationActionRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.InvitationService.InvitationDecline(params.Context, &invitation.InvitationDeclineRequest{
		Id:     request.GetId(),
		UserId: user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return &invitation.InvitationActionResponse{
		Success: true,
		Error:   0,
		Info:    legacy.ConvertInvitationLegacy(resp.GetInvitation()),
	}, nil
}

func InvitationDeclineByLinkMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &invitation.InvitationActionRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.InvitationService.InvitationDecline(params.Context, &invitation.InvitationDeclineRequest{
		Id:     request.GetId(),
		UserId: "",
	})
	if err != nil {
		return nil, err
	}

	return &invitation.InvitationActionResponse{
		Success: true,
		Error:   0,
		Info:    legacy.ConvertInvitationLegacy(resp.GetInvitation()),
	}, nil
}

func InvitationCancelByIDMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &invitation.CancelInvitationByIdRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.InvitationService.InvitationCancel(params.Context, &invitation.InvitationCancelRequest{
		Id:     request.GetId(),
		UserId: user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return &invitation.InvitationActionResponse{
		Success: true,
		Error:   0,
		Info:    legacy.ConvertInvitationLegacy(resp.GetInvitation()),
	}, nil
}

func InvitationReferLandlordMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	request := &invitation.LandlordReferenceSaveRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.Host == nil {
		request.Host = &invitation.User{}
	}

	request.Host.UserId = user.GetAccount().GetIdValue()

	if _, err := connections.InvitationService.LandlordReferenceSave(params.Context, request); err != nil {
		return nil, err
	}

	return true, nil
}

func InvitationCancelAllMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &invitation.CancelInvitationsByPropertyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	_, err := connections.InvitationService.InvitationCancelAll(
		params.Context,
		&invitation.InvitationCancelAllRequest{
			PropertyId: request.GetProperty(),
			UserId:     user.GetAccount().GetIdValue(),
		},
	)
	if err != nil {
		return nil, err
	}

	return true, nil
}
