package resolvers

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
)

func initListing(rootQueries, rootMutations *graphql.Fields) {
	queryVersion := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ListingVersionQuery",
		Description: "Listing version query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Get version",
				Args:        util.BindArguments(listing.VersionGetRequest{}),
				Resolve:     ListingVersionQueryGetResolver,
				Type:        util.BindOutput(listing.VersionGetResponse{}, "", nil),
			},
			"list": &graphql.Field{
				Description: "List versions",
				Args:        util.BindArguments(listing.VersionListRequest{}),
				Resolve:     ListingVersionQueryListResolver,
				Type:        util.BindOutput(listing.VersionListResponse{}, "", nil),
			},
		},
	}))

	queryValuation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ListingValuationQuery",
		Description: "Listing valuation query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Get valuation",
				Args:        util.BindArguments(listing.ValuationGetRequest{}),
				Resolve:     ListingValuationQueryGetResolver,
				Type:        util.BindOutput(listing.ValuationGetResponse{}, "", nil),
			},
		},
	}))

	queryPreface := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ListingPrefaceQuery",
		Description: "Listing preface query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Get preface",
				Args:        util.BindArguments(listing.PrefaceGetRequest{}),
				Resolve:     ListingPrefaceQueryGetResolver,
				Type:        util.BindOutput(listing.PrefaceGetResponse{}, "", nil),
			},
		},
	}))

	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ListingQuery",
		Description: "Listing query root",
		Fields: graphql.Fields{
			"version": &graphql.Field{
				Type: queryVersion,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return queryVersion, nil
				},
			},
			"valuation": &graphql.Field{
				Type: queryValuation,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return queryValuation, nil
				},
			},
			"preface": &graphql.Field{
				Type: queryPreface,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return queryPreface, nil
				},
			},
		},
	}))

	(*rootQueries)["listing"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutationVersionInvitation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ListingVersionInvitationMutation",
		Description: "Listing version invitation mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Create version invitation",
				Args:        util.BindArguments(listing.VersionInvitationCreateRequest{}),
				Resolve:     ListingVersionInvitationMutationCreateResolver,
				Type:        graphql.Boolean,
			},
		},
	}))

	mutationVersionViewing := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ListingVersionViewingMutation",
		Description: "Listing version viewing mutation root",
		Fields: graphql.Fields{
			"availability": &graphql.Field{
				Description: "Create version invitation",
				Args:        util.BindArguments(listing.VersionViewingAvailabilityRequest{}),
				Resolve:     ListingVersionViewingMutationAvailabilityResolver,
				Type:        util.BindOutput(listing.VersionViewingAvailabilityResponse{}, "", nil),
			},
		},
	}))

	mutationVersion := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ListingVersionMutation",
		Description: "Listing version mutation root",
		Fields: graphql.Fields{
			"invitation": &graphql.Field{
				Type: mutationVersionInvitation,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return mutationVersionInvitation, nil
				},
			},
			"viewing": &graphql.Field{
				Type: mutationVersionViewing,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return mutationVersionViewing, nil
				},
			},
			"create": &graphql.Field{
				Description: "Create version",
				Args:        util.BindArguments(listing.VersionCreateRequest{}),
				Resolve:     ListingVersionMutationCreateResolver,
				Type:        util.BindOutput(listing.VersionCreateResponse{}, "", nil),
			},
			"update": &graphql.Field{
				Description: "Update version",
				Args:        util.BindArguments(listing.VersionUpdateRequest{}),
				Resolve:     ListingVersionMutationUpdateResolver,
				Type:        util.BindOutput(listing.VersionUpdateResponse{}, "", nil),
			},
			"validate": &graphql.Field{
				Description: "Validate version",
				Args:        util.BindArguments(listing.VersionValidateRequest{}),
				Resolve:     ListingVersionMutationValidateResolver,
				Type:        util.BindOutput(listing.VersionValidateResponse{}, "", nil),
			},
			"submit": &graphql.Field{
				Description: "Submit version",
				Args:        util.BindArguments(listing.VersionSubmitRequest{}),
				Resolve:     ListingVersionMutationSubmitResolver,
				Type:        util.BindOutput(listing.VersionSubmitResponse{}, "", nil),
			},
			"unsubmit": &graphql.Field{
				Description: "Unsubmit version",
				Args:        util.BindArguments(listing.VersionUnsubmitRequest{}),
				Resolve:     ListingVersionMutationUnsubmitResolver,
				Type:        util.BindOutput(listing.VersionUnsubmitResponse{}, "", nil),
			},
		},
	}))

	mutationPreface := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ListingPrefaceMutation",
		Description: "Listing preface mutation root",
		Fields: graphql.Fields{
			"save": &graphql.Field{
				Description: "Save version",
				Args:        util.BindArguments(listing.PrefaceSaveRequest{}),
				Resolve:     ListingPrefaceMutationSaveResolver,
				Type:        util.BindOutput(listing.PrefaceSaveResponse{}, "", nil),
			},
		},
	}))

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ListingMutation",
		Description: "Listing mutation root",
		Fields: graphql.Fields{
			"version": &graphql.Field{
				Type: mutationVersion,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return mutationVersion, nil
				},
			},
			"preface": &graphql.Field{
				Type: mutationPreface,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return mutationPreface, nil
				},
			},
		},
	}))

	(*rootMutations)["listing"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func ListingVersionQueryGetResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.VersionGetRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.Spec.UserId = user.GetAccount().GetIdValue()

	return connections.ListingService.VersionGet(p.Context, request)
}

func ListingVersionQueryListResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.VersionListRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.Spec.UserId = user.GetAccount().GetIdValue()

	return connections.ListingService.VersionList(p.Context, request)
}

func ListingValuationQueryGetResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.ValuationGetRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()

	return connections.ListingService.ValuationGet(p.Context, request)
}

func ListingPrefaceQueryGetResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.PrefaceGetRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.ClientId = user.Session

	return connections.ListingService.PrefaceGet(p.Context, request)
}

func ListingVersionInvitationMutationCreateResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.VersionInvitationCreateRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.Spec.UserId = user.GetAccount().GetIdValue()

	_, err := connections.ListingService.VersionInvitationCreate(p.Context, request)
	if err != nil {
		return nil, err
	}

	return true, nil
}

func ListingVersionViewingMutationAvailabilityResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.VersionViewingAvailabilityRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.Spec.UserId = user.GetAccount().GetIdValue()

	return connections.ListingService.VersionViewingAvailability(p.Context, request)
}

func ListingVersionMutationCreateResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.VersionCreateRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.Spec.UserId = user.GetAccount().GetIdValue()

	return connections.ListingService.VersionCreate(p.Context, request)
}

func ListingVersionMutationUpdateResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.VersionUpdateRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.Spec.UserId = user.GetAccount().GetIdValue()

	return connections.ListingService.VersionUpdate(p.Context, request)
}

func ListingVersionMutationValidateResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.VersionValidateRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.UserId = user.GetAccount().GetIdValue()

	return connections.ListingService.VersionValidate(p.Context, request)
}

func ListingVersionMutationSubmitResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.VersionSubmitRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.UserId = user.GetAccount().GetIdValue()

	return connections.ListingService.VersionSubmit(p.Context, request)
}

func ListingVersionMutationUnsubmitResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.VersionUnsubmitRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.UserId = user.GetAccount().GetIdValue()

	return connections.ListingService.VersionUnsubmit(p.Context, request)
}

func ListingPrefaceMutationSaveResolver(p graphql.ResolveParams) (interface{}, error) {
	user := p.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &listing.PrefaceSaveRequest{}
	if err := util.UnmarshalArguments(p.Args, request); err != nil {
		return nil, err
	}

	request.ClientId = user.Session

	return connections.ListingService.PrefaceSave(p.Context, request)
}
