package resolvers

import (
	"context"
	"encoding/json"
	"fmt"
	"reflect"
	"regexp"
	"sort"
	"strings"

	"github.com/eientei/wsgraphql/mutcontext"
	"github.com/graphql-go/graphql"
	account "gitlab.mashplace.com/mashplace/backend/account/api"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	document "gitlab.mashplace.com/mashplace/backend/document/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	messaging "gitlab.mashplace.com/mashplace/backend/messaging/api"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	viewing "gitlab.mashplace.com/mashplace/backend/viewing/api"
)

type JoinRequest struct {
	messaging.JoinRequest
	Scope messaging.ChannelScope `json:"scope"`
}

type LeaveRequest struct {
	messaging.LeaveRequest
	Scope messaging.ChannelScope `json:"scope"`
}

type HistoryRequest struct {
	messaging.HistoryRequest
	Scope messaging.ChannelScope `json:"scope"`
}

type SendRequest struct {
	messaging.SendRequest
	Scope messaging.ChannelScope `json:"scope"`
}

type EditRequest struct {
	messaging.EditRequest
	Scope messaging.ChannelScope `json:"scope"`
}

type DeleteRequest struct {
	messaging.DeleteRequest
	Scope messaging.ChannelScope `json:"scope"`
}

type SeenRequest struct {
	messaging.SeenRequest
	Scope messaging.ChannelScope `json:"scope"`
}

type MarkAsReadRequest struct {
	messaging.MarkAsReadRequest
	Scope messaging.ChannelScope `json:"scope"`
}

type TypingRequest struct {
	messaging.TypingRequest
	Scope messaging.ChannelScope `json:"scope"`
}

type SDPRequest struct {
	Ref  messaging.Viewing_Reference
	Peer messaging.Viewing_Peer
	SDP  string `json:"sdp"`
}

type ViewingRegisterRequest struct {
	Ref messaging.Viewing_Reference
}

type ViewingStateRequest struct {
	Ref   messaging.Viewing_Reference
	State messaging.Viewing_State
}

type NotificationSpec struct {
	notification.NotificationSpec
	Title string
}

var regexpWhitespace = regexp.MustCompile("\\s+")

//nolint:gocyclo,gocognit // this func can have high cyclomatic and cognitive complexity
func initMessaging(rootQueries, rootMutations, rootSubscriptions *graphql.Fields) {
	util.BindOutput(messaging.EmbedPreviewValue{}, "", util.EnrichmentMap{
		"meta": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(messaging.EmbedPreviewValue_Meta{}, "", nil)
			},
		},
		"audio": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(messaging.EmbedPreviewValue_Audio{}, "", nil)
			},
		},
		"video": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(messaging.EmbedPreviewValue_Video{}, "", nil)
			},
		},
		"image": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(messaging.EmbedPreviewValue_Image{}, "", nil)
			},
		},
	})

	util.BindOutput(messaging.Embed{}, "", util.EnrichmentMap{
		"embed_type": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.DecomposeType(reflect.TypeOf(messaging.EmbedType(0)), false).GetOutput()
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ev, _ := p.Source.(*messaging.Embed)

				return messaging.EmbedType(ev.GetEmbedType()), nil
			},
		},
		"embed_value": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ev, ok := p.Source.(*messaging.Embed)
				if !ok {
					return "", nil
				}

				if messaging.EmbedType(ev.EmbedType) == messaging.EmbedType_EMBED_TYPE_PREVIEW {
					return "", nil
				}

				return ev.GetEmbedValue(), nil
			},
		},
		"account": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(account.Account{}, "", nil)
			},
			OutputNullable: true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ev, ok := p.Source.(*messaging.Embed)
				if ok && ev.EmbedType == int32(messaging.EmbedType_EMBED_TYPE_USER) {
					acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{
						Id: ev.GetEmbedValue(),
					})
					if err == nil {
						return &account.Account{
							Id:        acc.GetAccount().GetId(),
							CreatedAt: acc.GetAccount().GetCreatedAt(),
							Details:   acc.GetAccount().GetDetails(),
						}, nil
					}
				}

				return nil, nil
			},
		},
		"property": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(property.Property{}, "", nil)
			},
			OutputNullable: true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				embed, ok := p.Source.(*messaging.Embed)
				if !ok ||
					messaging.EmbedType(embed.GetEmbedType()) != messaging.EmbedType_EMBED_TYPE_PROPERTY {
					return nil, nil
				}

				var test bool

				user, _ := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if user != nil {
					test = user.GetAccount().GetTestValue()
				}

				prop, err := connections.PropertyService.Get(p.Context, &property.GetRequest{
					Id:   embed.GetEmbedValue(),
					Test: test,
				})

				return prop.GetProperty(), err
			},
		},
		"viewing": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(viewing.Viewing{}, "", nil)
			},
			OutputNullable: true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				embed, ok := p.Source.(*messaging.Embed)
				if !ok ||
					messaging.EmbedType(embed.GetEmbedType()) != messaging.EmbedType_EMBED_TYPE_VIEWING {
					return nil, nil
				}

				user, _ := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if user == nil {
					return nil, nil
				}

				view, err := connections.ViewingService.ViewingGet(
					p.Context,
					&viewing.ViewingGetRequest{
						Spec: &viewing.ViewingSpec{
							UserId: user.GetAccount().GetIdValue(),
							Select: &viewing.ViewingSpecSelect{
								ViewingId: embed.GetEmbedValue(),
							},
						},
					},
				)

				return view.GetViewing(), err
			},
		},
		"offer": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(offer.Offer{}, "", nil)
			},
			OutputNullable: true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ev, ok := p.Source.(*messaging.Embed)
				if ok && ev.EmbedType == int32(messaging.EmbedType_EMBED_TYPE_OFFER) {
					user, _ := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
					if user == nil {
						return nil, nil
					}

					off, err := connections.OfferService.Get(p.Context, &offer.GetRequest{
						Id:   ev.GetEmbedValue(),
						User: user.GetAccount().GetIdValue(),
					})
					if err == nil {
						return off.GetOffer(), nil
					}
				}

				return nil, nil
			},
		},
		"file": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(document.File{}, "", nil)
			},
			OutputNullable: true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ev, ok := p.Source.(*messaging.Embed)
				if ok && ev.EmbedType == int32(messaging.EmbedType_EMBED_TYPE_FILE) {
					file, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{
						Id: ev.GetEmbedValue(),
					})
					if err == nil {
						return file.GetFile(), nil
					}
				}

				return nil, nil
			},
		},
		"preview": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(messaging.EmbedPreviewValue{}, "", nil)
			},
			OutputNullable: true,
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				ev, ok := p.Source.(*messaging.Embed)
				if ok && ev.EmbedType == int32(messaging.EmbedType_EMBED_TYPE_PREVIEW) {
					v := &messaging.EmbedPreviewValue{}

					err := json.Unmarshal([]byte(ev.EmbedValue), v)
					if err != nil {
						return nil, err
					}

					return v, nil
				}

				return nil, nil
			},
		},
	})

	util.BindOutput(messaging.Message{}, "", util.EnrichmentMap{
		"channel": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, ok := p.Source.(*messaging.Message)
				if ok {
					return "resolved:" + ev.GetChannel(), nil
				}

				return "", nil
			},
		},
		"scope": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.DecomposeType(reflect.TypeOf(messaging.ChannelScope(0)), false).GetOutput()
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, ok := p.Source.(*messaging.Message)
				if ok {
					return messaging.ChannelScope(ev.Scope), nil
				}

				return messaging.ChannelScope(0), nil
			},
		},
		"event": &util.Enrichment{
			DeprecationReason: "use notification",
			Output: func(bind util.BindType) graphql.Output {
				return util.DecomposeType(reflect.TypeOf(notification.EventType(0)), false).GetOutput()
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, ok := p.Source.(*messaging.Message)
				if ok {
					_, ok = notification.EventType_name[ev.GetEvent()]
					if !ok {
						return notification.EventType(0), nil
					}

					return notification.EventType(ev.GetEvent()), nil
				}

				return notification.EventType(0), nil
			},
		},
		"notification": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutputName(
					NotificationSpec{},
					"notification details",
					util.EnrichmentMap{
						"title": &util.Enrichment{
							Output: func(bind util.BindType) graphql.Output {
								return graphql.String
							},
							Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
								ev, ok := p.Source.(*NotificationSpec)
								if !ok {
									return "", nil
								}

								return ev.Title, nil
							},
						},
					},
					"notification_api_NotificationSpec",
				)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, ok := p.Source.(*messaging.Message)
				if !ok {
					return nil, nil
				}

				var n *NotificationSpec

				for _, em := range ev.Embeds {
					switch messaging.EmbedType(em.EmbedType) {
					case messaging.EmbedType_EMBED_TYPE_NOTIFICATION:
						if n == nil {
							n = &NotificationSpec{}
						}

						_ = json.Unmarshal([]byte(em.EmbedValue), &n.NotificationSpec)
					case messaging.EmbedType_EMBED_TYPE_TITLE:
						if n == nil {
							n = &NotificationSpec{}
						}

						n.Title = em.EmbedValue
					}
				}

				return n, nil
			},
		},
	})

	util.BindOutput(messaging.Channel{}, "", util.EnrichmentMap{
		"channel": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, ok := p.Source.(*messaging.Channel)
				if ok {
					return "resolved:" + ev.GetChannel(), nil
				}

				return "", nil
			},
		},
		"scope": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.DecomposeType(reflect.TypeOf(messaging.ChannelScope(0)), false).GetOutput()
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, _ := p.Source.(*messaging.Channel)

				return messaging.ChannelScope(ev.GetScope()), nil
			},
		},
	})

	util.BindOutput(messaging.JoinEvent{}, "", util.EnrichmentMap{
		"channel": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, ok := p.Source.(*messaging.JoinEvent)
				if ok {
					return "resolved:" + ev.GetChannel(), nil
				}

				return "", nil
			},
		},
		"scope": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.DecomposeType(reflect.TypeOf(messaging.ChannelScope(0)), false).GetOutput()
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, _ := p.Source.(*messaging.JoinEvent)

				return messaging.ChannelScope(ev.GetScope()), nil
			},
		},
	})

	util.BindOutput(messaging.LeaveEvent{}, "", util.EnrichmentMap{
		"channel": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, ok := p.Source.(*messaging.LeaveEvent)
				if ok {
					return "resolved:" + ev.GetChannel(), nil
				}

				return "", nil
			},
		},
		"scope": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.DecomposeType(reflect.TypeOf(messaging.ChannelScope(0)), false).GetOutput()
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, _ := p.Source.(*messaging.LeaveEvent)

				return messaging.ChannelScope(ev.GetScope()), nil
			},
		},
	})

	util.BindOutput(messaging.SeenEvent{}, "", util.EnrichmentMap{
		"channel": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, ok := p.Source.(*messaging.SeenEvent)
				if ok {
					return "resolved:" + ev.GetChannel(), nil
				}

				return "", nil
			},
		},
		"scope": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.DecomposeType(reflect.TypeOf(messaging.ChannelScope(0)), false).GetOutput()
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, _ := p.Source.(*messaging.SeenEvent)

				return messaging.ChannelScope(ev.GetScope()), nil
			},
		},
	})

	util.BindOutput(messaging.DeleteEvent{}, "", util.EnrichmentMap{
		"channel": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, ok := p.Source.(*messaging.DeleteEvent)
				if ok {
					return "resolved:" + ev.GetChannel(), nil
				}

				return "", nil
			},
		},
		"scope": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.DecomposeType(reflect.TypeOf(messaging.ChannelScope(0)), false).GetOutput()
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, _ := p.Source.(*messaging.DeleteEvent)

				return messaging.ChannelScope(ev.GetScope()), nil
			},
		},
	})

	util.BindOutput(messaging.TypingEvent{}, "", util.EnrichmentMap{
		"channel": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, ok := p.Source.(*messaging.TypingEvent)
				if ok {
					return "resolved:" + ev.GetChannel(), nil
				}

				return "", nil
			},
		},
		"scope": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return util.DecomposeType(reflect.TypeOf(messaging.ChannelScope(0)), false).GetOutput()
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				ev, _ := p.Source.(*messaging.TypingEvent)

				return messaging.ChannelScope(ev.GetScope()), nil
			},
		},
	})

	util.BindOutput(messaging.ViewingEvent{}, "", util.EnrichmentMap{
		"subject": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
					Name: "ViewingEventSubject",
					Fields: graphql.Fields{
						"viewingState": &graphql.Field{
							Description: "Viewing broadcasting state",
							Type:        util.BindOutput(&messaging.Viewing_StateValue{}, "", nil),
							Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
								ev, _ := p.Source.(*messaging.ViewingEvent)

								return ev.GetViewingState().GetValue(), nil
							},
						},
						"sdpOffer": &graphql.Field{
							Description: "SDP offer payload from peer to broadcaster",
							Type:        util.BindOutput(&messaging.Viewing_SDP{}, "", nil),
							Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
								ev, _ := p.Source.(*messaging.ViewingEvent)

								return ev.GetSdpOffer(), nil
							},
						},
						"sdpAnswer": &graphql.Field{
							Description: "SDP answer payload from broadcaster to peer",
							Type:        util.BindOutput(&messaging.Viewing_SDP{}, "", nil),
							Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
								ev, _ := p.Source.(*messaging.ViewingEvent)

								return ev.GetSdpAnswer(), nil
							},
						},
					},
				}))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				return p.Source, nil
			},
		},
	})

	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "MessagingQuery",
		Description: "Messaging query root",
		Fields: graphql.Fields{
			"viewingMetadata": &graphql.Field{
				Description: "Returns available viewing broadcasting metadata",
				Args:        util.BindArguments(messaging.Viewing_Reference{}),
				Type:        util.BindOutput(messaging.Viewing{}, "Viewing broadcasting metadata", nil),
				Resolve:     MessagingViewingMetadataResolver,
			},
		},
	}))

	(*rootQueries)["messaging"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	common := util.EnrichmentMap{
		"errors": &util.Enrichment{
			OutputNullable:    true,
			DeprecationReason: "Do not use",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(
					graphql.NewNonNull(util.DecomposedTypes[reflect.TypeOf(messaging.MessagingError(0))].GetOutput()),
				)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return nil, nil
			},
		},
		"success": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "Do not use",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.Boolean
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				return true, nil
			},
		},
	}

	util.BindOutputName(messaging.DeleteResponse{}, "", common, "messaging_api_DeleteResponseOutput")
	util.BindOutputName(messaging.EditResponse{}, "", common, "messaging_api_EditResponseOutput")
	util.BindOutputName(messaging.HistoryResponse{}, "", common, "messaging_api_HistoryResponseOutput")
	util.BindOutputName(messaging.JoinResponse{}, "", common, "messaging_api_JoinResponseOutput")
	util.BindOutputName(messaging.LeaveResponse{}, "", common, "messaging_api_LeaveResponseOutput")
	util.BindOutputName(messaging.MarkAsReadResponse{}, "", common, "messaging_api_MarkAsReadResponseOutput")
	util.BindOutputName(messaging.PongResponse{}, "", common, "messaging_api_PongResponseOutput")
	util.BindOutputName(messaging.PresenceResponse{}, "", common, "messaging_api_PresenceResponseOutput")
	util.BindOutputName(messaging.SeenResponse{}, "", common, "messaging_api_SeenResponseOutput")
	util.BindOutputName(messaging.SendResponse{}, "", common, "messaging_api_SendResponseOutput")
	util.BindOutputName(messaging.TypingResponse{}, "", common, "messaging_api_TypingResponseOutput")
	util.BindOutputName(
		messaging.UnseenNotificationsResponse{},
		"",
		common,
		"messaging_api_UnseenNotificationsResponse",
	)

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "MessagingMutation",
		Description: "Messaging mutation root",
		Fields: graphql.Fields{
			"join": &graphql.Field{
				Description: "Joins to given channel",
				Args:        util.BindArguments(JoinRequest{}),
				Resolve:     MessagingMutationJoinResolver,
				Type:        util.BindOutput(messaging.JoinResponse{}, "", nil),
			},
			"leave": &graphql.Field{
				Description: "Leaves given channel",
				Args:        util.BindArguments(LeaveRequest{}),
				Resolve:     MessagingMutationLeaveResolver,
				Type:        util.BindOutput(messaging.LeaveResponse{}, "", nil),
			},
			"history": &graphql.Field{
				Description: "Asks for history of a given channel",
				Args:        util.BindArguments(HistoryRequest{}),
				Resolve:     MessagingMutationHistoryResolver,
				Type:        util.BindOutput(messaging.HistoryResponse{}, "", nil),
			},
			"send": &graphql.Field{
				Description: "Sends message to a given channel",
				Args:        util.BindArguments(SendRequest{}),
				Resolve:     MessagingMutationSendResolver,
				Type:        util.BindOutput(messaging.SendResponse{}, "", nil),
			},
			"edit": &graphql.Field{
				Description: "Edits message in a given channel",
				Args:        util.BindArguments(EditRequest{}),
				Resolve:     MessagingMutationEditResolver,
				Type:        util.BindOutput(messaging.EditResponse{}, "", nil),
			},
			"delete": &graphql.Field{
				Description: "Deletes message in a given channel",
				Args:        util.BindArguments(DeleteRequest{}),
				Resolve:     MessagingMutationDeleteResolver,
				Type:        util.BindOutput(messaging.DeleteResponse{}, "", nil),
			},
			"seen": &graphql.Field{
				Description: "Marks message as seen in a given channel",
				Args:        util.BindArguments(SeenRequest{}),
				Resolve:     MessagingMutationSeenResolver,
				Type:        util.BindOutput(messaging.SeenResponse{}, "", nil),
			},
			"pong": &graphql.Field{
				Description: "Replies to ping",
				Args:        util.BindArguments(messaging.PongRequest{}),
				Resolve:     MessagingMutationPongResolver,
				Type:        util.BindOutput(messaging.PongResponse{}, "", nil),
			},
			"markasread": &graphql.Field{
				Description: "Marks all messages in channel as seen",
				Args:        util.BindArguments(MarkAsReadRequest{}),
				Resolve:     MessagingMutationMarkAsReadResolver,
				Type:        util.BindOutput(messaging.MarkAsReadResponse{}, "", nil),
			},
			"presence": &graphql.Field{
				Description: "Notifies about user presence",
				Args:        util.BindArguments(messaging.PresenceRequest{}),
				Resolve:     MessagingMutationPresenceResolver,
				Type:        util.BindOutput(messaging.PresenceResponse{}, "", nil),
			},
			"typing": &graphql.Field{
				Description: "Notifies that the user is typing",
				Args:        util.BindArguments(TypingRequest{}),
				Resolve:     MessagingMutationTypingResolver,
				Type:        util.BindOutput(messaging.TypingResponse{}, "", nil),
			},
			"unseen": &graphql.Field{
				Description: "Get unseen notifications",
				Args:        util.BindArguments(messaging.UnseenNotificationsRequest{}),
				Resolve:     MessagingMutationUnseenResolver,
				Type:        util.BindOutput(messaging.UnseenNotificationsResponse{}, "", nil),
			},
			"viewingRegister": &graphql.Field{
				Description: "Register a viewing and notify members with 'viewingState' subject",
				Args:        util.BindArguments(ViewingRegisterRequest{}),
				Resolve:     MessagingViewingRegisterResolver,
				Type:        graphql.Boolean,
			},
			"viewingOffer": &graphql.Field{
				Description: "Save an SDP offer from peer and notify broadcaster with 'sdpOffer' subject",
				Args:        util.BindArguments(SDPRequest{}),
				Resolve:     MessagingViewingOfferResolver,
				Type:        graphql.Boolean,
			},
			"viewingAnswer": &graphql.Field{
				Description: "Save an SDP answer from broadcaster and notify peer with 'sdpAnswer' subject",
				Args:        util.BindArguments(SDPRequest{}),
				Resolve:     MessagingViewingAnswerResolver,
				Type:        graphql.Boolean,
			},
			"viewingState": &graphql.Field{
				Description: "Set broadcasting state and notify members with 'viewingState' subject",
				Args:        util.BindArguments(ViewingStateRequest{}),
				Resolve:     MessagingSetViewingStateResolver,
				Type:        graphql.Boolean,
			},
		},
	}))

	(*rootMutations)["messaging"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}

	subscription := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "MessagingSubscription",
		Description: "Messaging subscription root",
		Fields: graphql.Fields{
			"connect": &graphql.Field{
				Description: "Connects to messaging interface",
				Args:        util.BindArguments(messaging.ConnectRequest{}),
				Resolve:     MessagingSubscriptionConnectResolver,
				Type: util.BindOutput(messaging.Event{}, "Messaging events", util.EnrichmentMap{
					"event": &util.Enrichment{
						Output: func(bind util.BindType) graphql.Output {
							return graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
								Name: "MessagingEvent",
								Fields: graphql.Fields{
									"init": &graphql.Field{
										Type: util.BindOutput(messaging.InitEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetInit(), nil
										},
									},
									"ping": &graphql.Field{
										Type: util.BindOutput(messaging.PingEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetPing(), nil
										},
									},
									"info": &graphql.Field{
										Type: graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
											Name: "MessagingEventInfo",
											Fields: graphql.Fields{
												"joined": &graphql.Field{
													Type: util.BindOutput(messaging.Channel{}, "", nil),
													Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
														ev, _ := p.Source.(*messaging.InfoEvent)

														return ev.GetJoined(), nil
													},
												},
												"left": &graphql.Field{
													Type: util.BindOutput(messaging.Channel{}, "", nil),
													Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
														ev, _ := p.Source.(*messaging.InfoEvent)

														return ev.GetLeft(), nil
													},
												},
											},
										})),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetInfo(), nil
										},
									},
									"join": &graphql.Field{
										Type: util.BindOutput(messaging.JoinEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetJoin(), nil
										},
									},
									"leave": &graphql.Field{
										Type: util.BindOutput(messaging.LeaveEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetLeave(), nil
										},
									},
									"seen": &graphql.Field{
										Type: util.BindOutput(messaging.SeenEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetSeen(), nil
										},
									},
									"delete": &graphql.Field{
										Type: util.BindOutput(messaging.DeleteEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetDelete(), nil
										},
									},
									"send": &graphql.Field{
										Type: util.BindOutput(messaging.SendEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetSend(), nil
										},
									},
									"edit": &graphql.Field{
										Type: util.BindOutput(messaging.EditEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetEdit(), nil
										},
									},
									"history": &graphql.Field{
										Type: util.BindOutput(messaging.HistoryEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetHistory(), nil
										},
									},
									"presence": &graphql.Field{
										Type: util.BindOutput(messaging.PresenceEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetPresence(), nil
										},
									},
									"typing": &graphql.Field{
										Type: util.BindOutput(messaging.TypingEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (i interface{}, e error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetTyping(), nil
										},
									},
									"viewing": &graphql.Field{
										Type: util.BindOutput(&messaging.ViewingEvent{}, "", nil),
										Resolve: func(p graphql.ResolveParams) (interface{}, error) {
											ev, _ := p.Source.(*messaging.Event)

											return ev.GetViewing(), nil
										},
									},
								},
							}))
						},
						Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
							return p.Source, nil
						},
					},
				}),
			},
		},
	}))

	(*rootSubscriptions)["messaging"] = &graphql.Field{
		Type: subscription,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return subscription, nil
		},
	}
}

var (
	errNotAllowed     = grpcerr.New(grpcerr.AuthError, "not allowed")
	errUnknownChannel = grpcerr.New(grpcerr.NotFound, "unknown channel")
)

//nolint:gocyclo,gocognit // try to reduce the high cyclomatic and cognitive complexity of this func
func channelID(ctx context.Context, scope int32, channel, user string) (string, error) {
	if scope < 1000 {
		return "", errNotAllowed
	}

	var test bool

	acc, _ := ctx.Value(ws.CtxUser).(*auth.AuthResponse)
	if acc != nil {
		test = acc.GetAccount().GetTestValue()
	}

	const strResolved = "resolved:"

	if strings.Index(channel, strResolved) == 0 {
		channel = channel[len(strResolved):]
		parts := regexpWhitespace.Split(channel, -1)
		owns := 0

		for _, u := range parts {
			if u == user {
				owns++
			}
		}

		if owns > 1 {
			return "", errNotAllowed
		}

		if owns == 0 {
			return "", errNotAllowed
		}

		switch messaging.ChannelScope(scope) {
		case messaging.ChannelScope_USER_SCOPE:
			if len(parts) != 2 {
				return "", errUnknownChannel
			}
		case messaging.ChannelScope_USER_PROPERTY_SCOPE:
			if len(parts) != 3 {
				return "", errUnknownChannel
			}
		}

		return channel, nil
	}

	switch messaging.ChannelScope(scope) {
	case messaging.ChannelScope_PROPERTY_SCOPE:
		prop, err := connections.PropertyService.Get(ctx, &property.GetRequest{
			Id:   channel,
			Test: test,
		})
		if err != nil || prop == nil {
			return "", errUnknownChannel
		}
	case messaging.ChannelScope_USER_SCOPE:
		if channel == user {
			return "", errNotAllowed
		}

		other, err := connections.AccountService.Get(ctx, &account.GetRequest{
			Id: channel,
		})

		if err != nil || other == nil {
			return "", errUnknownChannel
		}

		pair := []string{
			user,
			other.GetAccount().GetIdValue(),
		}

		sort.Strings(pair)

		channel = strings.Join(pair, " ")
	case messaging.ChannelScope_USER_PROPERTY_SCOPE:
		parts := regexpWhitespace.Split(channel, -1)
		if len(parts) != 3 {
			return "", errUnknownChannel
		}

		owns := 0

		for _, u := range parts {
			if u == user {
				owns++
			}
		}

		if owns > 1 {
			return "", errNotAllowed
		}

		if owns == 0 {
			return "", errNotAllowed
		}

		prop, err := connections.PropertyService.Get(ctx, &property.GetRequest{
			Id:   parts[0],
			Test: test,
		})

		if err != nil || prop == nil {
			return "", errUnknownChannel
		}

		if parts[1] == user {
			return "", errNotAllowed
		}

		other, err := connections.AccountService.Get(ctx, &account.GetRequest{
			Id: parts[1],
		})
		if err != nil || other == nil {
			return "", errUnknownChannel
		}

		pair := []string{
			user,
			other.GetAccount().GetIdValue(),
		}

		sort.Strings(pair)

		channel = strings.Join(append([]string{prop.GetProperty().GetId()}, pair...), " ")
	}

	return channel, nil
}

func MessagingMutationJoinResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &JoinRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.JoinRequest.Scope = int32(request.Scope)

	if request.JoinRequest.Scope < 1000 {
		request.JoinRequest.Channel = user.GetAccount().GetIdValue()
	} else {
		ch, err := channelID(params.Context, request.GetScope(), request.GetChannel(), user.GetAccount().GetIdValue())
		if err != nil {
			return nil, err
		}

		request.JoinRequest.Channel = ch
	}

	return connections.MessagingService.Join(params.Context, &request.JoinRequest)
}

func MessagingMutationLeaveResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &LeaveRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.LeaveRequest.Scope = int32(request.Scope)

	if request.LeaveRequest.Scope < 1000 {
		request.LeaveRequest.Channel = user.GetAccount().GetIdValue()
	} else {
		ch, err := channelID(params.Context, request.GetScope(), request.GetChannel(), user.GetAccount().GetIdValue())
		if err != nil {
			return nil, err
		}

		request.LeaveRequest.Channel = ch
	}

	return connections.MessagingService.Leave(params.Context, &request.LeaveRequest)
}

func MessagingMutationHistoryResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &HistoryRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.HistoryRequest.Scope = int32(request.Scope)
	if request.HistoryRequest.Scope < 1000 {
		request.HistoryRequest.Channel = user.GetAccount().GetIdValue()
	} else {
		ch, err := channelID(params.Context, request.GetScope(), request.GetChannel(), user.GetAccount().GetIdValue())
		if err != nil {
			return nil, err
		}

		request.HistoryRequest.Channel = ch
	}

	return connections.MessagingService.History(params.Context, &request.HistoryRequest)
}

func MessagingMutationSendResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &SendRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.SendRequest.Scope = int32(request.Scope)

	if request.SendRequest.Scope < 1000 {
		request.SendRequest.Channel = user.GetAccount().GetIdValue()
	} else {
		ch, err := channelID(params.Context, request.GetScope(), request.GetChannel(), user.GetAccount().GetIdValue())
		if err != nil {
			return nil, err
		}

		request.SendRequest.Channel = ch
	}

	return connections.MessagingService.Send(params.Context, &request.SendRequest)
}

func MessagingMutationEditResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &EditRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.EditRequest.Scope = int32(request.Scope)

	if request.EditRequest.Scope < 1000 {
		request.EditRequest.Channel = user.GetAccount().GetIdValue()
	} else {
		ch, err := channelID(params.Context, request.GetScope(), request.GetChannel(), user.GetAccount().GetIdValue())
		if err != nil {
			return nil, err
		}

		request.EditRequest.Channel = ch
	}

	return connections.MessagingService.Edit(params.Context, &request.EditRequest)
}

func MessagingMutationDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &DeleteRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.DeleteRequest.Scope = int32(request.Scope)

	if request.DeleteRequest.Scope < 1000 {
		request.DeleteRequest.Channel = user.GetAccount().GetIdValue()
	} else {
		ch, err := channelID(params.Context, request.GetScope(), request.GetChannel(), user.GetAccount().GetIdValue())
		if err != nil {
			return nil, err
		}

		request.DeleteRequest.Channel = ch
	}

	return connections.MessagingService.Delete(params.Context, &request.DeleteRequest)
}

func MessagingMutationSeenResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &SeenRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.SeenRequest.Scope = int32(request.Scope)

	if request.SeenRequest.Scope < 1000 {
		request.SeenRequest.Channel = user.GetAccount().GetIdValue()
	} else {
		ch, err := channelID(params.Context, request.GetScope(), request.GetChannel(), user.GetAccount().GetIdValue())
		if err != nil {
			return nil, err
		}

		request.SeenRequest.Channel = ch
	}

	return connections.MessagingService.Seen(params.Context, &request.SeenRequest)
}

func MessagingMutationMarkAsReadResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &MarkAsReadRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.MarkAsReadRequest.Scope = int32(request.Scope)

	if request.MarkAsReadRequest.Scope < 1000 {
		request.MarkAsReadRequest.Channel = user.GetAccount().GetIdValue()
	} else {
		ch, err := channelID(params.Context, request.GetScope(), request.GetChannel(), user.GetAccount().GetIdValue())
		if err != nil {
			return nil, err
		}

		request.MarkAsReadRequest.Channel = ch
	}

	return connections.MessagingService.MarkAsRead(params.Context, &request.MarkAsReadRequest)
}

func MessagingMutationPresenceResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &messaging.PresenceRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()

	return connections.MessagingService.Presence(params.Context, request)
}

func MessagingMutationTypingResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &TypingRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.TypingRequest.Scope = int32(request.Scope)

	if request.TypingRequest.Scope < 1000 {
		request.TypingRequest.Channel = user.GetAccount().GetIdValue()
	} else {
		ch, err := channelID(params.Context, request.GetScope(), request.GetChannel(), user.GetAccount().GetIdValue())
		if err != nil {
			return nil, err
		}

		request.TypingRequest.Channel = ch
	}

	return connections.MessagingService.Typing(params.Context, &request.TypingRequest)
}

func MessagingMutationPongResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &messaging.PongRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()

	return connections.MessagingService.Pong(params.Context, request)
}

func MessagingSubscriptionConnectResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	ctx := params.Context.(mutcontext.MutableContext)
	subscription := ctx.Value(ws.CtxSubscription)

	if subscription == nil {
		request := &messaging.ConnectRequest{}

		if err := util.UnmarshalArguments(params.Args, request); err != nil {
			return nil, err
		}

		request.User = user.GetAccount().GetIdValue()

		client, err := connections.MessagingService.Connect(params.Context, request)
		if err != nil {
			return nil, err
		}

		subscription = client

		ctx.Set(ws.CtxSubscription, client)

		go func() {
			<-params.Context.Done()

			_ = client.CloseSend()

			ctx.Set(ws.CtxSubscription, nil)
		}()
	}

	client, ok := subscription.(messaging.MessagingService_ConnectClient)
	if !ok {
		_ = ctx.Cancel()
		return nil, nil
	}

	return client.Recv()
}

func MessagingMutationUnseenResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &messaging.UnseenNotificationsRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()

	if request.GetScope() < 1000 {
		request.Channel = user.GetAccount().GetIdValue()
	} else {
		return nil, nil
	}

	return connections.MessagingService.UnseenNotifications(params.Context, request)
}

func MessagingViewingMetadataResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return (*messaging.Viewing)(nil), errs.ErrNoUser
	}

	ref := &messaging.Viewing_Reference{}

	if err := util.UnmarshalArguments(params.Args, ref); err != nil {
		return (*messaging.Viewing)(nil), err
	}

	data, err := connections.MessagingService.GetViewing(params.Context, &messaging.GetViewingRequest{
		Ref:    ref,
		Sender: user.GetAccount().GetIdValue(),
	})

	return data.GetViewing(), err
}

func MessagingViewingRegisterResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return false, errs.ErrNoUser
	}

	input := &ViewingStateRequest{}

	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return false, err
	}

	prop, err := connections.PropertyService.Get(params.Context, &property.GetRequest{
		Id:   input.Ref.GetProperty(),
		Test: user.GetAccount().GetTestValue(),
	})
	if err != nil {
		return false, err
	}

	if prop.GetProperty() == nil {
		return false, fmt.Errorf("%w: property %s", errs.ErrNotFound, input.Ref.GetProperty())
	}

	view, err := connections.ViewingService.ViewingGet(params.Context, &viewing.ViewingGetRequest{
		Spec: &viewing.ViewingSpec{
			UserId: user.GetAccount().GetIdValue(),
			Select: &viewing.ViewingSpecSelect{
				PropertyId: input.Ref.GetProperty(),
				ViewingId:  input.Ref.GetViewRequest(),
			},
		},
	})
	if err != nil {
		return false, err
	}

	if view.GetViewing() == nil {
		// TODO Parse view.GetErrors()
		return false,
			fmt.Errorf(
				"%w: view request %s not found or access denied",
				errs.ErrServiceError,
				input.Ref.GetViewRequest(),
			)
	}

	known := map[string]bool{}
	people := prop.GetProperty().GetPeople()

	if landlord := people.GetLandlord().GetValue(); landlord != "" {
		known[landlord] = true
	}

	if manager := people.GetManager().GetValue(); manager != "" {
		known[manager] = true
	}

	if manager := view.GetViewing().GetManager(); manager != "" {
		known[manager] = true
	}

	if sender := view.GetViewing().GetSender(); sender != "" {
		known[sender] = true
	}

	members := make([]string, 0, len(known))
	for m := range known {
		members = append(members, m)
	}

	if len(members) == 0 {
		return false, fmt.Errorf("%w: no members found", errs.ErrInternalError)
	}

	resp, err := connections.MessagingService.RegisterViewing(params.Context, &messaging.RegisterViewingRequest{
		Ref:     &input.Ref,
		Sender:  user.Account.GetIdValue(),
		Members: members,
	})

	return resp.GetSuccess(), err
}

func MessagingViewingOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return false, errs.ErrNoUser
	}

	input := &SDPRequest{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return false, err
	}

	resp, err := connections.MessagingService.SetSDPOffer(params.Context, &messaging.SetSDPRequest{
		Ref:    &input.Ref,
		Sender: user.GetAccount().GetIdValue(),
		Sdp: &messaging.Viewing_SDP{
			Peer:    &input.Peer,
			Payload: input.SDP,
		},
	})

	return resp.GetSuccess(), err
}

func MessagingViewingAnswerResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return false, errs.ErrNoUser
	}

	input := &SDPRequest{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return false, err
	}

	resp, err := connections.MessagingService.SetSDPAnswer(params.Context, &messaging.SetSDPRequest{
		Ref:    &input.Ref,
		Sender: user.GetAccount().GetIdValue(),
		Sdp: &messaging.Viewing_SDP{
			Peer:    &input.Peer,
			Payload: input.SDP,
		},
	})

	return resp.GetSuccess(), err
}

func MessagingSetViewingStateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return false, errs.ErrNoUser
	}

	input := &ViewingStateRequest{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return false, err
	}

	resp, err := connections.MessagingService.SetViewingState(params.Context, &messaging.SetViewingStateRequest{
		Ref:    &input.Ref,
		Sender: user.GetAccount().GetIdValue(),
		State:  input.State,
	})

	return resp.GetSuccess(), err
}
