package resolvers

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
)

func initNotification(rootQueries, rootMutations *graphql.Fields) {
	sendgridListsQuery := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "NotificationSendgridListsQuery",
		Description: "Notification sendgrid lists query root",
		Fields: graphql.Fields{
			"list": &graphql.Field{
				Description: "Returns subscription list",
				Resolve:     NotificationSendgridListsFetchQueryResolver,
				Type: util.BindOutput(
					notification.SendgridListSubscriptionFetchResponse{},
					"",
					nil,
				),
			},
		},
	}))

	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "NotificationQuery",
		Description: "Notification query root",
		Fields: graphql.Fields{
			"sendgrid_lists": &graphql.Field{
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return sendgridListsQuery, nil
				},
				Type: sendgridListsQuery,
			},
		},
	}))

	(*rootQueries)["notification"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	subscription := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "NotificationSubscriptionMutation",
		Description: "Notification subscription mutation root",
		Fields: graphql.Fields{
			"leads": &graphql.Field{
				DeprecationReason: "Use sendgrid_lists { subscribe }",
				Description:       "Subscribe to leads list",
				Args:              util.BindArguments(notification.ListSubscribeLeads{}),
				Resolve:           NotificationSubscriptionMutationLeadsResolver,
				Type:              graphql.Boolean,
			},
		},
	}))

	sendgridListsMutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "NotificationSendgridListsMutation",
		Description: "Notification sendgrid lists mutation root",
		Fields: graphql.Fields{
			"subscribe": &graphql.Field{
				Description: "Subscribe to sendgrid mail list",
				Args:        util.BindArguments(notification.SendgridListSubscriptionCreateRequest{}),
				Resolve:     NotificationSendgridListsSubscribeMutationResolver,
				Type: util.BindOutput(
					notification.SendgridListSubscriptionCreateResponse{},
					"",
					nil,
				),
			},
		},
	}))

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "NotificationMutation",
		Description: "Notification mutation root",
		Fields: graphql.Fields{
			"subscribe": &graphql.Field{
				DeprecationReason: "Use sendgrid_lists { subscribe }",
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return subscription, nil
				},
				Type: subscription,
			},
			"sendgrid_lists": &graphql.Field{
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return sendgridListsMutation, nil
				},
				Type: sendgridListsMutation,
			},
		},
	}))

	(*rootMutations)["notification"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func NotificationSubscriptionMutationLeadsResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &notification.ListSubscribeLeads{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user != nil {
		if request.Email == "" {
			request.Email = user.GetAccount().GetContactValue().GetEmailPrimaryValue()
		}

		if request.Phone == "" {
			request.Phone = user.GetAccount().GetContactValue().GetPhonePrimaryValue()
		}
	}

	firstName := user.GetAccount().GetDetailsValue().GetFirstNameValue()
	lastName := user.GetAccount().GetDetailsValue().GetLastNameValue()

	if request.Name != "" {
		parts := regexpWhitespace.Split(request.Name, 2)
		switch len(parts) {
		case 1:
			firstName, lastName = parts[0], ""
		case 2:
			firstName, lastName = parts[0], parts[1]
		}
	}

	_, err := connections.NotificationService.SendgridListSubscriptionCreate(
		params.Context,
		&notification.SendgridListSubscriptionCreateRequest{
			Types: []notification.SendgridListSubscriptionType{
				notification.SendgridListSubscriptionType_SENDGRID_LIST_SUBSCRIPTION_TYPE_LEADS,
			},
			UserId:    user.GetAccount().GetIdValue(),
			Email:     request.GetEmail(),
			Phone:     request.GetPhone(),
			FirstName: firstName,
			LastName:  lastName,
			Address: &notification.SendgridListSubscribeAddress{
				AddressLine1: user.GetAccount().GetLocationValue().GetAddress1Value(),
				AddressLine2: user.GetAccount().GetLocationValue().GetAddress2Value(),
				AddressLine3: user.GetAccount().GetLocationValue().GetAddress3Value(),
				AddressLine4: user.GetAccount().GetLocationValue().GetAddress4Value(),
				City:         user.GetAccount().GetLocationValue().GetCityValue(),
				Country:      user.GetAccount().GetLocationValue().GetCountryValue(),
				Postcode:     user.GetAccount().GetLocationValue().GetPostcodeValue(),
			},
		},
	)
	if err != nil {
		return false, err
	}

	return true, nil
}

func NotificationSendgridListsSubscribeMutationResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &notification.SendgridListSubscriptionCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user != nil {
		request.UserId = user.GetAccount().GetIdValue()

		if request.Email == "" {
			request.Email = user.GetAccount().GetContactValue().GetEmailPrimaryValue()
		}

		if request.Phone == "" {
			request.Phone = user.GetAccount().GetContactValue().GetPhonePrimaryValue()
		}

		if request.FirstName == "" {
			request.FirstName = user.GetAccount().GetDetailsValue().GetFirstNameValue()
		}

		if request.LastName == "" {
			request.LastName = user.GetAccount().GetDetailsValue().GetLastNameValue()
		}

		if request.GetAddress() == nil {
			request.Address = &notification.SendgridListSubscribeAddress{}
		}

		if request.GetAddress().GetAddressLine1() == "" {
			request.Address.AddressLine1 = user.GetAccount().GetLocationValue().GetAddress1Value()
		}

		if request.GetAddress().GetAddressLine2() == "" {
			request.Address.AddressLine2 = user.GetAccount().GetLocationValue().GetAddress2Value()
		}

		if request.GetAddress().GetAddressLine3() == "" {
			request.Address.AddressLine3 = user.GetAccount().GetLocationValue().GetAddress3Value()
		}

		if request.GetAddress().GetAddressLine4() == "" {
			request.Address.AddressLine4 = user.GetAccount().GetLocationValue().GetAddress4Value()
		}

		if request.GetAddress().GetCity() == "" {
			request.Address.City = user.GetAccount().GetLocationValue().GetCityValue()
		}

		if request.GetAddress().GetCountry() == "" {
			request.Address.Country = user.GetAccount().GetLocationValue().GetCountryValue()
		}

		if request.GetAddress().GetPostcode() == "" {
			request.Address.Postcode = user.GetAccount().GetLocationValue().GetPostcodeValue()
		}
	}

	return connections.NotificationService.SendgridListSubscriptionCreate(params.Context, request)
}

func NotificationSendgridListsFetchQueryResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &notification.SendgridListSubscriptionFetchRequest{
		Force: false,
		Emails: []*notification.SendgridListSubscriptionFetchEmail{
			{
				Email:  user.GetAccount().GetContactValue().GetEmailPrimaryValue(),
				UserId: user.GetAccount().GetIdValue(),
			},
		},
	}

	return connections.NotificationService.SendgridListSubscriptionFetch(params.Context, request)
}
