package resolvers

import (
	"bytes"
	"context"
	"fmt"
	"io"
	"net/http"
	"time"

	"github.com/graphql-go/graphql"
	account "gitlab.mashplace.com/mashplace/backend/account/api"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	document "gitlab.mashplace.com/mashplace/backend/document/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	mixpanel "gitlab.mashplace.com/mashplace/backend/mixpanel/api"
	oauth "gitlab.mashplace.com/mashplace/backend/oauth/api"
)

type OauthAuthSignup struct {
	Token oauth.TokenAccessRequest `json:"token"`
	Info  auth.SignupRequest       `json:"info"`
}

type OauthAuthLogin struct {
	Token oauth.TokenAccessRequest `json:"token"`
	Info  auth.OauthRequest        `json:"info"`
}

func initOAuth(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "OauthQuery",
		Description: "Oauth query root",
		Fields: graphql.Fields{
			"facebookUserInfoWithToken": &graphql.Field{
				Description: "Get facebook user info using access token",
				Args:        util.BindArguments(oauth.TokenAccessRequest{}),
				Resolve:     OauthFacebookUserInfoWithTokenResolver,
				Type:        util.BindOutput(oauth.FacebookGetUserInfoResponse{}, "Facebook user info", nil),
			},
		},
	}))

	(*rootQueries)["oauth"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "OauthMutation",
		Description: "Oauth mutation root",
		Fields: graphql.Fields{
			"facebookLogin": &graphql.Field{
				Description: "Login using facebook token",
				Args:        util.BindArguments(OauthAuthLogin{}),
				Resolve:     OauthFacebookLogin,
				Type: util.BindOutput(auth.OauthResponse{}, "Facebook login", util.EnrichmentMap{
					"account": &util.Enrichment{
						Output: func(bind util.BindType) graphql.Output {
							return util.BindOutput(&account.Account{}, "", nil)
						},
						Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
							resp, ok := p.Source.(*auth.OauthResponse)
							if !ok || resp.GetAccessToken() == "" {
								return &account.Account{}, nil
							}
							authResp, err := connections.AuthService.Auth(p.Context, &auth.AuthRequest{
								AccessToken: resp.GetAccessToken(),
								AdminSite:   false,
							})
							if err != nil {
								log.WithError(err).Error()
							}
							return authResp.GetAccount(), err
						},
					},
				}),
			},
			"facebookSignup": &graphql.Field{
				Description: "Signup using facebook token",
				Args:        util.BindArguments(OauthAuthSignup{}),
				Resolve:     OauthFacebookSignup,
				Type:        util.BindOutput(auth.SignupResponse{}, "Facebook signup", nil),
			},
			"googleSignIn": &graphql.Field{
				Description: "Sign in using google authorization code",
				Args:        util.BindArguments(oauth.GoogleSignInRequest{}),
				Resolve:     OauthGoogleSignIn,
				Type:        util.BindOutput(auth.OauthResponse{}, "Google sign in", nil),
			},
		},
	}))

	(*rootMutations)["oauth"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func OauthFacebookUserInfoWithTokenResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &oauth.TokenAccessRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.OauthService.FacebookGetUserInfoWithToken(params.Context, request)
}

func OauthFacebookSignup(params graphql.ResolveParams) (interface{}, error) {
	log.Infof("signup %+v", params.Args)
	request := &OauthAuthSignup{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	log.Infof("signup2 %+v", connections.OauthService)
	resp, err := connections.OauthService.FacebookGetUserInfoWithToken(params.Context, &request.Token)
	log.Infof("signup3 %+v %+v", resp, err)
	if err != nil {
		return nil, err
	}
	remote := params.Context.Value(grpcctx.CtxRemote).(string)
	log.Infof("signup4 %+v", remote)

	if request.Info.Account == nil {
		request.Info.Account = &account.Account{}
	}

	if request.Info.GetAccount().Contact == nil {
		request.Info.Account.Contact = &account.Account_ContactValue{ContactValue: &account.AccountContact{}}
	}

	if request.Info.GetAccount().Details == nil {
		request.Info.Account.Details = &account.Account_DetailsValue{DetailsValue: &account.AccountDetails{}}
	}

	request.Info.GetAccount().Admin = &account.Account_AdminValue{AdminValue: false}

	if len(request.Info.GetAccount().GetContactValue().GetEmailPrimaryValue()) == 0 {
		request.Info.GetAccount().GetContactValue().EmailPrimary =
			&account.AccountContact_EmailPrimaryValue{EmailPrimaryValue: resp.GetEmail()}
	}
	if len(request.Info.GetAccount().GetDetailsValue().GetFirstNameValue()) == 0 {
		request.Info.GetAccount().GetDetailsValue().FirstName =
			&account.AccountDetails_FirstNameValue{FirstNameValue: resp.GetFirstName()}
	}
	if len(request.Info.GetAccount().GetDetailsValue().GetLastNameValue()) == 0 {
		request.Info.GetAccount().GetDetailsValue().LastName =
			&account.AccountDetails_LastNameValue{LastNameValue: resp.GetLastName()}
	}
	request.Info.Remote = &auth.SignupRequest_RemoteValue{RemoteValue: remote}
	request.Info.OauthId = &auth.SignupRequest_OauthIdValue{OauthIdValue: resp.GetId()}
	request.Info.OauthType = &auth.SignupRequest_OauthTypeValue{OauthTypeValue: "facebook"}
	log.Infof("signup5 %+v", &request.Info)

	return connections.AuthService.Signup(params.Context, &request.Info)
}

func OauthFacebookLogin(params graphql.ResolveParams) (r interface{}, err error) {
	request := &OauthAuthLogin{}
	if err = util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	log.Infof("oauth, %+v", connections.OauthService)
	resp, err := connections.OauthService.FacebookGetUserInfoWithToken(params.Context, &request.Token)
	log.Infof("%+v - %+v", resp, err)
	if err != nil {
		return nil, err
	}
	remote := params.Context.Value(grpcctx.CtxRemote).(string)
	log.Info("remote")

	request.Info.Email = &auth.OauthRequest_EmailValue{EmailValue: resp.GetEmail()}
	request.Info.Remote = &auth.OauthRequest_RemoteValue{RemoteValue: remote}
	request.Info.OauthId = &auth.OauthRequest_OauthIdValue{OauthIdValue: resp.GetId()}
	request.Info.OauthType = &auth.OauthRequest_OauthTypeValue{OauthTypeValue: "facebook"}

	log.Infof("%+v", &request.Info)

	return connections.AuthService.OauthLogin(params.Context, &request.Info)
}

func OauthGoogleSignIn(params graphql.ResolveParams) (r interface{}, err error) {
	request := &oauth.GoogleSignInRequest{}
	if err = util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	log.Infof("google signin auth code %v \n", request.GetAuthorizationCode())
	resp, err := connections.OauthService.GoogleSignIn(params.Context, request)
	log.Infof("%+v - %+v", resp, err)
	if err != nil {
		return nil, err
	}

	remote := params.Context.Value(grpcctx.CtxRemote).(string)

	respAcc, err := connections.AccountService.GetByEmail(params.Context, &account.GetByEmailRequest{
		Email: resp.GetUserInfo().GetEmail(),
	})
	if err != nil {
		log.Infof("getByEmail error %+v \n", err)
		if err.Error() == "mongo: no documents in result" {
			return &auth.OauthResponse{
				Errors: []auth.AuthErrors{
					auth.AuthErrors_AUTH_ERROR_ACCOUNT_DOES_NOT_EXIST,
				},
			}, nil
		}

		return nil, err
	}

	logInResp, err := oauthGoogleLogIn(params.Context, resp.UserInfo, remote)
	if err != nil {
		return nil, err
	}

	if logInResp.Success && len(logInResp.GetErrors()) == 0 {
		avatarFile, _ := getGoogleAvatar(params.Context, respAcc.GetAccount().GetIdValue(), resp.GetUserInfo())
		_, err = connections.AccountService.Update(params.Context, &account.UpdateRequest{
			Account: &account.Account{
				Id: &account.Account_IdValue{IdValue: respAcc.GetAccount().GetIdValue()},
				Details: &account.Account_DetailsValue{DetailsValue: &account.AccountDetails{
					Avatar: &account.AccountDetails_AvatarValue{AvatarValue: avatarFile},
				}},
			},
		})
		if err != nil {
			return nil, err
		}

		if respAcc.GetAccount() != nil {
			if respAcc.GetAccount().GetDetails() == nil {
				respAcc.Account.Details = &account.Account_DetailsValue{DetailsValue: &account.AccountDetails{}}
			}

			respAcc.Account.GetDetailsValue().Avatar = &account.AccountDetails_AvatarValue{AvatarValue: avatarFile}
		}

		authLoginSuccess(params.Context, &auth.AuthResponse{
			Account: respAcc.GetAccount(),
		}, logInResp.GetAccessToken(), mixpanel.AuthMethod_GOOGLE_OAUTH)
	}

	return logInResp, err
}

/*func oauthGoogleSignUp(
	ctx context.Context,
	info *oauth.GoogleUserInfo,
	remote string,
) (*auth.SignupResponse, error) {
	request := &auth.SignupRequest{}

	avatarFile, _ := getGoogleAvatar(ctx, info)

	request.Account = &account.Account{
		Details: &account.Account_DetailsValue{DetailsValue: &account.AccountDetails{
			FirstName: &account.AccountDetails_FirstNameValue{FirstNameValue: info.GivenName},
			LastName:  &account.AccountDetails_LastNameValue{LastNameValue: info.FamilyName},
			Avatar:    &account.AccountDetails_AvatarValue{AvatarValue: avatarFile},
		}},
		Contact: &account.Account_ContactValue{ContactValue: &account.AccountContact{
			EmailPrimary: &account.AccountContact_EmailPrimaryValue{EmailPrimaryValue: info.Email},
		}},
		Admin: &account.Account_AdminValue{AdminValue: false},
	}

	request.Remote = &auth.SignupRequest_RemoteValue{RemoteValue: remote}
	request.OauthId = &auth.SignupRequest_OauthIdValue{OauthIdValue: info.GetId()}
	request.OauthType = &auth.SignupRequest_OauthTypeValue{OauthTypeValue: "google"}
	request.TosAccepted = &auth.SignupRequest_TosAcceptedValue{TosAcceptedValue: 2}

	resp, err := connections.AuthService.Signup(ctx, request)

	return resp, err
}*/

func oauthGoogleLogIn(ctx context.Context, info *oauth.GoogleUserInfo, remote string) (*auth.OauthResponse, error) {
	request := &auth.OauthRequest{}

	request.Email = &auth.OauthRequest_EmailValue{EmailValue: info.GetEmail()}
	request.Remote = &auth.OauthRequest_RemoteValue{RemoteValue: remote}
	request.OauthId = &auth.OauthRequest_OauthIdValue{OauthIdValue: info.GetId()}
	request.OauthType = &auth.OauthRequest_OauthTypeValue{OauthTypeValue: "google"}

	return connections.AuthService.OauthLogin(ctx, request)
}

func getGoogleAvatar(ctx context.Context, userid string, info *oauth.GoogleUserInfo) (string, error) {
	resp, err := http.Get(info.PictureUrl)
	if err != nil {
		return "", err
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	body, err := io.ReadAll(resp.Body)
	if err != nil {
		return "", nil
	}

	now := time.Now()
	name := fmt.Sprintf("avatar-%s-%04d-%02d-%02d", info.Email, now.Year(), now.Month(), now.Day())
	mime := resp.Header.Get("Content-type")
	rdr := bytes.NewReader(body)

	fileResp, err := document.UploadFile(
		ctx,
		connections.DocumentService,
		name,
		mime,
		&document.FileAttribute{
			User: userid,
		},
		rdr,
	)

	if err != nil {
		return "", fmt.Errorf("upload failure: %w", err)
	}

	return fileResp.GetId(), nil
}
