package resolvers

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	cardpayment "gitlab.mashplace.com/mashplace/backend/cardpayment/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	identity "gitlab.mashplace.com/mashplace/backend/identity/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	viewing "gitlab.mashplace.com/mashplace/backend/viewing/api"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

type ScheduleDepositParams struct {
	Months         int32   `json:"months"`
	MonthlyPayment float64 `json:"monthly_payment"`
	DueDay         int32   `json:"due_day"`
}

type DepositSelectRequest struct {
	Offer string `json:"offer"`
	User  string `json:"user"`
	Email string `json:"email"`
}

type DepositSelectResponse struct {
	Success         bool                  `json:"success"`
	Invoice         string                `json:"invoice"`
	Mandate         string                `json:"mandate"`
	ScheduleDeposit ScheduleDepositParams `json:"schedule_deposit"`
	Error           string                `json:"error"`
}

type DirectDebitRequest struct {
	Offer string `json:"offer"`
	User  string `json:"user"`
	Email string `json:"email"`
}

type DirectDebitResponse struct {
	Success   bool   `json:"success"`
	SessionID string `json:"session_id"`
	Error     string `json:"error"`
}

func initOffer(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "OfferQuery",
		Description: "Offer query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Retrieves specified offer",
				Args:        util.BindArguments(offer.GetRequest{}),
				Resolve:     OfferQueryGetResolver,
				Type:        util.BindOutput(offer.GetResponse{}, "Requested offer", nil),
			},
			"listbyproperty": &graphql.Field{
				Description: "Retrieves all offers by property",
				Args:        util.BindArguments(offer.GetRequest{}),
				Resolve:     OfferQueryListByPropertyResolver,
				Type:        util.BindOutput(offer.ListResponse{}, "Offer list", nil),
			},
			"myout": &graphql.Field{
				Description: "Retrieves current user outgoing offers",
				Args:        util.BindArguments(offer.ListRequest{}),
				Resolve:     OfferQueryMyoutResolver,
				Type:        util.BindOutput(offer.ListResponse{}, "Offer list", nil),
			},
			"myin": &graphql.Field{
				Description: "Retrieves current user incoming offers",
				Args:        util.BindArguments(offer.ListRequest{}),
				Resolve:     OfferQueryMyinResolver,
				Type:        util.BindOutput(offer.ListResponse{}, "Offer list", nil),
			},
			"serviceslist": &graphql.Field{
				Description: "List services",
				Args:        util.BindArguments(offer.ListServicesRequest{}),
				Resolve:     OfferQueryServicesListResolver,
				Type:        util.BindOutput(offer.ListServicesResponse{}, "Services list", nil),
			},
			"paymentschedule": &graphql.Field{
				Description: "Payment schedule for signed offer",
				Args:        util.BindArguments(offer.ScheduledPaymentListRequest{}),
				Resolve:     OfferQueryPaymentScheduleResolver,
				Type:        util.BindOutput(offer.ScheduledPaymentListResponse{}, "Scheduled payments", nil),
			},
			"getmandate": &graphql.Field{
				Description: "Retrieves specified Stripe's mandate",
				Args:        util.BindArguments(cardpayment.GetMandateRequest{}),
				Resolve:     OfferQueryGetMandateResolver,
				Type:        util.BindOutput(cardpayment.GetMandateResponse{}, "Requested Stripe's mandate", nil),
			},
			"depositselect": &graphql.Field{
				Description: "Request Stripe's Direct Debit params",
				Args:        util.BindArguments(DepositSelectRequest{}),
				Resolve:     OfferQueryDepositSelectResolver,
				Type: util.BindOutput(
					DepositSelectResponse{},
					"Return Stripe's Direct Debit params",
					nil),
			},
		},
	}))

	(*rootQueries)["offer"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	inviteRequest := util.BindArguments(offer.InviteRequest{})
	delete(inviteRequest, "site_url")

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "OfferMutation",
		Description: "Offer mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Create new offer",
				Args: func() graphql.FieldConfigArgument {
					input := util.BindArguments(offer.CreateRequest{})

					input["recaptcha_token"] = &graphql.ArgumentConfig{
						Type: graphql.String,
					}

					return input
				}(),
				Resolve: OfferMutationCreateResolver,
				Type:    util.BindOutput(offer.CreateResponse{}, "Creation status", nil),
			},
			"update": &graphql.Field{
				Description: "Updates existing offer",
				Args:        util.BindArguments(offer.UpdateRequest{}),
				Resolve:     OfferMutationUpdateResolver,
				Type:        util.BindOutput(offer.UpdateResponse{}, "Update status", nil),
			},
			"submit": &graphql.Field{
				Description: "Submit offer for review",
				Args:        util.BindArguments(offer.SubmitRequest{}),
				Resolve:     OfferMutationSubmitResolver,
				Type:        util.BindOutput(offer.SubmitResponse{}, "Submit status", nil),
			},
			"invite": &graphql.Field{
				Description: "Invite tenant to join offer",
				Args:        inviteRequest,
				Resolve:     OfferMutationInviteResolver,
				Type:        util.BindOutput(offer.InviteResponse{}, "Invite creation status", nil),
			},
			"removesecondlandlord": &graphql.Field{
				Description: "Remove second landlord from offer",
				Args:        util.BindArguments(offer.RemoveSecondLandlordRequest{}),
				Resolve:     OfferMutationRemovesecondlandlordResolver,
				Type: util.BindOutput(
					offer.RemoveSecondLandlordResponse{},
					"Remove second landlord status",
					nil,
				),
			},
			"accept": &graphql.Field{
				Description: "Accepts incoming offer",
				Args:        util.BindArguments(offer.AcceptRequest{}),
				Resolve:     OfferMutationAcceptResolver,
				Type:        util.BindOutput(offer.AcceptResponse{}, "Accept status", nil),
			},
			"deny": &graphql.Field{
				Description: "Denies incoming offer",
				Args:        util.BindArguments(offer.DenyRequest{}),
				Resolve:     OfferMutationDenyResolver,
				Type:        util.BindOutput(offer.DenyResponse{}, "Deny status", nil),
			},
			"requestchange": &graphql.Field{
				Description: "Request change for incoming offer",
				Args:        util.BindArguments(offer.RequestChangeRequest{}),
				Resolve:     OfferMutationRequestchangeResolver,
				Type:        util.BindOutput(offer.RequestChangeResponse{}, "Request change status", nil),
			},
			"cancel": &graphql.Field{
				Description: "Cancels own outgoing offer",
				Args:        util.BindArguments(offer.CancelRequest{}),
				Resolve:     OfferMutationCancelResolver,
				Type:        util.BindOutput(offer.CancelResponse{}, "Cancel status", nil),
			},
			"checks": &graphql.Field{
				Description: "Start tenant checks",
				Args:        util.BindArguments(offer.CheckRequest{}),
				Resolve:     OfferMutationCheckResolver,
				Type:        util.BindOutput(offer.CheckResponse{}, "Checks status", nil),
			},
			"payment": &graphql.Field{
				Description: "Request payment url",
				Args:        util.BindArguments(offer.PaymentRequest{}),
				Resolve:     OfferMutationPaymentResolver,
				Type:        util.BindOutput(offer.PaymentResponse{}, "Payment response", nil),
			},
			"servicesselect": &graphql.Field{
				Description: "Select services",
				Args:        util.BindArguments(offer.SelectServicesRequest{}),
				Resolve:     OfferMutationServicesSelectResolver,
				Type:        util.BindOutput(offer.SelectServicesResponse{}, "Services select response", nil),
			},
			"delete": &graphql.Field{
				Description: "Delete offer",
				Args:        util.BindArguments(offer.DeleteOfferByIdRequest{}),
				Resolve:     OfferMutationDeleteResolver,
				Type:        util.BindOutput(offer.DeleteOfferByIdResponse{}, "Delete response", nil),
			},
			"restore": &graphql.Field{
				Description: "Restore offer",
				Args:        util.BindArguments(offer.RestoreOfferByIdRequest{}),
				Resolve:     OfferMutationRestoreResolver,
				Type:        util.BindOutput(offer.RestoreOfferByIdResponse{}, "Restore response", nil),
			},
			"requiredguarantor": &graphql.Field{
				Description: "Request to add a new guarantor to the current tenant",
				Args:        util.BindArguments(offer.RequestRequiredGuarantorRequest{}),
				Resolve:     OfferMutationRequiredGuarantorResolver,
				Type: util.BindOutput(
					offer.RequestRequiredGuarantorResponse{},
					"Required guarantor response",
					nil,
				),
			},
			"addguarantors": &graphql.Field{
				Description: "Add new guarantors by tenant",
				Args:        util.BindArguments(offer.AddGuarantorsRequest{}),
				Resolve:     OfferMutationAddGuarantorsResolver,
				Type:        util.BindOutput(offer.AddGuarantorsResponse{}, "Add guarantors response", nil),
			},
			"addcontractclause": &graphql.Field{
				Description: "Add new contract clause by landlord",
				Args:        util.BindArguments(offer.AddContractClauseRequest{}),
				Resolve:     OfferMutationAddContractClauseResolver,
				Type:        util.BindOutput(offer.AddContractClauseResponse{}, "Add contract clause response", nil),
			},
			"updatecontractclauses": &graphql.Field{
				Description: "Update text of a clause or to remove clause by landlord",
				Args:        util.BindArguments(offer.UpdateContractClausesRequest{}),
				Resolve:     OfferMutationUpdateContractClauseResolver,
				Type: util.BindOutput(
					offer.UpdateContractClausesResponse{},
					"Update contract clause response",
					nil,
				),
			},
			"changestateclauses": &graphql.Field{
				Description: "Add new contract clause by landlord",
				Args:        util.BindArguments(offer.ChangeStateContractClausesRequest{}),
				Resolve:     OfferMutationChangeStateContractClauseResolver,
				Type: util.BindOutput(
					offer.ChangeStateContractClausesResponse{},
					"Change state contract clause response",
					nil,
				),
			},
			"progresstosigning": &graphql.Field{
				Description: "Progress to signing",
				Args:        util.BindArguments(offer.ProgressToSigningRequest{}),
				Resolve:     OfferMutationProgressToSigningResolver,
				Type:        util.BindOutput(offer.ProgressToSigningResponse{}, "Progress to signing response", nil),
			},
			"acceptupdate": &graphql.Field{
				Description: "Accept updates",
				Args:        util.BindArguments(offer.AcceptUpdateRequest{}),
				Resolve:     OfferMutationAcceptUpdateResolver,
				Type:        util.BindOutput(offer.AcceptUpdateResponse{}, "Accept update response", nil),
			},
			"renewtenancy": &graphql.Field{
				Description: "Renew tenancy",
				Args:        util.BindArguments(offer.RenewTenancyRequest{}),
				Resolve:     OfferMutationRenewTenancyResolver,
				Type:        util.BindOutput(offer.RenewTenancyResponse{}, "Renew tenancy response", nil),
			},
			"advancerequest": &graphql.Field{
				DeprecationReason: "use expenses to request advances",
				Description:       "Request advance payment",
				Args:              util.BindArguments(offer.ScheduledPaymentAdvanceCreateRequest{}),
				Resolve:           OfferMutationAdvanceRequestResolver,
				Type: util.BindOutputName(
					offer.ScheduledPaymentAdvanceCreateResponse{},
					"Request advance response",
					util.EnrichmentMap{
						"item": &util.Enrichment{
							DeprecationReason: "use items",
							Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
								svc, ok := p.Source.(*offer.ScheduledPaymentAdvanceCreateResponse)
								if !ok || len(svc.GetItems()) == 0 {
									return &offer.ScheduledPayment{}, nil
								}

								return svc.GetItems()[0], nil
							},
						},
					},
					"offer_api_ScheduledPaymentAdvanceCreateResponse",
				),
			},
			"advancecancel": &graphql.Field{
				DeprecationReason: "use expenses to cancel advances",
				Description:       "Cancel advance payment request",
				Args:              util.BindArguments(offer.ScheduledPaymentAdvanceCancelRequest{}),
				Resolve:           OfferMutationAdvanceCancelResolver,
				Type: util.BindOutput(
					offer.ScheduledPaymentAdvanceCancelResponse{},
					"Cancel advance response",
					nil,
				),
			},
			"rentpayment": &graphql.Field{
				DeprecationReason: "Do not use",
				Description:       "Create rent payment",
				Args:              util.BindArguments(offer.ScheduledPaymentPayRequest{}),
				Resolve:           OfferMutationRentPaymentResolver,
				Type: util.BindOutput(
					offer.ScheduledPaymentPayResponse{},
					"Create rent payment response",
					nil,
				),
			},
			"standingorder": &graphql.Field{
				Description: "Request standing order url",
				Args:        util.BindArguments(offer.StandingOrderRequest{}),
				Resolve:     OfferMutationStandingOrderResolver,
				Type:        util.BindOutput(offer.StandingOrderResponse{}, "Standing order response", nil),
			},
			"setupdirectdebit": &graphql.Field{
				Description: "Request first deposit replacement payment and Stripe's Direct Debit mandate",
				Args:        util.BindArguments(DirectDebitRequest{}),
				Resolve:     OfferMutationSetupDirectDebitResolver,
				Type: util.BindOutput(
					DirectDebitResponse{},
					"Create first deposit replacement payment and Stripe's Direct Debit mandate response",
					nil),
			},
			"rentprofilepdf": &graphql.Field{
				Description: "Request the rentprofile's check PDF file",
				Args:        util.BindArguments(identity.DownloadPDFRequest{}),
				Resolve:     OfferMutationDownloadRentprofilePDFResolver,
				Type: util.BindOutput(
					identity.DownloadPDFResponse{},
					"Get downloaded pdf file ID in response",
					nil),
			},
		},
	}))

	(*rootMutations)["offer"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func OfferMutationRenewTenancyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.RenewTenancyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.RenewTenancy(params.Context, request)
}

func OfferMutationAcceptUpdateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.AcceptUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.AcceptUpdate(params.Context, request)
}

func OfferMutationProgressToSigningResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ProgressToSigningRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ProgressToSigning(params.Context, request)
}

func OfferMutationAddContractClauseResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.AddContractClauseRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.AddContractClause(params.Context, request)
}

func OfferMutationUpdateContractClauseResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.UpdateContractClausesRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.UpdateContractClauses(params.Context, request)
}

func OfferMutationChangeStateContractClauseResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ChangeStateContractClausesRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ChangeStateContractClauses(params.Context, request)
}

func OfferMutationRequiredGuarantorResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.RequestRequiredGuarantorRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.RequestRequiredGuarantor(params.Context, request)
}

func OfferMutationAddGuarantorsResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.AddGuarantorsRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.AddGuarantors(params.Context, request)
}

func OfferQueryGetResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Get(params.Context, request)
}

func OfferQueryListByPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.OfferService.ListByProperty(params.Context, request)
}

func OfferQueryMyoutResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ListOutgoing(params.Context, request)
}

func OfferQueryServicesListResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ListServicesRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ServicesList(params.Context, request)
}

func OfferQueryMyinResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ListIncoming(params.Context, request)
}

func OfferQueryPaymentScheduleResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ScheduledPaymentListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ScheduledPaymentList(params.Context, request)
}

func OfferQueryGetMandateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &cardpayment.GetMandateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.CardPaymentService.GetMandate(params.Context, request)
}

func OfferQueryDepositSelectResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &DepositSelectRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.User == "" {
		request.User = user.GetAccount().GetIdValue()
	}

	if request.Email == "" {
		request.Email = user.GetAccount().GetContactValue().GetEmailPrimaryValue()
	}

	resp, err := connections.OfferService.DepositSelect(params.Context, &offer.DepositSelectRequest{
		User:  request.User,
		Offer: request.Offer,
	})
	if err != nil {
		log.Error(err)
		return &DepositSelectResponse{
			Success: false,
		}, err
	}

	mnd, err := connections.CardPaymentService.FilterMandates(params.Context, &cardpayment.FilterMandatesRequest{
		Params: &query.QueryParams{
			Filter: []*query.QueryParamsFilter{
				{
					Fields:   []string{"user_id"},
					Value:    []string{request.User},
					Operator: query.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"email"},
					Value:    []string{request.Email},
					Operator: query.QueryFilterOperator_EQUALS,
				},
			},
		},
	})
	if err != nil {
		log.Error(err)
	}

	var mandate *cardpayment.Mandate
	if mnd != nil && len(mnd.Items) > 0 {
		mandate = mnd.GetItems()[0]
	}

	return &DepositSelectResponse{
		Success: true,
		Invoice: resp.Invoice,
		Mandate: mandate.GetId(),
		ScheduleDeposit: ScheduleDepositParams{
			DueDay:         resp.GetDeposit().GetDueDay(),
			Months:         resp.GetDeposit().GetMonths(),
			MonthlyPayment: resp.GetDeposit().GetMonthlyPayment(),
		},
	}, nil
}

func OfferMutationCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &offer.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	// set default values
	if request.Offer == nil {
		request.Offer = &offer.Offer{}
	}

	if request.Offer.Conditions == nil {
		request.Offer.Conditions = &offer.OfferConditions{}
	}

	if request.Offer.Conditions.PaymentFrequency == nil {
		request.Offer.Conditions.PaymentFrequency = wrapperspb.Int32(1)
	}

	viewlist, _ := connections.ViewingService.ViewingList(params.Context, &viewing.ViewingListRequest{
		Spec: &viewing.ViewingSpec{
			UserId: user.GetAccount().GetIdValue(),
			Select: &viewing.ViewingSpecSelect{
				PropertyId: request.GetOffer().GetProperty(),
				ListingId:  request.GetOffer().GetPropertyListing(),
				States: []viewing.ViewingState{
					viewing.ViewingState_VIEWING_STATE_OFFER,
					viewing.ViewingState_VIEWING_STATE_COMPLETED,
					viewing.ViewingState_VIEWING_STATE_ACCEPTED,
					viewing.ViewingState_VIEWING_STATE_VIEWER_DECISION,
				},
			},
		},
	})

	if len(viewlist.GetViewings()) > 0 {
		request.Offer.Viewing = viewlist.GetViewings()[0].GetId()
	}

	request.User = user.GetAccount().GetIdValue()

	return connections.OfferService.Create(params.Context, request)
}

func OfferMutationUpdateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &offer.UpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Update(params.Context, request)
}

func OfferMutationAcceptResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.AcceptRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Accept(params.Context, request)
}

func OfferMutationDenyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.DenyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Deny(params.Context, request)
}

func OfferMutationCancelResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.CancelRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Cancel(params.Context, request)
}

func OfferMutationSubmitResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.SubmitRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Submit(params.Context, request)
}

func OfferMutationInviteResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.InviteRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	request.SiteUrl = connections.SiteURL
	return connections.OfferService.Invite(params.Context, request)
}

func OfferMutationRemovesecondlandlordResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &offer.RemoveSecondLandlordRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.RemoveSecondLandlord(params.Context, request)
}

func OfferMutationRequestchangeResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.RequestChangeRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.RequestChange(params.Context, request)
}

func OfferMutationCheckResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.CheckRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Check(params.Context, request)
}

func OfferMutationPaymentResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.PaymentRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Payment(params.Context, request)
}

func OfferMutationServicesSelectResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.SelectServicesRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ServicesSelect(params.Context, request)
}

func OfferMutationDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.DeleteOfferByIdRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.DeleteOfferById(params.Context, request)
}

func OfferMutationRestoreResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.RestoreOfferByIdRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.RestoreOfferById(params.Context, request)
}

func OfferMutationAdvanceRequestResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ScheduledPaymentAdvanceCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ScheduledPaymentAdvanceCreate(params.Context, request)
}

func OfferMutationAdvanceCancelResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ScheduledPaymentAdvanceCancelRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ScheduledPaymentAdvanceCancel(params.Context, request)
}

func OfferMutationRentPaymentResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	return &offer.ScheduledPaymentPayResponse{}, nil
}

func OfferMutationStandingOrderResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.StandingOrderRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.StandingOrder(params.Context, request)
}

func OfferMutationSetupDirectDebitResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &DirectDebitRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()

	resp, err := connections.OfferService.DirectDebit(params.Context, &offer.DirectDebitRequest{
		User:  request.User,
		Offer: request.Offer,
	})
	if err != nil {
		return nil, err
	}

	return &DirectDebitResponse{
		Success:   true,
		SessionID: resp.GetStripe().ClientSecret,
	}, nil
}

func OfferMutationDownloadRentprofilePDFResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &identity.DownloadPDFRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.IdentityService.DownloadPDF(params.Context, request)
	return resp, err
}
