package resolvers

import (
	"fmt"

	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
)

type orderSaveRequestProduct struct {
	invoice.Product
	Service    *extras.ServiceSnapshotStatus `json:"service" graphqlbind:"-"`
	Attributes map[string]string             `json:"attributes"`
}

type orderSaveRequest struct {
	CorrelationType   string                     `json:"correlation_type"`
	CorrelationID     string                     `json:"correlation_id"`
	CorrelationTarget string                     `json:"correlation_target"`
	Userid            string                     `json:"userid"`
	Userdata          *invoice.UserData          `json:"userdata"`
	Products          []*orderSaveRequestProduct `json:"products"`
	ClientID          string                     `json:"clientid"`
	OrderType         invoice.OrderType          `json:"order_type"`
	OrderID           string                     `json:"order_id"`
}

func initOrder(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "OrderQuery",
		Description: "Order query root",
		Fields: graphql.Fields{
			"load": &graphql.Field{
				Description: "Load saved cart or specific order by id",
				Args:        util.BindArguments(invoice.OrderLoadRequest{}),
				Resolve:     OrderQueryLoadResolver,
				Type:        util.BindOutput(invoice.OrderLoadResponse{}, "Saved order", nil),
			},
			"list": &graphql.Field{
				Description: "List orders",
				Args:        util.BindArguments(invoice.OrderListRequest{}),
				Resolve:     OrderQueryListResolver,
				Type:        util.BindOutput(invoice.OrderListResponse{}, "Order list", nil),
			},
			"products": &graphql.Field{
				Description: "List all user products",
				Args:        util.BindArguments(invoice.ProductListRequest{}),
				Resolve:     OrderQueryProductsResolver,
				Type:        util.BindOutput(invoice.ProductListResponse{}, "Product list", nil),
			},
		},
	}))

	(*rootQueries)["order"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "OrderMutation",
		Description: "Order mutation root",
		Fields: graphql.Fields{
			"save": &graphql.Field{
				Description: "Save order",
				Args:        util.BindArguments(orderSaveRequest{}),
				Resolve:     OrderMutationSaveResolver,
				Type:        util.BindOutput(invoice.OrderSaveResponse{}, "Save status", nil),
			},
			"commit": &graphql.Field{
				Description: "Commit order and generate invoice",
				Args:        util.BindArguments(invoice.OrderCommitRequest{}),
				Resolve:     OrderMutationCommitResolver,
				Type:        util.BindOutput(invoice.OrderCommitResponse{}, "Commit status", nil),
			},
			"render": &graphql.Field{
				Description: "Render documents for order product",
				Args:        util.BindArguments(invoice.ProductRenderDocumentsRequest{}),
				Resolve:     OrderMutationRenderResolver,
				Type:        util.BindOutput(invoice.ProductRenderDocumentsResponse{}, "Rendered product", nil),
			},
			"payment": &graphql.Field{
				Description: "Create payment for standing order/invoice",
				Args:        util.BindArguments(invoice.PaymentCreateRequest{}),
				Resolve:     OrderMutationPaymentResolver,
				Type:        util.BindOutput(invoice.PaymentCreateResponse{}, "Payment status", nil),
			},
			"cancel": &graphql.Field{
				Description: "Cancel current saved order",
				Args:        util.BindArguments(invoice.OrderCancelRequest{}),
				Resolve:     OrderMutationCancelResolver,
				Type:        graphql.Boolean,
			},
		},
	}))

	(*rootMutations)["order"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func OrderQueryLoadResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)

	request := &invoice.OrderLoadRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	switch {
	case user != nil:
		request.Clientid = ""
		request.Userid = user.GetAccount().GetIdValue()
	case tuser != nil:
		request.Clientid = tuser.Session
		request.Userid = ""
	default:
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	return connections.InvoiceService.OrderLoad(params.Context, request)
}

func OrderQueryListResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &invoice.OrderListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.Userid = user.GetAccount().GetIdValue()

	return connections.InvoiceService.OrderList(params.Context, request)
}

func OrderQuerySavePrefaceResolver(params graphql.ResolveParams) (interface{}, error) {
	tuser, ok := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if !ok {
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	request := &orderSaveRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var products []*invoice.Product

	for _, p := range request.Products {
		products = append(products, &p.Product)
	}

	return connections.InvoiceService.OrderSave(params.Context, &invoice.OrderSaveRequest{
		Userdata:          request.Userdata,
		Products:          products,
		Clientid:          tuser.Session,
		OrderType:         request.OrderType,
		CorrelationType:   request.CorrelationType,
		CorrelationId:     request.CorrelationID,
		CorrelationTarget: request.CorrelationTarget,
	})
}

func OrderQueryProductsResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &invoice.ProductListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()

	return connections.InvoiceService.ProductList(params.Context, request)
}

func OrderMutationSaveResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return OrderQuerySavePrefaceResolver(params)
	}

	request := &orderSaveRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var products []*invoice.Product

	for _, p := range request.Products {
		if p.Product.GetDetails().GetPerformance().GetDesiredPerformanceDate() != nil &&
			p.Product.GetDetails().GetTime().GetStart() == nil {
			p.Product.Details.Time = &invoice.ProductDetailsTime{
				Start: p.Product.GetDetails().GetPerformance().GetDesiredPerformanceDate(),
			}
		}

		products = append(products, &p.Product)
	}

	return connections.InvoiceService.OrderSave(params.Context, &invoice.OrderSaveRequest{
		Userid:            user.GetAccount().GetIdValue(),
		Userdata:          request.Userdata,
		Products:          products,
		OrderType:         request.OrderType,
		CorrelationType:   request.CorrelationType,
		CorrelationId:     request.CorrelationID,
		CorrelationTarget: request.CorrelationTarget,
		OrderId:           request.OrderID,
	})
}

func OrderMutationCommitResolver(params graphql.ResolveParams) (interface{}, error) {
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)

	request := &invoice.OrderCommitRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	switch {
	case user != nil:
		request.Userid = user.GetAccount().GetIdValue()
	case tuser != nil:
		request.Userid = tuser.Session
	default:
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	return connections.InvoiceService.OrderCommit(params.Context, request)
}

func OrderMutationRenderResolver(params graphql.ResolveParams) (interface{}, error) {
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)

	request := &invoice.ProductRenderDocumentsRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	switch {
	case user != nil:
		request.Clientid = ""
		request.Userid = user.GetAccount().GetIdValue()
	case tuser != nil:
		request.Clientid = tuser.Session
		request.Userid = ""
	default:
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	return connections.InvoiceService.ProductRenderDocuments(params.Context, request)
}

func OrderMutationPaymentResolver(params graphql.ResolveParams) (resp interface{}, err error) {
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)

	request := &invoice.PaymentCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	switch {
	case user != nil:
		request.User = user.GetAccount().GetIdValue()
	case tuser != nil:
		request.User = tuser.Session
	default:
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	return connections.InvoiceService.PaymentCreate(params.Context, request)
}

func OrderMutationCancelResolver(params graphql.ResolveParams) (interface{}, error) {
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)

	request := &invoice.OrderCancelRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	switch {
	case user != nil:
		request.Clientid = ""
		request.Userid = user.GetAccount().GetIdValue()
	case tuser != nil:
		request.Clientid = tuser.Session
		request.Userid = ""
	default:
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	_, err := connections.InvoiceService.OrderCancel(params.Context, request)
	if err != nil {
		return nil, err
	}

	return true, nil
}
