package resolvers

import (
	"fmt"

	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
)

type PaymentCreateInput struct {
	ByInvoiceID        string                           `json:"by_invoice_id,omitempty"`
	ByInvoiceReference string                           `json:"by_invoice_reference,omitempty"`
	PaymentSystem      invoice.PaymentSystemDataRequest `json:"payment_system,omitempty"`
}

func initPayment(rootQueries, rootMutations *graphql.Fields) {
	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "PaymentMutation",
		Description: "Payment mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Create payment for existent invoice.",
				Args:        util.BindArguments(PaymentCreateInput{}),
				Resolve:     resolvePaymentCreate,
				Type:        util.BindOutput(invoice.PaymentCreateResponse{}, "", nil),
			},
		},
	}))

	(*rootMutations)["payment"] = &graphql.Field{
		Type: mutation,
		Resolve: func(graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func resolvePaymentCreate(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := PaymentCreateInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &invoice.PaymentCreateRequest{
		User:   user.GetAccount().GetIdValue(),
		System: &input.PaymentSystem,
	}
	switch {
	case input.ByInvoiceID != "":
		forward.InvoiceId = input.ByInvoiceID
	case input.ByInvoiceReference != "":
		forward.InvoiceReference = input.ByInvoiceReference
	default:
		return nil,
			fmt.Errorf("%w: neither `by_invoice_id' nor 'by_invoice_reference' is specified", errs.ErrInvalidGQLInput)
	}

	return connections.InvoiceService.PaymentCreate(params.Context, forward)
}
