package resolvers

import (
	"context"
	"errors"
	"fmt"
	"time"

	account "gitlab.mashplace.com/mashplace/backend/account/api"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
	"gitlab.mashplace.com/mashplace/backend/common/generic/wrap"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/legacy"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	tenancy "gitlab.mashplace.com/mashplace/backend/tenancy/api"
	viewing "gitlab.mashplace.com/mashplace/backend/viewing/api"

	"github.com/graphql-go/graphql"
	"google.golang.org/protobuf/types/known/timestamppb"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

type IntentionRequest struct {
	Tenancy         *wrapperspb.StringValue   `json:"tenancy"`
	Intention       property.TenancyIntention `json:"intention"`
	TerminationDate *timestamppb.Timestamp    `json:"termination_date,omitempty"` // only for intention == TERMINATE
	PenaltyConsent  *wrapperspb.BoolValue     `json:"penalty_consent"`
}

type LandlordIntentionRequest struct {
	Tenancy         *wrapperspb.StringValue   `json:"tenancy"`
	Intention       property.TenancyIntention `json:"intention"`
	TerminationDate *timestamppb.Timestamp    `json:"termination_date,omitempty"` // only for intention == TERMINATE
	Property        *wrapperspb.StringValue   `json:"property"`
	OfferConditions *offer.OfferConditions    `json:"offer_conditions"`
	ExpirationDate  *timestamppb.Timestamp    `json:"expiration_date,omitempty"` // only for intention == RENEW
}

type LandlordIntentionResponse struct {
	Success bool   `json:"success"`
	Tenancy string `json:"tenancy"`
	Offer   string `json:"new_offer"`
}

type SessionIDRequest struct {
	SessionID string `json:"session_id"`
}

type GetTenancyRequest struct {
	ID string `json:"id"`
}

type GetActiveTenancyRequest struct {
	PropertyID string `json:"property_id"`
}

type GetTenancyByOfferRequest struct {
	OfferID string `json:"offer_id"`
}

type GetTenancyByPropertyRequest struct {
	PropertyID string `json:"property_id"`
}

type GetTenancyResponse struct {
	Success bool              `json:"success"`
	Tenancy *property.Tenancy `json:"tenancy"`
}

type GetTenancyClaimInput struct {
	ByClaimID   string `json:"by_claim_id,omitempty"`
	ByTenancyID string `json:"by_tenancy_id,omitempty"`
}

type CreateTenancyClaimInput struct {
	TenancyID string                                  `json:"tenancy_id,omitempty"`
	Damages   []*property.TenancyClaim_Damage_Details `json:"damages,omitempty"`
}

type UpdateTenancyClaimDamageInput struct {
	ByClaimID   string                        `json:"by_claim_id,omitempty"`
	ByTenancyID string                        `json:"by_tenancy_id,omitempty"`
	Damage      *property.TenancyClaim_Damage `json:"damage,omitempty"`
}

type RemoveTenancyClaimDamageInput struct {
	ByClaimID   string `json:"by_claim_id,omitempty"`
	ByTenancyID string `json:"by_tenancy_id,omitempty"`
	DamageID    string `json:"damage_id,omitempty"`
}

type RemoveTenancyClaimInput struct {
	ByClaimID   string `json:"by_claim_id,omitempty"`
	ByTenancyID string `json:"by_tenancy_id,omitempty"`
}

type TenancyClaimSubmitSettlementInput struct {
	ByClaimID   string `json:"by_claim_id,omitempty"`
	ByTenancyID string `json:"by_tenancy_id,omitempty"`
}

type TenancyClaimGetSettlementInput struct {
	ByClaimID   string `json:"by_claim_id,omitempty"`
	ByTenancyID string `json:"by_tenancy_id,omitempty"`
}

func CensoredAccount(acc *account.Account) *account.Account {
	return &account.Account{
		Id:        acc.GetId(),
		CreatedAt: acc.GetCreatedAt(),
		Details:   acc.GetDetails(),
		Contact:   acc.GetContact(),
	}
}

func initProperty(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "PropertyQuery",
		Description: "Property query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Retrieves specified property",
				Args:        util.BindArguments(property.GetRequest{}),
				Resolve:     PropertyQueryGetResolver,
				Type:        util.BindOutput(property.GetResponse{}, "Property information", nil),
			},
			"getprivate": &graphql.Field{
				Description: "Retrieves specified property",
				Args:        util.BindArguments(property.GetPrivateRequest{}),
				Resolve:     PropertyQueryGetPrivateResolver,
				Type:        util.BindOutput(property.GetResponse{}, "Property information", nil),
			},
			"getprivatepreface": &graphql.Field{
				Description: "Retrieves specified property for preface user",
				Args:        util.BindArguments(property.GetPrivateRequest{}),
				Resolve:     PropertyQueryGetPrivatePrefaceResolver,
				Type:        util.BindOutput(property.GetResponse{}, "Property information", nil),
			},
			"my": &graphql.Field{
				DeprecationReason: "Use list",
				Description:       "Retrieves owned property",
				Args:              util.BindArguments(property.PropertyListRequest{}),
				Resolve:           PropertyQueryListResolver,
				Type:              util.BindOutput(property.PropertyListResponse{}, "Property list information", nil),
			},
			"list": &graphql.Field{
				Description: "Retrieves owned property",
				Args:        util.BindArguments(property.PropertyListRequest{}),
				Resolve:     PropertyQueryListResolver,
				Type:        util.BindOutput(property.PropertyListResponse{}, "Property list information", nil),
			},
			"search": &graphql.Field{
				Description: "Searches property by provided criteria",
				Args:        util.BindArguments(property.PropertySearchRequest{}),
				Resolve:     PropertyQuerySearchResolver,
				Type: util.BindOutput(
					property.PropertySearchResponse{},
					"Property list information with pagination",
					nil,
				),
			},
			"searchinpolygon": &graphql.Field{
				Description: "Searches property within a geographical polygon",
				Args:        util.BindArguments(property.PropertySearchInPolygonRequest{}),
				Resolve:     PropertyQuerySearchInPolygonResolver,
				Type: util.BindOutput(
					property.PropertySearchResponse{},
					"Property list information with pagination",
					nil,
				),
			},
			"listview": &graphql.Field{
				DeprecationReason: "use viewing { list }",
				Description:       "Lists view requests for property",
				Args:              util.BindArguments(property.ListViewRequest{}),
				Resolve:           PropertyQueryListViewResolver,
				Type:              util.BindOutput(property.ListViewResponse{}, "Property view listing", nil),
			},
			"viewreserved": &graphql.Field{
				DeprecationReason: "use viewing { reserved }",
				Description:       "Reserved view intervals",
				Args:              util.BindArguments(property.ViewReservedRequest{}),
				Resolve:           PropertyQueryViewReservedResolver,
				Type:              util.BindOutput(property.ViewReservedResponse{}, "Reserved intervals", nil),
			},
			"gettenancy": &graphql.Field{
				Description: "Retrieves specified tenancy",
				Args:        util.BindArguments(GetTenancyRequest{}),
				Resolve:     PropertyQueryGetTenancyResolver,
				Type:        util.BindOutput(GetTenancyResponse{}, "Tenancy information", nil),
			},
			"gettenancybyoffer": &graphql.Field{
				Description: "Retrieves specified tenancy by offer",
				Args:        util.BindArguments(GetTenancyByOfferRequest{}),
				Resolve:     PropertyQueryGetTenancyByOfferResolver,
				Type:        util.BindOutput(GetTenancyResponse{}, "Tenancy information", nil),
			},
			"gettenancybyproperty": &graphql.Field{
				Description: "Retrieves list of tenancy by property id",
				Args:        util.BindArguments(GetTenancyByPropertyRequest{}),
				Resolve:     PropertyQueryGetTenancyByPropertyResolver,
				Type:        util.BindOutput(property.GetTenancyResponse{}, "List of Tenancy", nil),
			},
			"getactivetenancy": &graphql.Field{
				Description: "Retrieves active tenancy for property",
				Args:        util.BindArguments(GetActiveTenancyRequest{}),
				Resolve:     PropertyQueryGetActiveTenancyResolver,
				Type:        util.BindOutput(GetTenancyResponse{}, "Tenancy information", nil),
			},
			"tenancyClaimGet": &graphql.Field{
				Description: "Retrieve existent tenancy claim using one of provided IDs.",
				Args:        util.BindArguments(GetTenancyClaimInput{}),
				Resolve:     resolveGetTenancyClaim,
				Type:        util.BindOutput(property.TenancyClaim_GetResponse{}, "Complete tenancy claim data", nil),
			},
			"tenancyClaimGetSettlement": &graphql.Field{
				Description: "Retrieve tenancy claim settlement invoice using one of provided IDs.",
				Args:        util.BindArguments(TenancyClaimGetSettlementInput{}),
				Resolve:     resolveTenancyClaimGetSettlement,
				Type:        util.BindOutput(invoice.Invoice{}, "", nil),
			},
			"getaveragerentprices": &graphql.Field{
				Description: "Get average rent prices by bedrooms",
				Args:        util.BindArguments(property.GetAverageRentPricesRequest{}),
				Resolve:     PropertyQueryGetAverageRentPrices,
				Type: util.BindOutput(
					property.GetAverageRentPricesResponse{},
					"Average rent prices",
					nil,
				),
			},
			"valuation": &graphql.Field{
				Description: "Get property valuation",
				Args:        util.BindArguments(property.GetValuationRequest{}),
				Resolve:     PropertyQueryGetValuation,
				Type:        util.BindOutput(property.GetValuationResponse{}, "", nil),
			},
			"ownershipIsVerified": &graphql.Field{
				Description: "Get ownership verification",
				Args:        util.BindArguments(account.OwnershipIsVerifiedRequest{}),
				Resolve:     OwnershipIsVerified,
				Type:        util.BindOutput(account.OwnershipIsVerifiedResponse{}, "", nil),
			},
		},
	}))

	(*rootQueries)["property"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "PropertyMutation",
		Description: "Property mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Creates new property",
				Args: func() graphql.FieldConfigArgument {
					input := util.BindArguments(property.CreateRequest{})

					input["recaptcha_token"] = &graphql.ArgumentConfig{
						Type: graphql.String,
					}

					return input
				}(),
				Resolve: PropertyMutationCreateResolver,
				Type:    util.BindOutput(property.CreateResponse{}, "Create status", nil),
			},
			"update": &graphql.Field{
				Description: "Updates existing property",
				Args:        util.BindArguments(property.UpdateRequest{}),
				Resolve:     PropertyMutationUpdateResolver,
				Type:        util.BindOutput(property.UpdateResponse{}, "Create status", nil),
			},
			"delete": &graphql.Field{
				Description: "Deletes existing property",
				Args:        util.BindArguments(property.DeleteRequest{}),
				Resolve:     PropertyMutationDeleteResolver,
				Type:        util.BindOutput(property.DeleteResponse{}, "Delete status", nil),
			},
			"restore": &graphql.Field{
				Description: "Restore property",
				Args:        util.BindArguments(property.RestoreRequest{}),
				Resolve:     PropertyMutationRestoreResolver,
				Type:        util.BindOutput(property.RestoreResponse{}, "Restore status", nil),
			},
			"publish": &graphql.Field{
				Description: "Publish listing request",
				Args:        util.BindArguments(property.PublishRequest{}),
				Resolve:     PropertyMutationPublishResolver,
				Type:        util.BindOutput(property.PublishResponse{}, "Create listing status", nil),
			},
			"republish": &graphql.Field{
				Description: "Republish listing request",
				Args:        util.BindArguments(property.RepublishRequest{}),
				Resolve:     PropertyMutationRepublishResolver,
				Type:        util.BindOutput(property.RepublishResponse{}, "Recreate listing status", nil),
			},
			"publishconsent": &graphql.Field{
				Description: "Publish listing consent request",
				Args:        util.BindArguments(property.PublishConsentRequest{}),
				Resolve:     PropertyMutationPublishConsentResolver,
				Type:        util.BindOutput(property.PublishConsentResponse{}, "Consent listing status", nil),
			},
			"unpublish": &graphql.Field{
				Description: "Unpublish listing request",
				Args:        util.BindArguments(property.UnpublishRequest{}),
				Resolve:     PropertyMutationUnpublishResolver,
				Type:        util.BindOutput(property.UnpublishResponse{}, "Drop listing status", nil),
			},
			"claim": &graphql.Field{
				Description: "Claim property",
				Args:        util.BindArguments(property.ClaimRequest{}),
				Resolve:     PropertyMutationClaimResolver,
				Type:        util.BindOutput(property.ClaimResponse{}, "Claim status", nil),
			},
			"commitstate": &graphql.Field{
				Description: "Commit state",
				Args:        util.BindArguments(property.CommitStateRequest{}),
				Resolve:     PropertyMutationCommitstateResolver,
				Type:        util.BindOutput(property.CommitStateResponse{}, "Commit status", nil),
			},
			"requestview": &graphql.Field{
				DeprecationReason: "use viewing { request }",
				Description:       "Request property view",
				Args: func() graphql.FieldConfigArgument {
					input := util.BindArguments(property.RequestViewRequest{})

					input["recaptcha_token"] = &graphql.ArgumentConfig{
						Type: graphql.String,
					}

					return input
				}(),
				Resolve: PropertyMutationRequestViewResolver,
				Type:    util.BindOutput(property.RequestViewResponse{}, "Request view status", nil),
			},
			"cancelview": &graphql.Field{
				DeprecationReason: "use viewing { status }",
				Description:       "Cancel property view",
				Args:              util.BindArguments(property.CancelViewRequest{}),
				Resolve: PropertyMutationStatusViewLegacyResolver(
					viewing.ViewingState_VIEWING_STATE_CANCELLED,
				),
				Type: util.BindOutput(property.CancelViewResponse{}, "Cancel view status", nil),
			},
			"acceptview": &graphql.Field{
				DeprecationReason: "use viewing { status }",
				Description:       "Accept property view",
				Args:              util.BindArguments(property.AcceptViewRequest{}),
				Resolve: PropertyMutationStatusViewLegacyResolver(
					viewing.ViewingState_VIEWING_STATE_ACCEPTED,
				),
				Type: util.BindOutput(property.AcceptViewResponse{}, "Accept view status", nil),
			},
			"rejectview": &graphql.Field{
				DeprecationReason: "use viewing { status }",
				Description:       "Reject property view",
				Args:              util.BindArguments(property.RejectViewRequest{}),
				Resolve: PropertyMutationStatusViewLegacyResolver(
					viewing.ViewingState_VIEWING_STATE_REJECTED,
				),
				Type: util.BindOutput(property.RejectViewResponse{}, "Reject view status", nil),
			},
			"completeview": &graphql.Field{
				DeprecationReason: "use viewing { status }",
				Description:       "Complete property view",
				Args:              util.BindArguments(property.CompleteViewRequest{}),
				Resolve: PropertyMutationStatusViewLegacyResolver(
					viewing.ViewingState_VIEWING_STATE_COMPLETED,
				),
				Type: util.BindOutput(property.CompleteViewResponse{}, "Complete view status", nil),
			},
			"failview": &graphql.Field{
				DeprecationReason: "use viewing { status }",
				Description:       "Fail property view",
				Args:              util.BindArguments(property.FailViewRequest{}),
				Resolve: PropertyMutationStatusViewLegacyResolver(
					viewing.ViewingState_VIEWING_STATE_FAILED,
				),
				Type: util.BindOutput(property.FailViewResponse{}, "Fail view status", nil),
			},
			"changebankdetails": &graphql.Field{
				Name:    "Change bank details",
				Args:    util.BindArguments(property.ChangeBankDetailsRequest{}),
				Resolve: PropertyMutationChangeBankDetails,
				Type:    util.BindOutput(property.ChangeBankDetailsResponse{}, "Change bank details status", nil),
			},
			"withdraw": &graphql.Field{
				Description: "Withdraw property",
				Args:        util.BindArguments(property.WithdrawnRequest{}),
				Resolve:     adminMutationWithdrawPropertyResolver,
				Type:        util.BindOutput(property.WithdrawnResponse{}, "Withdraw property response", nil),
			},
			"savereminder": &graphql.Field{
				Description: "Save listing reminder",
				Args:        util.BindArguments(property.SaveListingReminderRequest{}),
				Resolve:     PropertyMutationSaveReminderResolver,
				Type: util.BindOutput(
					property.SaveListingReminderResponse{},
					"Save listing reminder response",
					nil,
				),
			},
			"leavefromlisting": &graphql.Field{
				Description: "Leave from listing",
				Args:        util.BindArguments(property.LeaveFromListingFormRequest{}),
				Resolve:     PropertyMutationLeaveFromListingResolver,
				Type: util.BindOutput(
					property.LeaveFromListingFormResponse{},
					"Leave from listing response",
					nil,
				),
			},
			"setLandlordIntention": &graphql.Field{
				Description: "Set landlord intention",
				Args:        util.BindArguments(LandlordIntentionRequest{}),
				Resolve:     PropertySetLandlordIntentionResolver,
				Type:        util.BindOutput(LandlordIntentionResponse{}, "Set landlord intention response", nil),
			},
			"setTenantsIntention": &graphql.Field{
				Description: "Set tenants intention",
				Args:        util.BindArguments(IntentionRequest{}),
				Resolve:     PropertySetTenantsIntentionResolver,
				Type:        util.BindOutput(property.UpdateTenancyResponse{}, "Set tenants intention response", nil),
			},
			"commitpreface": &graphql.Field{
				Description: "Create from preface",
				Args:        util.BindArguments(SessionIDRequest{}),
				Resolve:     PropertyMutationCreateFromPrefaceResolver,
				Type:        util.BindOutput(property.CreateResponse{}, "Create from preface", nil),
			},
			"tenancyClaimCreate": &graphql.Field{
				Description: "Create tenancy claim for completed tenancy (property supervisors only).",
				Args:        util.BindArguments(CreateTenancyClaimInput{}),
				Resolve:     resolveCreateTenancyClaim,
				Type:        util.BindOutput(property.TenancyClaim_CreateResponse{}, "", nil),
			},
			"tenancyClaimUpdateDamage": &graphql.Field{
				Description: "Update specified tenancy claim damage using one of provided IDs.",
				Args:        util.BindArguments(UpdateTenancyClaimDamageInput{}),
				Resolve:     resolveUpdateTenancyClaimDamage,
				Type:        util.BindOutput(property.TenancyClaim_UpdateDamageResponse{}, "", nil),
			},
			"tenancyClaimRemoveDamage": &graphql.Field{
				Description: "Remove specified claim damage using one of provided IDs (property supervisors only).",
				Args:        util.BindArguments(RemoveTenancyClaimDamageInput{}),
				Resolve:     resolveRemoveTenancyClaimDamage,
				Type:        util.BindOutput(property.TenancyClaim_RemoveDamageResponse{}, "", nil),
			},
			"tenancyClaimRemove": &graphql.Field{
				Description: "Completely remove tenancy claim using one of provided IDs (property supervisors only).",
				Args:        util.BindArguments(RemoveTenancyClaimInput{}),
				Resolve:     resolveRemoveTenancyClaim,
				Type:        util.BindOutput(property.TenancyClaim_RemoveResponse{}, "", nil),
			},
			"tenancyClaimSubmitSettlement": &graphql.Field{
				Description: "Move tenancy claim state into SETTLEMENT" +
					" and create invoice to pay damages using one of provided IDs (tenants only).",
				Args:    util.BindArguments(TenancyClaimSubmitSettlementInput{}),
				Resolve: resolveTenancyClaimSubmitSettlement,
				Type:    util.BindOutput(invoice.Invoice{}, "", nil),
			},
		},
	}))

	(*rootMutations)["property"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func ForeignInvite(user *auth.AuthResponse, prop *property.PropertyPeople, inv *property.InvitationStatus) bool {
	if !ForeignProperty(user, prop) {
		return false
	}

	var foreignID, foreignContact bool

	switch user.GetAccount().GetIdValue() {
	case inv.GetInviterId():
	case inv.GetInviteeId():
	default:
		foreignID = true
	}

	switch {
	case user.GetAccount().GetContactValue().GetEmailPrimaryValue() == inv.GetInviteeEmail():
	case user.GetAccount().GetContactValue().GetPhonePrimaryValue() == inv.GetInviteePhone():
	default:
		foreignContact = true

		for _, p := range user.GetAccount().GetContactValue().GetPhonesSecondaryValue().GetValue() {
			if p.GetNumber() == inv.GetInviteePhone() {
				foreignContact = false
			}
		}
	}

	return foreignID && foreignContact
}

func ForeignProperty(user *auth.AuthResponse, prop *property.PropertyPeople) bool {
	if user == nil || prop == nil {
		return true
	}

	if HasRole(user, AdminRole) {
		return false
	}

	switch user.GetAccount().GetIdValue() {
	case prop.GetLandlord().GetValue():
	case prop.GetSubmitter().GetValue():
	case prop.GetManager().GetValue():
	default:
		return true
	}

	return false
}

func PropertyQueryGetPrefaceResolver(params graphql.ResolveParams, session string) (interface{}, error) {
	resp, err := connections.PropertyService.GetPreface(
		params.Context,
		&property.GetPrefaceRequest{
			By: &property.GetPrefaceRequest_ClientId{ClientId: session},
		},
	)
	if err != nil && !errors.Is(err, grpcerr.NotFound) {
		return nil, err
	}

	if resp == nil {
		resp = &property.GetPrefaceResponse{}
	}

	if resp.GetPreface() == nil {
		resp.Preface = &property.Preface{}
	}

	if resp.GetPreface().GetClientId() == "" {
		resp.GetPreface().ClientId = session
	}

	return &property.GetResponse{
		Success:      true,
		Property:     renderPrefaceProperty(resp.GetPreface()),
		Verification: renderPrefaceVerification(resp.GetVerification()),
	}, nil
}

func PropertyQueryGetResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	request := &property.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if tuser != nil && request.GetId() == tuser.Session {
		return PropertyQueryGetPrefaceResolver(params, tuser.Session)
	}

	if user != nil {
		request.User = user.GetAccount().GetIdValue()
		request.Test = user.GetAccount().GetTestValue()
	}

	resp, err := connections.PropertyService.Get(params.Context, request)
	if err != nil {
		return nil, err
	}

	prop := resp.Property

	if prop.GetId() != "" && request.User != "" {
		rtoStateName, err := getRTOStateName(params.Context, prop.GetId(), request.User)
		if err == nil && rtoStateName != "" {
			resp.Property.RtoState = rtoStateName
		}
	}

	// It's a temporary solution until the property service is refactored, where there will be 2 implementations of
	// property.Get method (for users and for services)
	if user == nil && (prop.GetIsArchived().GetValue() ||
		(prop.Status != property.PropertyStatus_LISTED &&
			prop.Status != property.PropertyStatus_OFFERING &&
			prop.Status != property.PropertyStatus_RENTED &&
			prop.Status != property.PropertyStatus_AWAITING_DEPOSIT)) {
		return nil, errs.ErrNoUser
	}
	return resp, nil
}

func getRTOStateName(context context.Context, propID string, userID string) (string, error) {
	rtoResp, err := connections.RtoServiceClient.GetByPropertyID(context,
		&rto.GetByPropertyIDRequest{
			PropertyId: propID,
			UserId:     userID,
		})
	if err != nil {
		return "", err
	}

	if rtoResp.Rto != nil {
		rtoStateName, ok := rto.RtoState_name[int32(rtoResp.Rto.State)]
		if ok {
			return rtoStateName, nil
		}
	}

	return "", nil
}

func PropertyQueryGetPrivateResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	request := &property.GetPrivateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.IsAdmin = HasRole(user, AdminRole)
	request.Test = user.GetAccount().GetTestValue()

	resp, err := connections.PropertyService.GetPrivate(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyQueryGetPrivatePrefaceResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &property.GetPrivateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	id := request.GetId()

	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user != nil {
		if id == "" {
			id = user.GetSession()
		}

		return PropertyQueryGetPrefaceResolver(params, id)
	}

	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if tuser != nil {
		if id == "" {
			id = tuser.Session
		}

		return PropertyQueryGetPrefaceResolver(params, id)
	}

	return nil, errs.ErrNoUser
}

func PropertyQueryListPrefaceResolver(params graphql.ResolveParams) (interface{}, error) {
	tuser, ok := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if !ok {
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	resp, err := connections.PropertyService.GetPreface(
		params.Context,
		&property.GetPrefaceRequest{
			By: &property.GetPrefaceRequest_ClientId{ClientId: tuser.Session},
		},
	)
	if err != nil {
		if errors.Is(err, grpcerr.NotFound) {
			return &property.PropertyListResponse{
				Success:    true,
				Properties: []*property.Property{},
				Total:      0,
			}, nil
		}

		return nil, err
	}

	return &property.PropertyListResponse{
		Success: true,
		Properties: []*property.Property{
			renderPrefaceProperty(resp.GetPreface()),
		},
		Total: 1,
	}, nil
}

func PropertyQueryListResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return PropertyQueryListPrefaceResolver(params)
	}

	request := &property.PropertyListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = wrap.StringValueFromOneOf(user.GetAccount().GetId(), user.GetAccount().GetIdValue())
	request.Test = user.GetAccount().GetTestValue()

	resp, err := connections.PropertyService.List(params.Context, request)
	if err != nil {
		return nil, err
	}
	var propertyIDs []string
	for _, property := range resp.GetProperties() {
		propertyIDs = append(propertyIDs, property.Id)
	}
	if len(propertyIDs) > 0 {
		rtosByPropertyIDs, err := connections.RtoServiceClient.GetByPropertyIDs(params.Context,
			&rto.GetByPropertyIDsRequest{
				PropertyIds: propertyIDs,
				UserId:      user.GetAccount().GetIdValue(),
			})
		if err != nil {
			return nil, err
		}
		for _, rtoModel := range rtosByPropertyIDs.GetRtos() {
			for _, property := range resp.GetProperties() {
				if property.Id == rtoModel.Property {
					rtoStateName, ok := rto.RtoState_name[int32(rtoModel.State)]
					if ok {
						property.RtoState = rtoStateName
					}
					break
				}
			}
		}
	}

	return resp, nil
}

func PropertyQuerySearchResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &property.PropertySearchRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.Search(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyQuerySearchInPolygonResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &property.PropertySearchInPolygonRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.SearchInPolygon(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyQueryViewReservedResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &property.ViewReservedRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.ViewingService.ViewingReserved(
		grpcctx.WithServiceUser(params.Context),
		&viewing.ViewingReservedRequest{
			Spec: &viewing.ViewingSpec{
				Select: &viewing.ViewingSpecSelect{
					PropertyId: request.GetProperty(),
				},
			},
		},
	)
	if err != nil {
		return nil, err
	}

	var list []*property.TakenInterval

	for _, r := range resp.GetIntervals() {
		list = append(list, &property.TakenInterval{
			TimeStartsec: wrapperspb.Int32(r.GetTimeStartsec()),
			TimeEndsec:   wrapperspb.Int32(r.GetTimeEndsec()),
			Day:          wrapperspb.Int32(r.GetDay()),
			Month:        wrapperspb.Int32(r.GetMonth()),
			Year:         wrapperspb.Int32(r.GetYear()),
		})
	}

	return &property.ViewReservedResponse{
		Success: true,
		Errors:  nil,
		List:    list,
	}, nil
}

func PropertyQueryGetTenancyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &GetTenancyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.TenancyService.TenancyGet(params.Context, &tenancy.TenancyGetRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
			Details: &tenancy.TenancySpec_TenancyId{
				TenancyId: request.ID,
			},
		},
	})
	if err != nil {
		return nil, err
	}

	return &GetTenancyResponse{
		Success: true,
		Tenancy: legacy.ConvertTenancyLegacy(resp.GetTenancy()),
	}, nil
}

func PropertyQueryGetTenancyByOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &GetTenancyByOfferRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.TenancyService.TenancyGet(params.Context, &tenancy.TenancyGetRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
			Details: &tenancy.TenancySpec_Select{
				Select: &tenancy.TenancySpecSelect{
					By: &tenancy.TenancySpecSelect_OfferId{
						OfferId: request.OfferID,
					},
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	return &GetTenancyResponse{
		Success: true,
		Tenancy: legacy.ConvertTenancyLegacy(resp.GetTenancy()),
	}, nil
}

func PropertyQueryGetTenancyByPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	request := &GetTenancyByPropertyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.TenancyService.TenancyList(params.Context, &tenancy.TenancyListRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
			Details: &tenancy.TenancySpec_Select{
				Select: &tenancy.TenancySpecSelect{
					By: &tenancy.TenancySpecSelect_PropertyId{
						PropertyId: request.PropertyID,
					},
				},
			},
		},
	})

	return &property.GetTenancyResponse{
		Success: true,
		Tenancy: legacy.ConvertTenanciesLegacy(resp.GetTenancies()),
	}, err
}

func PropertyQueryGetActiveTenancyResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	request := &GetActiveTenancyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.TenancyService.TenancyGet(params.Context, &tenancy.TenancyGetRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
			Details: &tenancy.TenancySpec_Select{
				Select: &tenancy.TenancySpecSelect{
					States: []tenancy.TenancyState{tenancy.TenancyState_TENANCY_ACTIVE},
					By: &tenancy.TenancySpecSelect_PropertyId{
						PropertyId: request.PropertyID,
					},
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	res := &GetTenancyResponse{
		Success: true,
	}

	res.Tenancy = legacy.ConvertTenancyLegacy(resp.GetTenancy())

	return res, nil
}

func convertTenancyClaimLegacy(claim *tenancy.TenancyClaim) *property.TenancyClaim {
	return &property.TenancyClaim{
		// Id: claim.GetTenancyId(),
		Id:         claim.ClaimId,
		TenancyId:  claim.GetTenancyId(),
		PropertyId: claim.GetPropertyId(),
		CreatedAt:  claim.GetCreatedAt(),
		UpdatedAt:  claim.GetUpdatedAt(),
		State:      property.TenancyClaim_State(claim.GetState()),
		TotalCost:  claim.GetTotalCost(),
		Damages:    convertTenancyClaimDamagesLegacy(claim.GetDamages()),
	}
}

func convertTenancyClaimDamage(m *property.TenancyClaim_Damage) *tenancy.TenancyClaimDamage {
	return &tenancy.TenancyClaimDamage{
		Id:      m.GetId(),
		Details: convertTenancyClaimDetail(m.GetDetails()),
		Dispute: convertTenancyClaimDispute(m.GetDispute()),
	}
}

func convertTenancyClaimDispute(m *property.TenancyClaim_Damage_Dispute) *tenancy.TenancyClaimDamageDispute {
	return &tenancy.TenancyClaimDamageDispute{
		Explanation: m.GetExplanation(),
		Pay:         m.GetPay(),
		Proofs:      convertTenancyClaimProofs(m.GetProofs()),
	}
}

func convertTenancyClaimDetails(ms []*property.TenancyClaim_Damage_Details) (ts []*tenancy.TenancyClaimDamageDetails) {
	for _, m := range ms {
		ts = append(ts, convertTenancyClaimDetail(m))
	}

	return
}

func convertTenancyClaimDetail(m *property.TenancyClaim_Damage_Details) *tenancy.TenancyClaimDamageDetails {
	return &tenancy.TenancyClaimDamageDetails{
		Name:        m.GetName(),
		Description: m.GetDescription(),
		Cost:        m.GetCost(),
		Proofs:      convertTenancyClaimProofs(m.GetProofs()),
	}
}

func convertTenancyClaimProofs(ms []*property.TenancyClaim_Proof) (ts []*tenancy.TenancyClaimProof) {
	for _, m := range ms {
		ts = append(ts, &tenancy.TenancyClaimProof{
			FileId:   m.GetFileId(),
			Filename: m.GetFilename(),
		})
	}

	return
}

func convertTenancyClaimDamagesLegacy(ms []*tenancy.TenancyClaimDamage) (ts []*property.TenancyClaim_Damage) {
	for _, m := range ms {
		ts = append(ts, &property.TenancyClaim_Damage{
			Id:      m.GetId(),
			Details: convertTenancyClaimDetailsLegacy(m.GetDetails()),
			Dispute: convertTenancyClaimDisputeLegacy(m.GetDispute()),
		})
	}

	return
}

func convertTenancyClaimDisputeLegacy(m *tenancy.TenancyClaimDamageDispute) *property.TenancyClaim_Damage_Dispute {
	return &property.TenancyClaim_Damage_Dispute{
		Explanation: m.GetExplanation(),
		Pay:         m.GetPay(),
		Proofs:      convertTenancyClaimProofsLegacy(m.GetProofs()),
	}
}

func convertTenancyClaimDetailsLegacy(m *tenancy.TenancyClaimDamageDetails) *property.TenancyClaim_Damage_Details {
	return &property.TenancyClaim_Damage_Details{
		Name:        m.GetName(),
		Description: m.GetDescription(),
		Cost:        m.GetCost(),
		Proofs:      convertTenancyClaimProofsLegacy(m.GetProofs()),
	}
}

func convertTenancyClaimProofsLegacy(ms []*tenancy.TenancyClaimProof) (ts []*property.TenancyClaim_Proof) {
	for _, m := range ms {
		ts = append(ts, &property.TenancyClaim_Proof{
			FileId:   m.GetFileId(),
			Filename: m.GetFilename(),
		})
	}

	return
}

func resolveGetTenancyClaim(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := GetTenancyClaimInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &tenancy.TenancyClaimGetRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
		},
	}
	switch {
	case input.ByClaimID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	resp, err := connections.TenancyService.TenancyClaimGet(params.Context, forward)
	if err != nil {
		return nil, err
	}

	if resp.GetClaim().GetState() == 0 {
		return nil, nil
	}

	return &property.TenancyClaim_GetResponse{
		Claim: convertTenancyClaimLegacy(resp.GetClaim()),
	}, nil
}

func resolveTenancyClaimGetSettlement(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := TenancyClaimGetSettlementInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &tenancy.TenancyClaimSettlementRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
		},
	}

	switch {
	case input.ByClaimID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	settlement, err := connections.TenancyService.TenancyClaimGetSettlement(params.Context, forward)
	if err != nil {
		return nil, err
	}

	// TODO resolve user matching
	// Any from tenants can request settlement after creation,
	// but invoice is associated with only one creator (tenant).
	// Thus here we can potentially get an error, if different tenant user make source GQL-request.
	// This case can be resolved by adding creator ID to settlement data.

	inv, err := connections.InvoiceService.InvoiceGet(params.Context, &invoice.InvoiceGetRequest{
		InvoiceId:        settlement.GetSettlement().GetInvoiceId(),
		InvoiceReference: settlement.GetSettlement().GetInvoiceReference(),
		User:             user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return inv.GetInvoice(), nil
}

func renderPrefaceProperty(p *property.Preface) *property.Property {
	return &property.Property{
		Id:          p.GetClientId(),
		CreatedAt:   p.GetCreatedAt(),
		UpdatedAt:   p.GetUpdatedAt(),
		Location:    p.GetLocation(),
		Details:     p.GetDetails(),
		Rental:      p.GetRental(),
		Description: p.GetDescription(),
		Valuation:   p.GetValuation(),
	}
}

func renderPrefaceVerification(p *property.PrefaceVerification) *property.PropertyVerification {
	return &property.PropertyVerification{
		Location:    p.GetLocation(),
		Details:     p.GetDetails(),
		Rental:      p.GetRental(),
		Description: p.GetDescription(),
	}
}

func PropertyQueryGetAverageRentPrices(params graphql.ResolveParams) (interface{}, error) {
	request := &property.GetAverageRentPricesRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.GetAverageRentPrices(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyMutationCreateTransientResolver(params graphql.ResolveParams) (interface{}, error) {
	tuser, ok := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if !ok {
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	request := &property.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.SavePreface(
		params.Context,
		&property.SavePrefaceRequest{
			ClientId:    tuser.Session,
			Location:    request.GetProperty().GetLocation(),
			Details:     request.GetProperty().GetDetails(),
			Rental:      request.GetProperty().GetRental(),
			Description: request.GetProperty().GetDescription(),
		},
	)
	if err != nil {
		return nil, err
	}

	return &property.CreateResponse{
		Success:      true,
		Property:     renderPrefaceProperty(resp.GetPreface()),
		Verification: renderPrefaceVerification(resp.GetVerification()),
	}, nil
}

func PropertyMutationCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return PropertyMutationCreateTransientResolver(params)
	}

	request := &property.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	/*v, ok := params.Args["recaptcha_token"]
	if !ok {
		return nil, grpcerr.InvalidData
	}

	err := recaptcha.CheckRecaptchaToken(params.Context, v.(string))
	if err != nil {
		return nil, err
	}*/

	request.User = user.GetAccount().GetIdValue()
	resp, err := connections.PropertyService.Create(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func PropertyMutationUpdateTransientResolver(params graphql.ResolveParams) (interface{}, error) {
	tuser, ok := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if !ok {
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	request := &property.UpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.SavePreface(
		params.Context,
		&property.SavePrefaceRequest{
			ClientId:    tuser.Session,
			Location:    request.GetProperty().GetLocation(),
			Details:     request.GetProperty().GetDetails(),
			Rental:      request.GetProperty().GetRental(),
			Description: request.GetProperty().GetDescription(),
		},
	)
	if err != nil {
		return nil, err
	}

	return &property.UpdateResponse{
		Success:      true,
		Property:     renderPrefaceProperty(resp.GetPreface()),
		Verification: renderPrefaceVerification(resp.GetVerification()),
	}, nil
}

func PropertyMutationUpdateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return PropertyMutationUpdateTransientResolver(params)
	}

	request := &property.UpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()

	resp, err := connections.PropertyService.Update(params.Context, request)
	if err != nil {
		return nil, err
	}

	if request.GetProperty().GetPendingExtraServices() != nil {
		var products []*invoice.Product

		pid := request.GetProperty().GetId()
		if pid == "" {
			pid = request.GetProperty().GetLegacyId().GetValue()
		}

		for _, p := range request.GetProperty().GetPendingExtraServices().GetExtraServices() {
			products = append(products, &invoice.Product{
				ServiceId: p,
				Details: &invoice.ProductDetails{
					Property: &invoice.ProductDetailsProperty{
						Id: wrapperspb.String(pid),
					},
				},
			})
		}

		_, err = connections.InvoiceService.OrderSave(params.Context, &invoice.OrderSaveRequest{
			Userid:            request.GetUser(),
			Products:          products,
			Periodic:          false,
			CorrelationType:   invoice.CorrelationTypeProperty,
			CorrelationId:     pid,
			CorrelationTarget: invoice.CorrelationTargetListingServices,
		})
		if err != nil {
			return nil, err
		}
	}

	return resp, nil
}

func PropertyMutationDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &property.DeleteRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()

	resp, err := connections.PropertyService.Delete(params.Context, request)
	if err != nil {
		return nil, err
	}

	_, err = connections.OfferService.DeleteOffersByProperty(params.Context, &offer.DeleteOffersByPropertyRequest{
		Property: request.GetId(),
		User:     user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyMutationRestoreResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &property.RestoreRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.Restore(params.Context, request)
}

func PropertyMutationPublishResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.PublishRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.Publish(params.Context, request)
}

func PropertyMutationRepublishResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.RepublishRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()

	resp, err := connections.TenancyService.TenancyGet(params.Context, &tenancy.TenancyGetRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
			Details: &tenancy.TenancySpec_Select{
				Select: &tenancy.TenancySpecSelect{
					States: []tenancy.TenancyState{tenancy.TenancyState_TENANCY_ACTIVE},
					By: &tenancy.TenancySpecSelect_PropertyId{
						PropertyId: request.GetProperty(),
					},
				},
			},
		},
	})
	if err != nil && !errors.Is(err, grpcerr.NotFound) {
		return nil, err
	}

	if resp != nil {
		llIntention := resp.GetTenancy().GetLandlordIntention()
		now := time.Now().UTC()

		if resp.GetTenancy().GetTenantsIntention().GetChoice() != tenancy.TenancyIntentionChoice_INTENTION_TERMINATE &&
			llIntention.GetChoice() == tenancy.TenancyIntentionChoice_INTENTION_RENEW &&
			!now.Truncate(time.Hour*24).After(llIntention.GetExpirationDate().AsTime().UTC()) {
			return nil, grpcerr.New(grpcerr.InvalidState, "active tenancy is waiting for renewal")
		}
	}

	return connections.PropertyService.Republish(params.Context, request)
}

func PropertyMutationPublishConsentResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.PublishConsentRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.PublishConsent(params.Context, request)
}

func PropertyMutationUnpublishResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.UnpublishRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.Unpublish(params.Context, request)
}

func PropertyMutationClaimResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.ClaimRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.Claim(params.Context, request)
}

func PropertyMutationCommitstateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.CommitStateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.CommitState(params.Context, request)
}

func PropertyQueryListViewResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.ListViewRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.ViewingService.ViewingList(params.Context, &viewing.ViewingListRequest{
		Spec: &viewing.ViewingSpec{
			UserId: user.GetAccount().GetIdValue(),
			Select: &viewing.ViewingSpecSelect{
				States:     nil,
				PropertyId: request.GetProperty().GetValue(),
				ListingId:  request.GetListing(),
				OfferId:    "",
				ViewingId:  "",
				Token:      "",
			},
		},
	})
	if err != nil {
		return nil, err
	}

	return &property.ListAllViewResponse{
		Success: true,
		Errors:  nil,
		List:    legacy.ConvertLegacyViewingsLegacy(resp.GetViewings()),
	}, nil
}

func PropertyMutationRequestViewResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &property.RequestViewRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.ViewingService.ViewingSave(
		grpcctx.ValuesBackgroundContext(params.Context),
		&viewing.ViewingSaveRequest{
			ViewingId: request.GetRequest().GetId(),
			Token:     request.GetToken(),
			Property:  request.GetRequest().GetProperty(),
			User:      user.GetAccount().GetIdValue(),
			Time: &viewing.ViewingTime{
				TimeStartsec: request.GetRequest().GetTime().GetTimeStartsec().GetValue(),
				TimeEndsec:   request.GetRequest().GetTime().GetTimeEndsec().GetValue(),
				Day:          request.GetRequest().GetTime().GetDay().GetValue(),
				Month:        request.GetRequest().GetTime().GetMonth().GetValue(),
				Year:         request.GetRequest().GetTime().GetYear().GetValue(),
			},
			Details: &viewing.ViewingDetails{
				Flatmates:              request.GetRequest().GetDetails().GetFlatmates().GetValue(),
				MoveinDate:             viewing.MoveinDate(request.GetRequest().GetDetails().GetMoveinDate()),
				HavePets:               request.GetRequest().GetDetails().GetHavePets().GetValue(),
				PetsDescription:        request.GetRequest().GetDetails().GetPetsDescription().GetValue(),
				HaveHouseBenefits:      request.GetRequest().GetDetails().GetHaveHouseBenefits().GetValue(),
				BenefitDetails:         request.GetRequest().GetDetails().GetBenefitDetails().GetValue(),
				ViewingDescription:     request.GetRequest().GetDetails().GetViewingDescription().GetValue(),
				CombinedHousehold:      request.GetRequest().GetDetails().GetCombinedHousehold().GetValue(),
				CountryCourtJudgements: request.GetRequest().GetDetails().GetCountryCourtJudgements().GetValue(),
				Video:                  request.GetRequest().GetDetails().GetVideo().GetValue(),
			},
			Remotely: request.GetRequest().GetRemotely().GetValue(),
		},
	)
	if err != nil {
		return nil, err
	}

	return &property.RequestViewResponse{
		Success: true,
		Errors:  nil,
		Request: legacy.ConvertViewingLegacy(resp.GetViewing()),
	}, nil
}

func PropertyMutationStatusViewLegacyResolver(
	status viewing.ViewingState,
) func(params graphql.ResolveParams) (interface{}, error) {
	return func(params graphql.ResolveParams) (interface{}, error) {
		user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
		if user == nil {
			return nil, errs.ErrNoUser
		}

		request := &property.CancelViewRequest{}

		if err := util.UnmarshalArguments(params.Args, request); err != nil {
			return nil, err
		}
		request.User = user.GetAccount().GetIdValue()

		resp, err := connections.ViewingService.ViewingStatus(
			grpcctx.ValuesBackgroundContext(params.Context),
			&viewing.ViewingStatusRequest{
				Spec: &viewing.ViewingSpec{
					UserId: user.GetAccount().GetIdValue(),
					Select: &viewing.ViewingSpecSelect{
						States:     nil,
						PropertyId: request.GetProperty(),
						ListingId:  "",
						OfferId:    "",
						ViewingId:  request.GetId(),
						Token:      request.GetToken(),
					},
				},
				Transition: status,
			},
		)
		if err != nil {
			return nil, err
		}

		return &property.CancelViewResponse{
			Success: true,
			Errors:  nil,
			Request: legacy.ConvertViewingLegacy(resp.GetViewing()),
		}, nil
	}
}

func PropertyMutationChangeBankDetails(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.ChangeBankDetailsRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	if remote, ok := params.Context.Value(grpcctx.CtxRemote).(string); ok {
		request.Remote = remote
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.ChangeBankDetails(grpcctx.ValuesBackgroundContext(params.Context), request)
}

func adminMutationWithdrawPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	req := &property.WithdrawnRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	request := &property.WithdrawnRequest{
		Id:        req.Id,
		User:      user.GetAccount().GetIdValue(),
		AdminSite: false,
	}

	return connections.PropertyService.ServiceWithdrawn(params.Context, request)
}

func PropertyMutationSaveReminderResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &property.SaveListingReminderRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.PropertyService.SaveListingReminder(params.Context, request)
}

func PropertyMutationLeaveFromListingResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.LeaveFromListingFormRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.LeaveFromListingForm(params.Context, request)
}

func PropertySetLandlordIntentionResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &LandlordIntentionRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	ctx := params.Context

	spec := &tenancy.TenancySpec{
		UserId: user.GetAccount().GetIdValue(),
	}

	if request.Tenancy.GetValue() != "" {
		spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: request.Tenancy.GetValue()}
	} else {
		spec.Details = &tenancy.TenancySpec_Select{
			Select: &tenancy.TenancySpecSelect{
				States: []tenancy.TenancyState{tenancy.TenancyState_TENANCY_ACTIVE},
				By: &tenancy.TenancySpecSelect_PropertyId{
					PropertyId: request.Property.GetValue(),
				},
			},
		}
	}

	resp, err := connections.TenancyService.TenancyLandlordIntention(ctx, &tenancy.TenancyLandlordIntentionRequest{
		Spec:            spec,
		Choice:          tenancy.TenancyIntentionChoice(request.Intention),
		TerminationDate: request.TerminationDate,
		ExpirationDate:  request.ExpirationDate,
	})
	if err != nil {
		return nil, err
	}

	res := &LandlordIntentionResponse{
		Success: true,
		Tenancy: resp.GetTenancy().GetId(),
	}

	if request.OfferConditions != nil && request.Intention == property.TenancyIntention_INTENTION_RENEW {
		var ren *offer.CreateResponse

		ren, err = connections.OfferService.CreateForRenew(ctx, &offer.CreateForRenewRequest{
			User: user.GetAccount().GetIdValue(),
			Offer: &offer.Offer{
				Id:         resp.GetTenancy().GetOffer(),
				Property:   resp.GetTenancy().GetProperty(),
				Conditions: request.OfferConditions,
			},
			Tenancy: resp.GetTenancy().GetId(),
		})
		if err != nil {
			return nil, err
		}

		res.Offer = ren.GetOffer().GetId()
	}

	return res, nil
}

func PropertySetTenantsIntentionResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &IntentionRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var penaltyConsent *timestamppb.Timestamp

	if request.PenaltyConsent.GetValue() {
		penaltyConsent = timestamppb.Now()
	}

	_, err := connections.TenancyService.TenancyTenantIntention(params.Context, &tenancy.TenancyTenantIntentionRequest{
		Spec: &tenancy.TenancySpec{
			UserId:  user.GetAccount().GetIdValue(),
			Details: &tenancy.TenancySpec_TenancyId{TenancyId: request.Tenancy.GetValue()},
		},
		Choice:           tenancy.TenancyIntentionChoice(request.Intention),
		TerminationDate:  request.TerminationDate,
		PenaltyConsented: penaltyConsent,
	})
	if err != nil {
		return nil, err
	}

	return &property.UpdateTenancyResponse{
		Success: true,
		Errors:  nil,
	}, nil
}

func PropertyMutationCreateFromPrefaceResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &SessionIDRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.CreateFromPreface(params.Context, &property.CreateFromPrefaceRequest{
		By: &property.CreateFromPrefaceRequest_ClientId{
			ClientId: request.SessionID,
		},
		Owner: user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func resolveCreateTenancyClaim(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := CreateTenancyClaimInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	resp, err := connections.TenancyService.TenancyClaimCreate(params.Context, &tenancy.TenancyClaimCreateRequest{
		Spec: &tenancy.TenancySpec{
			UserId:  user.GetAccount().GetIdValue(),
			Details: &tenancy.TenancySpec_TenancyId{TenancyId: input.TenancyID},
		},
		Details: convertTenancyClaimDetails(input.Damages),
	})
	if err != nil {
		return nil, err
	}

	return &property.TenancyClaim_CreateResponse{
		Claim: convertTenancyClaimLegacy(resp.GetClaim()),
	}, nil
}

func resolveUpdateTenancyClaimDamage(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := UpdateTenancyClaimDamageInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &tenancy.TenancyClaimUpdateDamageRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
		},
		Damage: convertTenancyClaimDamage(input.Damage),
	}

	switch {
	case input.ByClaimID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	resp, err := connections.TenancyService.TenancyClaimUpdateDamage(params.Context, forward)
	if err != nil {
		return nil, err
	}

	return &property.TenancyClaim_UpdateDamageResponse{
		Claim: convertTenancyClaimLegacy(resp.GetClaim()),
	}, nil
}

func resolveRemoveTenancyClaimDamage(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := RemoveTenancyClaimDamageInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &tenancy.TenancyClaimRemoveDamageRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
		},
		DamageId: input.DamageID,
	}

	switch {
	case input.ByClaimID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	resp, err := connections.TenancyService.TenancyClaimRemoveDamage(params.Context, forward)
	if err != nil {
		return nil, err
	}

	return &property.TenancyClaim_RemoveDamageResponse{
		Claim: convertTenancyClaimLegacy(resp.GetClaim()),
	}, nil
}

func resolveRemoveTenancyClaim(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := RemoveTenancyClaimInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &tenancy.TenancyClaimRemoveRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
		},
	}
	switch {
	case input.ByClaimID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	_, err := connections.TenancyService.TenancyClaimRemove(params.Context, forward)
	if err != nil {
		return nil, err
	}

	return &property.TenancyClaim_RemoveResponse{
		Success: true,
	}, nil
}

func resolveTenancyClaimSubmitSettlement(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := TenancyClaimSubmitSettlementInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &tenancy.TenancyClaimSettlementRequest{
		Spec: &tenancy.TenancySpec{
			UserId: user.GetAccount().GetIdValue(),
		},
	}

	switch {
	case input.ByClaimID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.Spec.Details = &tenancy.TenancySpec_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	settlement, err := connections.TenancyService.TenancyClaimSubmitSettlement(params.Context, forward)
	if err != nil {
		return nil, err
	}

	inv, err := connections.InvoiceService.InvoiceGet(params.Context, &invoice.InvoiceGetRequest{
		InvoiceId:        settlement.GetSettlement().GetInvoiceId(),
		InvoiceReference: settlement.GetSettlement().GetInvoiceReference(),
		User:             user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return inv.GetInvoice(), nil
}

func PropertyQueryGetValuation(params graphql.ResolveParams) (interface{}, error) {
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	req := &property.GetValuationRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	req.User = user.GetAccount().GetIdValue()

	resp, err := connections.PropertyService.GetValuation(params.Context, req)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func OwnershipIsVerified(params graphql.ResolveParams) (interface{}, error) {
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	req := &account.OwnershipIsVerifiedRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	resp, err := connections.AccountService.OwnershipIsVerified(params.Context, req)
	if err != nil {
		return nil, err
	}

	return resp, nil
}
