package resolvers

import (
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	"gitlab.mashplace.com/mashplace/backend/offer/api/rto"

	"github.com/graphql-go/graphql"
)

func initRto(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "RtoQuery",
		Description: "Rto query root",
		Fields: graphql.Fields{
			"getbypropertyid": &graphql.Field{
				Description: "Retrieves specified rto by property id",
				Args:        util.BindArguments(rto.GetByPropertyIDRequest{}),
				Resolve:     RtoQueryGetByPropertyIDResolver,
				Type: util.BindOutput(rto.RtoResponse{},
					"Requested rto by property id", nil),
			},
			"getbypropertyids": &graphql.Field{
				Description: "Retrieves specified rtos by property ids",
				Args:        util.BindArguments(rto.GetByPropertyIDsRequest{}),
				Resolve:     RtoQueryGetByPropertyIDsResolver,
				Type: util.BindOutput(rto.RtosResponse{},
					"Requested rtos by property ids", nil),
			},
		},
	}))

	(*rootQueries)["rto"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "RtoMutation",
		Description: "Rto mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Rto create",
				Args:        util.BindArguments(rto.CreateRequest{}),
				Resolve:     RtoMutationCreateResolver,
				Type:        util.BindOutput(rto.RtoResponse{}, "Rto create", nil),
			},
			"createbyguest": &graphql.Field{
				Description: "Rto create by guest",
				Args:        util.BindArguments(rto.CreateByGuestRequest{}),
				Resolve:     RtoMutationCreateByGuestResolver,
				Type:        util.BindOutput(rto.RtoResponse{}, "Rto create by guest", nil),
			},
			"updatelandlord": &graphql.Field{
				Description: "Rto update landlord",
				Args:        util.BindArguments(rto.UpdateLandlordRequest{}),
				Resolve:     RtoMutationUpdateLandlordResolver,
				Type:        util.BindOutput(rto.RtoResponse{}, "Rto update landlord", nil),
			},
			"createtenancy": &graphql.Field{
				Description: "Rto create tenancy",
				Args:        util.BindArguments(rto.CreateTenancyRequest{}),
				Resolve:     RtoMutationCreateTenancyResolver,
				Type: util.BindOutput(rto.RtoResponse{},
					"Rto create tenancy", nil),
			},
			"updatetenancy": &graphql.Field{
				Description: "Rto update tenancy",
				Args:        util.BindArguments(rto.UpdateLeadTenantUserRequest{}),
				Resolve:     RtoMutationUpdateLeadTenantUserResolver,
				Type: util.BindOutput(rto.RtoResponse{},
					"Rto create tenancy", nil),
			},
			"updaterentplan": &graphql.Field{
				Description: "Rto update rent plan",
				Args:        util.BindArguments(rto.UpdateRentPlanRequest{}),
				Resolve:     RtoMutationUpdateRentPlanResolver,
				Type:        util.BindOutput(rto.RtoResponse{}, "Rto update rent plan", nil),
			},
			"updatertoenddate": &graphql.Field{
				Description: "Rto update rto end date",
				Args:        util.BindArguments(rto.UpdateEndDateRequest{}),
				Resolve:     RtoMutationUpdateRtoEndDateResolver,
				Type:        util.BindOutput(rto.RtoResponse{}, "Rto update rto end date", nil),
			},
			"renewal": &graphql.Field{
				Description: "Rto renewal",
				Args:        util.BindArguments(rto.RenewalRequest{}),
				Resolve:     RtoMutationRenewalResolver,
				Type:        util.BindOutput(rto.RtoResponse{}, "Rto renewal", nil),
			},
			"answerfromtenant": &graphql.Field{
				Description: "Rto answer from tenant",
				Args:        util.BindArguments(rto.AnswerFromTenantRequest{}),
				Resolve:     RtoMutationAnswerFromTenantResolver,
				Type: util.BindOutput(rto.RtoResponse{},
					"Rto answer from tenant", nil),
			},
		},
	}))

	(*rootMutations)["rto"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func RtoMutationAnswerFromTenantResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &rto.AnswerFromTenantRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.TenantId = user.GetAccount().GetIdValue()

	return connections.RtoServiceClient.AnswerFromTenant(params.Context, request)
}

func RtoQueryGetByPropertyIDResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &rto.GetByPropertyIDRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.UserId = user.GetAccount().GetIdValue()

	return connections.RtoServiceClient.GetByPropertyID(params.Context, request)
}

func RtoQueryGetByPropertyIDsResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &rto.GetByPropertyIDsRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.UserId = user.GetAccount().GetIdValue()

	return connections.RtoServiceClient.GetByPropertyIDs(params.Context, request)
}

func RtoMutationCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &rto.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.LandlordId = user.GetAccount().GetIdValue()

	return connections.RtoServiceClient.Create(params.Context, request)
}

func RtoMutationCreateByGuestResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &rto.CreateByGuestRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.RtoServiceClient.CreateByGuest(params.Context, request)
}

func RtoMutationUpdateLandlordResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &rto.UpdateLandlordRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.LandlordId = user.GetAccount().GetIdValue()

	return connections.RtoServiceClient.UpdateLandlord(params.Context, request)
}

func RtoMutationCreateTenancyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &rto.CreateTenancyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.LandlordId = user.GetAccount().GetIdValue()

	return connections.RtoServiceClient.CreateTenancy(params.Context, request)
}

func RtoMutationUpdateLeadTenantUserResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &rto.UpdateLeadTenantUserRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.RtoServiceClient.UpdateLeadTenantUser(params.Context, request)
}

func RtoMutationUpdateRentPlanResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &rto.UpdateRentPlanRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.LandlordId = user.GetAccount().GetIdValue()

	return connections.RtoServiceClient.UpdateRentPlan(params.Context, request)
}

func RtoMutationUpdateRtoEndDateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &rto.UpdateEndDateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.LandlordId = user.GetAccount().GetIdValue()

	return connections.RtoServiceClient.UpdateEndDate(params.Context, request)
}

func RtoMutationRenewalResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &rto.RenewalRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.LandlordId = user.GetAccount().GetIdValue()

	return connections.RtoServiceClient.Renewal(params.Context, request)
}
