package resolvers

import (
	"github.com/graphql-go/graphql"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	signature "gitlab.mashplace.com/mashplace/backend/signature/api"
)

func initSignature(rootQueries, rootMutations *graphql.Fields) {
	sigresp := util.BindOutput(signature.SignatureResponse{}, "signature response", nil)

	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "SignatureQuery",
		Description: "Signature query root",
		Fields: graphql.Fields{
			"info": &graphql.Field{
				Description: "Get signature info by signature request id",
				Args:        util.BindArguments(signature.GetInfoRequest{}),
				Resolve:     SignatureQueryInfoResolver,
				Type:        sigresp,
			},
			"pdf": &graphql.Field{
				Description: "Get pdf media object id by signature request id",
				Args:        util.BindArguments(signature.GetPDFRequest{}),
				Resolve:     SignatureQueryPdfResolver,
				Type:        util.BindOutput(signature.GetPDFResponse{}, "pdf media id", nil),
			},
		},
	}))

	(*rootQueries)["signature"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "SignatureMutation",
		Description: "Signature mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Create signature request",
				Args:        util.BindArguments(signature.CreateRequest{}),
				Resolve:     SignatureMutationCreateResolver,
				Type:        sigresp,
			},
			"cancel": &graphql.Field{
				Description: "Cancel signature request",
				Args:        util.BindArguments(signature.CancelRequest{}),
				Resolve:     SignatureMutationCancelResolver,
				Type:        util.BindOutput(signature.CancelResponse{}, "cancel status", nil),
			},
		},
	}))

	(*rootMutations)["signature"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func convertSigningLegacy(signing *signature.Signing) *signature.SignatureResponse {
	var isComplete, isDeclined, hasError bool

	switch signing.GetStatus() {
	case signature.SigningStatus_SIGNING_STATUS_COMPLETE:
		isComplete = true
	case signature.SigningStatus_SIGNING_STATUS_DECLINED:
		isDeclined = true
	case signature.SigningStatus_SIGNING_STATUS_ERROR:
		hasError = true
	}

	return &signature.SignatureResponse{
		SignatureRequestId: signing.GetSignatureRequestId(),
		IsComplete:         isComplete,
		IsDeclined:         isDeclined,
		HasError:           hasError,
		Signatures:         signing.GetSignatures(),
	}
}

func SignatureQueryInfoResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &signature.GetInfoRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.SignatureService.SigningGet(params.Context, &signature.SigningGetRequest{
		SignatureRequestId: request.GetSignatureRequestId(),
	})
	if err != nil {
		return nil, err
	}

	return convertSigningLegacy(resp.GetSigning()), nil
}

func SignatureQueryPdfResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &signature.GetPDFRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.SignatureService.SigningSignedDocument(
		params.Context,
		&signature.SigningSignedDocumentRequest{
			SignatureRequestId: request.GetSignatureRequestId(),
		},
	)
	if err != nil {
		return nil, err
	}

	return &signature.GetPDFResponse{
		Fileid: resp.GetSignedDocumentId(),
	}, nil
}

func SignatureMutationCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &signature.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.SignatureService.SigningCreate(params.Context, &signature.SigningCreateRequest{
		Title:          request.GetTitle(),
		Subject:        request.GetSubject(),
		Message:        request.GetMessage(),
		DocumentFileId: request.GetContractId(),
		Signers:        request.GetSigners(),
		Metadata: &signature.SigningMetadata{
			OfferId:    request.GetOfferId(),
			PropertyId: request.GetPropertyId(),
			ViewingId:  request.GetViewingId(),
		},
		DocumentForms: request.GetFormFieldsPerDocument(),
	})
	if err != nil {
		return nil, err
	}

	return convertSigningLegacy(resp.GetSigning()), nil
}

func SignatureMutationCancelResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &signature.CancelRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	_, err := connections.SignatureService.SigningCancel(params.Context, &signature.SigningCancelRequest{
		SignatureRequestId: request.GetSignatureRequestId(),
	})
	if err != nil {
		return nil, err
	}

	return &signature.CancelResponse{Success: true}, nil
}
