package resolvers

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/graphql/util"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/middleware"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers/errs"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	viewing "gitlab.mashplace.com/mashplace/backend/viewing/api"
)

func initViewing(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ViewingQuery",
		Description: "Viewing query root",
		Fields: graphql.Fields{
			"list": &graphql.Field{
				Description: "Lists view requests for property",
				Args:        util.BindArguments(viewing.ViewingListRequest{}),
				Resolve:     ViewingQueryListViewResolver,
				Type:        util.BindOutput(viewing.ViewingListResponse{}, "Property view listing", nil),
			},
			"reserved": &graphql.Field{
				Description: "Reserved view intervals",
				Args:        util.BindArguments(viewing.ViewingReservedRequest{}),
				Resolve:     ViewingQueryViewReservedResolver,
				Type:        util.BindOutput(viewing.ViewingReservedResponse{}, "Reserved intervals", nil),
			},
		},
	}))

	(*rootQueries)["viewing"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ViewingMutation",
		Description: "Viewing mutation root",
		Fields: graphql.Fields{
			"request": &graphql.Field{
				Description: "Request property view",
				Args: func() graphql.FieldConfigArgument {
					input := util.BindArguments(viewing.ViewingSaveRequest{})

					input["recaptcha_token"] = &graphql.ArgumentConfig{
						Type: graphql.String,
					}

					return input
				}(),
				Resolve: ViewingMutationRequestResolver,
				Type:    util.BindOutput(viewing.ViewingSaveResponse{}, "Request view status", nil),
			},
			"status": &graphql.Field{
				Description: "Cancel property view",
				Args:        util.BindArguments(property.CancelViewRequest{}),
				Resolve:     ViewingMutationStatusResolver,
				Type:        util.BindOutput(property.CancelViewResponse{}, "Cancel view status", nil),
			},
		},
	}))

	(*rootMutations)["viewing"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func ViewingQueryViewReservedResolver(params graphql.ResolveParams) (interface{}, error) {
	var request viewing.ViewingReservedRequest

	if err := util.UnmarshalArguments(params.Args, &request); err != nil {
		return nil, err
	}

	return connections.ViewingService.ViewingReserved(grpcctx.WithServiceUser(params.Context), &request)
}

func ViewingQueryListViewResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &viewing.ViewingListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.Spec == nil {
		request.Spec = &viewing.ViewingSpec{}
	}

	request.Spec.UserId = user.GetAccount().GetIdValue()

	return connections.ViewingService.ViewingList(params.Context, request)
}

func ViewingMutationRequestResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &viewing.ViewingSaveRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.ViewingService.ViewingSave(grpcctx.ValuesBackgroundContext(params.Context), request)
}

func ViewingMutationStatusResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &viewing.ViewingStatusRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.Spec == nil {
		request.Spec = &viewing.ViewingSpec{}
	}

	request.Spec.UserId = user.GetAccount().GetIdValue()

	return connections.ViewingService.ViewingStatus(grpcctx.ValuesBackgroundContext(params.Context), request)
}
