package rest

import (
	"encoding/json"
	"io"
	"net/http"

	"gitlab.mashplace.com/mashplace/backend/common/generic/env"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	insurance "gitlab.mashplace.com/mashplace/backend/insurance/api"
)

var AppliedAPIKey = env.ResolveString("APPLIED_API_KEY", "test")

func AppliedHeartbeat(w http.ResponseWriter, r *http.Request) {
	if r.Method != http.MethodGet {
		http.Error(w, "method not allowed", http.StatusMethodNotAllowed)
		return
	}

	if r.Header.Get("X-API-Key") != AppliedAPIKey {
		w.WriteHeader(http.StatusForbidden)
		return
	}

	resp, err := connections.InsuranceService.AppliedHeartbeat(r.Context(), &insurance.AppliedHeartbeatRequest{})
	if err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	w.WriteHeader(int(resp.Code))
	return
}

func AppliedSchemeConfiguration(w http.ResponseWriter, r *http.Request) {
	if r.Method != http.MethodGet {
		http.Error(w, "method not allowed", http.StatusMethodNotAllowed)
		return
	}

	if r.Header.Get("X-API-Key") != AppliedAPIKey {
		w.WriteHeader(http.StatusForbidden)
		return
	}

	resp, err := connections.InsuranceService.AppliedSchemeConfiguration(
		r.Context(),
		&insurance.AppliedSchemeConfigurationRequest{},
	)
	if err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	bs, err := json.Marshal(resp)
	if err != nil {
		log.WithError(err).WithField("resp", string(bs)).Error("AppliedSchemeConfiguration marshalling error")

		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)

	_, _ = w.Write(bs)
	return
}

func AppliedNewBusinessQuote(w http.ResponseWriter, r *http.Request) {
	if r.Method != http.MethodPost {
		http.Error(w, "method not allowed", http.StatusMethodNotAllowed)
		return
	}

	if r.Header.Get("X-API-Key") != AppliedAPIKey {
		w.WriteHeader(http.StatusForbidden)
		return
	}

	bs, err := io.ReadAll(r.Body)
	if err != nil {
		log.WithError(err).Error("reading AppliedNewBusinessQuote request body")

		w.WriteHeader(http.StatusBadRequest)
		return
	}

	req := &insurance.AppliedNewBusinessQuoteRequest{}

	err = json.Unmarshal(bs, req)
	if err != nil {
		log.WithError(err).WithField("body", string(bs)).Error("AppliedNewBusinessQuote unmarshalling error")

		w.WriteHeader(http.StatusBadRequest)
		return
	}

	if req.Id == "" || req.Risk == "" || req.RiskVariationId == "" || req.PolicyLineVersion == "" ||
		req.PolicyLine == "" || req.StartDate == "" || req.EndDate == "" {
		log.WithError(err).WithField("body", string(bs)).Error("AppliedNewBusinessQuote request body is not full")

		w.WriteHeader(http.StatusBadRequest)
		return
	}

	resp, err := connections.InsuranceService.AppliedNewBusinessQuote(
		r.Context(),
		req,
	)
	if err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	bs, err = json.Marshal(resp)
	if err != nil {
		log.WithError(err).WithField("resp", string(bs)).Error("AppliedNewBusinessQuote marshalling error")

		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)

	_, _ = w.Write(bs)
	return
}
