package rest

import (
	"encoding/json"
	"fmt"
	"io"
	"net/http"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	grabdata "gitlab.mashplace.com/mashplace/backend/grabdata/api"
)

const (
	// 20 MB
	maxRequestSize int64 = 20 * 1024 * 1024
)

func LettingAgents(w http.ResponseWriter, r *http.Request) {
	if r.Method != http.MethodPost {
		http.Error(w, "method not allowed", http.StatusMethodNotAllowed)
		return
	}

	if r.ContentLength == 0 {
		log.Error("empty request, content-length is zero")
		http.Error(w, "empty request, content-length is zero", http.StatusUnprocessableEntity)
		return
	}

	if r.ContentLength > maxRequestSize {
		msg := fmt.Sprintf("request is too big (max %d allowed)", maxRequestSize)
		log.Error(msg)
		http.Error(w, msg, http.StatusUnprocessableEntity)
		return
	}

	body, err := io.ReadAll(r.Body)
	if err != nil {
		log.WithError(err).Error("read body failed")
		http.Error(w, "read body failed", http.StatusUnprocessableEntity)
		return
	}

	stream, err := connections.GrabDataService.UploadLettingAgents(r.Context())
	if err != nil {
		log.WithError(err).Error("connect to upload service failed")
		http.Error(w, "connect to upload service failed", http.StatusInternalServerError)
		return
	}

	err = stream.Send(&grabdata.UploadRequest{Gzip: body})
	if err != nil {
		log.WithError(err).Error("send agents gzip file to upload service failed")
		http.Error(w, "send agents gzip file to upload service failed", http.StatusInternalServerError)
		return
	}

	var resp *grabdata.UploadResponse
	resp, err = stream.CloseAndRecv()
	if err != nil {
		log.WithError(err).Error("close upload service failed")
		http.Error(w, "close upload service failed", http.StatusInternalServerError)
		return
	}

	var b []byte
	b, err = json.Marshal(resp)
	if err != nil {
		log.WithError(err).Error("marshal response to json failed")
		http.Error(w, "marshal response to json failed", http.StatusInternalServerError)
		return
	}

	w.Header().Set("Content-Type", "application/json")

	_, err = w.Write(b)
	if err != nil {
		log.WithError(err).Error("write response failed")
		http.Error(w, "write response failed", http.StatusInternalServerError)
		return
	}
}

func AreaGuides(w http.ResponseWriter, r *http.Request) {
	if r.Method != http.MethodPost {
		http.Error(w, "method not allowed", http.StatusMethodNotAllowed)
		return
	}

	if r.ContentLength == 0 {
		log.Error("empty request, content-length is zero")
		http.Error(w, "empty request, content-length is zero", http.StatusUnprocessableEntity)
		return
	}

	if r.ContentLength > maxRequestSize {
		msg := fmt.Sprintf("request is too big (max %d allowed)", maxRequestSize)
		log.Error(msg)
		http.Error(w, msg, http.StatusUnprocessableEntity)
		return
	}

	stream, err := connections.GrabDataService.ImportAreaGuides(r.Context())
	if err != nil {
		log.WithError(err).Error("connect to upload service failed")
		http.Error(w, "connect to upload service failed", http.StatusInternalServerError)
		return
	}

	csvFile, err := io.ReadAll(r.Body)
	if err != nil {
		log.WithError(err).Error("read body failed")
		http.Error(w, "read body failed", http.StatusUnprocessableEntity)
		return
	}

	err = stream.Send(&grabdata.ImportAreaGuideRequest{CsvFile: csvFile})
	if err != nil {
		log.WithError(err).Error("send area guides csv file to upload service failed")
		http.Error(w, "send area guides csv file to upload service failed", http.StatusInternalServerError)
		return
	}

	var resp *grabdata.ImportAreaGuideResponse
	resp, err = stream.CloseAndRecv()
	if err != nil {
		log.WithError(err).Error("close upload service failed")
		http.Error(w, "close upload service failed", http.StatusInternalServerError)
		return
	}

	var b []byte
	b, err = json.Marshal(resp)
	if err != nil {
		log.WithError(err).Error("marshal response to json failed")
		http.Error(w, "marshal response to json failed", http.StatusInternalServerError)
		return
	}

	w.Header().Set("Content-Type", "application/json")

	_, err = w.Write(b)
	if err != nil {
		log.WithError(err).Error("write response failed")
		http.Error(w, "write response failed", http.StatusInternalServerError)
		return
	}
}
