package rest

import (
	"encoding/json"
	"net/http"
	"reflect"
	"strings"

	"gitlab.mashplace.com/mashplace/backend/common/generic/bind"
	"gitlab.mashplace.com/mashplace/backend/common/generic/wrap"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
)

type IvrResponse struct {
	Reference   string      `json:"reference"`
	IsAvailable bool        `json:"is_available"`
	RealEstate  interface{} `json:"real_estate"`
}

type IvrError struct {
	Error string `json:"error"`
}

func IvrHandler(w http.ResponseWriter, r *http.Request) {
	parts := strings.Split(r.URL.Path, "/")
	if len(parts) < 2 {
		w.WriteHeader(http.StatusBadRequest)
		bs, _ := json.Marshal(IvrError{Error: "invalid path"})
		_, _ = w.Write(bs)
		return
	}
	lasts := parts[len(parts)-2:]
	kind, value := lasts[0], lasts[1]

	var resp *property.LookupIvrResponse
	var err error
	switch kind {
	case "id":
		resp, err = connections.PropertyService.LookupIvr(r.Context(), &property.LookupIvrRequest{
			Id: wrap.StringValue(&value),
		})
	case "ivr":
		resp, err = connections.PropertyService.LookupIvr(r.Context(), &property.LookupIvrRequest{
			Ivr: wrap.StringValue(&value),
		})
	default:
		w.WriteHeader(http.StatusBadRequest)
		bs, _ := json.Marshal(IvrError{Error: "invalid id type"})
		_, _ = w.Write(bs)
		return
	}

	if err != nil {
		w.WriteHeader(http.StatusBadRequest)
		bs, _ := json.Marshal(IvrError{Error: "invalid id"})
		_, _ = w.Write(bs)
		return
	}

	structtype := reflect.TypeOf(resp.Property)
	strct := bind.GetStruct(structtype)
	s, _ := strct.GetValue(reflect.ValueOf(resp.Property), &bind.Config{})

	var det map[string]interface{}
	if maps, ok := s.Interface().(map[string]interface{}); ok {
		det = map[string]interface{}{
			"id":          maps["id"],
			"ivr":         maps["ivr"],
			"created_at":  maps["created_at"],
			"status":      maps["status"],
			"location":    maps["location"],
			"details":     maps["details"],
			"description": maps["description"],
			"rental":      maps["rental"],
		}
	}

	bs, _ := json.Marshal(IvrResponse{
		Reference:   value,
		IsAvailable: resp.GetProperty().GetStatus() == property.PropertyStatus_LISTED,
		RealEstate:  det,
	})

	w.WriteHeader(http.StatusOK)
	_, _ = w.Write(bs)
	return
}
