package rest

import (
	"context"
	"fmt"
	"net/http"
	"net/url"
	"strconv"
	"strings"

	viewing "gitlab.mashplace.com/mashplace/backend/viewing/api"

	"google.golang.org/protobuf/types/known/wrapperspb"

	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
)

func queryOrDefaultInt32Wrap(values url.Values, key string, def int32) int32 {
	vs := values[key]
	if len(vs) == 0 {
		return def
	}

	v, err := strconv.ParseInt(vs[0], 10, 32)
	if err != nil {
		return def
	}

	return int32(v)
}

type propertyViewingParam struct {
	ViewID     string
	PropertyID string
	Action     string
	Token      string
	Comment    *wrapperspb.StringValue
	Rating     *wrapperspb.Int32Value
	Values     url.Values
}

func propertyViewingParamParse(request *http.Request, prefix string) (param *propertyViewingParam, err error) {
	next := strings.TrimPrefix(request.URL.Path, prefix)
	parts := strings.Split(next, "/")

	param = &propertyViewingParam{}

	switch len(parts) {
	case 0:
		return nil, fmt.Errorf("invalid action")
	case 1:
		param.Action = parts[0]
	default:
		param.ViewID = parts[0]
		param.Action = parts[1]
	}

	param.Values = request.URL.Query()

	param.Token = param.Values.Get("token")

	if len(param.Token) == 0 {
		return nil, fmt.Errorf("invalid token")
	}

	if vs, ok := param.Values["comment"]; ok {
		if len(vs) > 0 {
			param.Comment = wrapperspb.String(vs[0])
		}
	}

	if vs, ok := param.Values["rating"]; ok {
		var v int64

		if len(vs) > 0 {
			v, err = strconv.ParseInt(vs[0], 10, 32)
			if err != nil {
				return nil, err
			}

			param.Rating = wrapperspb.Int32(int32(v))
		}
	}

	return
}

func propertyViewingHandleAction(ctx context.Context, param *propertyViewingParam) error {
	var transition viewing.ViewingState

	switch param.Action {
	case "accept":
		transition = viewing.ViewingState_VIEWING_STATE_ACCEPTED
	case "reject":
		transition = viewing.ViewingState_VIEWING_STATE_REJECTED
	case "cancel":
		transition = viewing.ViewingState_VIEWING_STATE_CANCELLED
	case "complete":
		transition = viewing.ViewingState_VIEWING_STATE_COMPLETED
	case "fail":
		transition = viewing.ViewingState_VIEWING_STATE_FAILED
	case "reschedule":
		view, err := connections.ViewingService.ViewingGet(ctx, &viewing.ViewingGetRequest{
			Spec: &viewing.ViewingSpec{
				Select: &viewing.ViewingSpecSelect{
					Token:     param.Token,
					ViewingId: param.ViewID,
				},
			},
		})
		if err != nil {
			return err
		}

		param.ViewID, param.PropertyID = view.GetViewing().GetId(), view.GetViewing().GetProperty()

		req := view.GetViewing()
		if req.Time == nil {
			req.Time = &viewing.ViewingTime{}
		}

		req.Time.Year = queryOrDefaultInt32Wrap(param.Values, "year", req.GetTime().GetYear())
		req.Time.Month = queryOrDefaultInt32Wrap(param.Values, "month", req.GetTime().GetMonth())
		req.Time.Day = queryOrDefaultInt32Wrap(param.Values, "day", req.GetTime().GetDay())
		req.Time.TimeStartsec = queryOrDefaultInt32Wrap(param.Values, "startsec", req.GetTime().GetTimeStartsec())
		req.Time.TimeEndsec = queryOrDefaultInt32Wrap(param.Values, "endsec", req.GetTime().GetTimeEndsec())

		_, err = connections.ViewingService.ViewingSave(ctx, &viewing.ViewingSaveRequest{
			ViewingId: req.GetId(),
			Token:     param.Token,
			Property:  req.GetProperty(),
			Listing:   req.GetListing(),
			Time:      req.GetTime(),
			Details:   req.GetDetails(),
			Remotely:  req.GetRemotely(),
		})
		if err != nil {
			return err
		}

		return nil
	default:
		return fmt.Errorf("unknown action")
	}

	resp, err := connections.ViewingService.ViewingStatus(ctx, &viewing.ViewingStatusRequest{
		Spec: &viewing.ViewingSpec{
			Select: &viewing.ViewingSpecSelect{
				Token:     param.Token,
				ViewingId: param.ViewID,
			},
		},
		Comment:    param.Comment.GetValue(),
		Rating:     param.Rating.GetValue(),
		Transition: transition,
	})
	if err != nil {
		return err
	}

	param.ViewID, param.PropertyID = resp.GetViewing().GetId(), resp.GetViewing().GetProperty()

	return nil
}

func PropertyViewing(prefix, success, failure string) http.HandlerFunc {
	return func(writer http.ResponseWriter, request *http.Request) {
		param, err := propertyViewingParamParse(request, prefix)
		if err != nil {
			http.Redirect(writer, request, failure+"?error="+url.QueryEscape(err.Error()), http.StatusFound)
			return
		}

		err = propertyViewingHandleAction(request.Context(), param)
		if err != nil {
			http.Redirect(writer, request, failure+"?error="+url.QueryEscape(err.Error()), http.StatusFound)
			return
		}

		target := strings.NewReplacer("$view", param.ViewID, "$property", param.PropertyID).Replace(success)

		http.Redirect(writer, request, target, http.StatusFound)
	}
}
