package rest

import (
	"context"
	"crypto/sha256"
	"errors"
	"fmt"
	"io"
	"net/http"
	"strings"

	cardpayment "gitlab.mashplace.com/mashplace/backend/cardpayment/api"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	signature "gitlab.mashplace.com/mashplace/backend/signature/api"
)

func StripeHandler(req *http.Request) (int, string, error) {
	err := req.ParseMultipartForm(MultipartMemoryBuffer)
	if err != nil {
		return Err500Result(err)
	}

	jsonList, ok := req.MultipartForm.Value["json"]
	if !ok {
		return Err500Result(err)
	}
	jsonStr := strings.Join(jsonList, "")

	log.Info(jsonStr)

	data := []byte(jsonStr)

	actual := sha256.New().Sum(data)

	stream, err := connections.CardPaymentService.HandleCallback(context.Background())
	if err != nil {
		return Err500Result(err)
	}

	err = stream.Send(&cardpayment.HandleCallbackRequest{
		Init: &cardpayment.HandleCallbackInit{
			Sha256: actual,
		},
	})
	if err != nil {
		return Err500Result(err)
	}

	for len(data) > 0 {
		var trans []byte
		if len(data) > signature.ChunkSize {
			trans = data[:signature.ChunkSize]
			data = data[signature.ChunkSize:]
		} else {
			trans = data
			data = nil
		}
		err = stream.Send(&cardpayment.HandleCallbackRequest{
			Chunk: &cardpayment.DataChunk{
				Data: trans,
			},
		})

		if err != nil {
			return Err500Result(err)
		}
	}

	err = stream.CloseSend()
	if err != nil && !errors.Is(err, io.EOF) {
		return Err500Result(err)
	}

	return http.StatusOK, "", nil
}

func StripeWebhooksHandler(w http.ResponseWriter, req *http.Request) {
	const MaxBodyBytes = int64(65536)
	req.Body = http.MaxBytesReader(w, req.Body, MaxBodyBytes)
	body, err := io.ReadAll(req.Body)
	if err != nil {
		log.Errorf("StripeWebhooksHandler error reading request body: %v\n", err)
		w.WriteHeader(http.StatusServiceUnavailable)
		return
	}

	resp, err := connections.CardPaymentService.HandleWebhook(req.Context(), &cardpayment.HandleWebhookRequest{
		StripeSignature: req.Header.Get("Stripe-Signature"),
		Data:            body,
	})

	if err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	switch resp.Status {
	case cardpayment.HttpStatusCode_OK:
		w.WriteHeader(http.StatusOK)
		return
	default:
		w.WriteHeader(http.StatusInternalServerError)
		return
	}
}

// StripeSuccessHandler - Successfully completed checkout session
func StripeSuccessHandler(w http.ResponseWriter, req *http.Request) {
	loc := fmt.Sprintf(dashboardURL, connections.SiteURL)

	if req.URL == nil {
		log.Errorln("StripeSuccessHandler empty URL")
		w.Header().Set("Location", loc)
		w.WriteHeader(http.StatusMovedPermanently)
		return
	}

	q := req.URL.Query()
	sessionID := q.Get("session_id")
	redirect := strings.TrimPrefix(q.Get("redirect"), "/")

	if redirect != "" {
		loc = fmt.Sprintf("https://%s/%s", connections.SiteURL, redirect)
	}

	if sessionID == "" {
		log.Errorln("StripeSuccessHandler session ID empty")
		w.Header().Set("Location", loc)
		w.WriteHeader(http.StatusMovedPermanently)
		return
	}

	resp, err := connections.CardPaymentService.HandleCheckout(req.Context(), &cardpayment.HandleCheckoutRequest{
		SessionId: sessionID,
		Success:   true,
	})
	if err != nil {
		log.Errorf("StripeCancelHandler error HandleCheckout: %v\n", err)
		w.Header().Set("Location", loc)
		w.WriteHeader(http.StatusMovedPermanently)
		return
	}

	if redirect == "" && resp.GetProperty() != "" {
		loc = fmt.Sprintf("https://%s/dashboard/%s/overview", connections.SiteURL, resp.GetProperty())
	}

	w.Header().Set("Location", loc)
	w.WriteHeader(http.StatusMovedPermanently)
}

// StripeCancelHandler - Interrupted checkout session
func StripeCancelHandler(w http.ResponseWriter, req *http.Request) {
	loc := fmt.Sprintf(dashboardURL, connections.SiteURL)

	if req.URL == nil {
		log.Errorln("StripeCancelHandler empty URL")
		w.Header().Set("Location", loc)
		w.WriteHeader(http.StatusMovedPermanently)
		return
	}

	q := req.URL.Query()
	sessionID := q.Get("session_id")
	redirect := strings.TrimPrefix(q.Get("redirect"), "/")

	if redirect != "" {
		loc = fmt.Sprintf("https://%s/%s", connections.SiteURL, redirect)
	}

	if sessionID == "" {
		log.Errorln("StripeCancelHandler session ID empty")
		w.Header().Set("Location", loc)
		w.WriteHeader(http.StatusMovedPermanently)
		return
	}

	resp, err := connections.CardPaymentService.HandleCheckout(req.Context(), &cardpayment.HandleCheckoutRequest{
		SessionId: sessionID,
		Success:   true,
	})
	if err != nil {
		log.Errorf("StripeCancelHandler error HandleCheckout: %v\n", err)
		w.Header().Set("Location", loc)
		w.WriteHeader(http.StatusMovedPermanently)
		return
	}

	if redirect == "" && resp.GetProperty() != "" {
		loc = fmt.Sprintf("https://%s/dashboard/%s/overview", connections.SiteURL, resp.GetProperty())
	}

	w.Header().Set("Location", loc)
	w.WriteHeader(http.StatusMovedPermanently)
}
