package rest

import (
	"context"
	"errors"
	"fmt"
	"io"
	"net/http"
	"net/url"
	"strings"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	tokenio "gitlab.mashplace.com/mashplace/backend/tokenio/api"
)

const (
	errorURL      = "https://%s/dashboard"
	dashboardURL  = "https://%s/dashboard"
	adminURL      = "https://admin.%s"
	termsDeclined = "terms declined"
)

func onError(w http.ResponseWriter, err error) {
	log.WithError(err).Error("TokenioHandler")
	w.Header().Set("Location", fmt.Sprintf(errorURL, connections.SiteURL))
	w.WriteHeader(http.StatusMovedPermanently)
	_, _ = w.Write(([]byte)(err.Error()))
}

func tokenioPaymentHandlerDetermineRedirect(
	ctx context.Context,
	q url.Values,
	resp *tokenio.PaymentHandlerResponse,
) (loc string, erred bool, err error) {
	erred = q.Get("error") != "" || q.Get("error_description") != ""

	if erred && q.Get("redirectError") != "" {
		loc = q.Get("redirectError")

		return
	} else if q.Get("redirectSuccess") != "" {
		loc = q.Get("redirectSuccess")

		return
	}

	inv, err := connections.InvoiceService.InvoiceGet(ctx, &invoice.InvoiceGetRequest{
		InvoiceId: resp.GetInvoiceId(),
	})
	if err != nil {
		log.WithError(err).Infof("TokenioPaymentHandler InvoiceService.InvoiceGet invoiceID=%+v\n", resp.GetInvoiceId())

		return
	}

	errorStr := q.Get("error_description")

	// check decline error
	if erred && strings.ToLower(errorStr) == termsDeclined {
		if inv.GetInvoice().GetCorrelationType() != "offer" {
			err = errors.New("payment declined")

			return
		}

		var offr *offer.GetResponse

		offr, err = connections.OfferService.ServiceGet(ctx, &offer.GetRequest{
			Id: inv.GetInvoice().GetCorrelationId(),
		})

		loc = fmt.Sprintf(dashboardURL, connections.SiteURL)

		if err != nil {
			log.WithError(err).
				Infof(
					"TokenioPaymentHandler OfferService.ServiceGet tResp.Offer=%+v\n",
					inv.GetInvoice().GetCorrelationId(),
				)
		} else {
			if offr.GetOffer().State == offer.OfferState_SIGNED {
				loc = fmt.Sprintf(
					dashboardURL+"/rent?offer=%s",
					connections.SiteURL,
					inv.GetInvoice().GetCorrelationId(),
				)
			} else {
				loc = fmt.Sprintf(
					dashboardURL+"/deposit?offer=%s",
					connections.SiteURL,
					inv.GetInvoice().GetCorrelationId(),
				)
			}
		}

		return
	}

	loc = fmt.Sprintf(
		dashboardURL+"/%s/overview",
		connections.SiteURL,
		inv.GetInvoice().GetProperty(),
	)

	return
}

func TokenioPaymentHandler(w http.ResponseWriter, r *http.Request) {
	var fallback string

	resp, err := connections.TokenioService.PaymentHandler(r.Context(), &tokenio.PaymentHandlerRequest{
		Data: []byte(r.URL.RawQuery),
	})
	if err != nil {
		log.WithError(err).Info("TokenioPaymentHandler TokenioService.PaymentHandler\n")

		fallback = err.Error()
	}

	q := r.URL.Query()

	loc, erred, _ := tokenioPaymentHandlerDetermineRedirect(r.Context(), q, resp)
	if loc == "" {
		loc = fmt.Sprintf(errorURL, connections.SiteURL)
	}

	erred = erred || fallback != ""

	p, err := url.Parse(loc)
	if err != nil {
		loc = fmt.Sprintf(errorURL, connections.SiteURL)

		p, _ = url.Parse(loc)
	}

	if erred {
		pq := p.Query()

		if q.Get("error") != "" {
			pq.Set("error", q.Get("error"))
		} else {
			pq.Set("error", fallback)
		}

		if q.Get("error_description") != "" {
			pq.Set("error_description", q.Get("error_description"))
		}

		if q.Get("message") != "" {
			pq.Set("message", q.Get("message"))
		}

		parts := strings.Split(q.Get("message"), ":")
		if len(parts) > 0 {
			pq.Set("error_code", parts[0])
		}

		if q.Get("request-id") != "" {
			pq.Set("request-id", q.Get("request-id"))
		}

		p.RawQuery = pq.Encode()

		loc = p.String()
	}

	w.Header().Set(
		"Location",
		loc,
	)

	w.WriteHeader(http.StatusFound)
}

func TokenioTokeninfoHandler(w http.ResponseWriter, r *http.Request) {
	body, err := io.ReadAll(r.Body)

	if len(body) == 0 && r.URL != nil {
		body = []byte(r.URL.RawQuery)
	}
	if len(body) == 0 {
		onError(w, err)
		return
	}

	resp, err := connections.TokenioService.GetTokenHandler(r.Context(), &tokenio.GetTokenHandlerRequest{
		Data: body,
	})
	if err != nil {
		onError(w, err)
		return
	}

	log.WithField("ValidTo", resp.GetValidTo()).
		WithField("Status", resp.GetStatus()).
		WithField("Token", resp.GetToken()).
		Info()
	w.Header().Set("Location", fmt.Sprintf(adminURL, connections.SiteURL))
	w.WriteHeader(http.StatusFound)
}

func TokenioWebhookHandler(w http.ResponseWriter, r *http.Request) {
	body, err := io.ReadAll(r.Body)
	if len(body) == 0 {
		log.Errorf("TokenioTokeninfoHandler error reading request body: %v\n", err)
		onError(w, err)
		return
	}

	_, err = connections.TokenioService.WebhookHandler(r.Context(), &tokenio.WebhookHandlerRequest{
		Event:     r.Header.Get("Token-Event"),
		Signature: r.Header.Get("Token-Signature"),
		Data:      body,
	})
	if err != nil {
		onError(w, err)
		return
	}

	w.WriteHeader(http.StatusOK)
}
