package rest

import (
	"context"
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"strings"

	"gitlab.mashplace.com/mashplace/backend/common/generic/env"
	"gitlab.mashplace.com/mashplace/backend/common/generic/wrap"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	twilio "gitlab.mashplace.com/mashplace/backend/twilio/api"
)

type TwilioIvrLookup struct {
	URL string `json:"URL"`
}

type TwilioCallbackResponse struct {
	Instruction         string `json:"instruction"`
	PostWorkActivitySID string `json:"post_work_activity_sid"`
	StatusCallbackURL   string `json:"status_callback_url"`
}

func TwilioIvrHandler(w http.ResponseWriter, r *http.Request) {
	body, _ := io.ReadAll(r.Body)

	if len(body) == 0 && r.URL != nil {
		body = []byte(r.URL.RawQuery)
	}

	if len(body) == 0 {
		http.Error(w, "Empty request", http.StatusBadRequest)
		return
	}

	resp, err := connections.TwilioService.IVRCallbackHandler(context.Background(), &twilio.IVRCallbackHandlerRequest{
		Data: string(body),
	})
	if err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}

	w.Header().Set("Content-Type", "application/xml")
	w.WriteHeader(http.StatusOK)

	_, _ = w.Write(([]byte)(resp.Data))

	return
}

func TwilioIvrLookupHandler(w http.ResponseWriter, r *http.Request) {
	ivr := strings.Replace(r.URL.Path, "/callback/twilio/ivr-lookup/", "", 1)
	if len(ivr) != 6 {
		w.WriteHeader(http.StatusNoContent)
		return
	}

	resp, err := connections.PropertyService.LookupIvr(r.Context(), &property.LookupIvrRequest{
		Ivr: wrap.StringValue(&ivr),
	})
	if err != nil {
		w.WriteHeader(http.StatusNotFound)
		return
	}

	url := fmt.Sprintf("https://%s/property/%s", connections.SiteURL, resp.GetProperty().GetId())

	bs, _ := json.Marshal(TwilioIvrLookup{
		URL: url,
	})

	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)

	_, _ = w.Write(bs)

	return
}

func TwilioAssignmentCallbackHandler(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)

	err := json.NewEncoder(w).Encode(&TwilioCallbackResponse{
		Instruction:         env.ResolveString("TWILIO_INSTRUCTION", "dequeue"),
		PostWorkActivitySID: env.ResolveString("TWILIO_POST_WORK_ACTIVITY_SID", "WAdc03e61ad7555b768c4ad3b9b1ba0677"),
		StatusCallbackURL:   "https://dev.mashplace.com/callback/twilio/complete-task",
	})

	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
}

func TwilioCompleteStatusHandler(w http.ResponseWriter, r *http.Request) {
	taskID := r.URL.Query().Get("TaskSid")

	resp, err := connections.TwilioService.TwillioCompleteTaskHandler(context.Background(), &twilio.TwilioCompleteTaskHandlerRequest{
		TaskID: string(taskID),
	})
	if err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}

	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)

	_, _ = w.Write(([]byte)(resp.Data))
}
