//go:generate go-bindata -pkg impl -ignore .*\.go .
//go:generate go fmt ./...
//go:generate go fix ./...

package impl

import (
	"bytes"
	"context"
	"crypto/sha1"
	"crypto/sha256"
	"encoding/hex"
	"errors"
	"html/template"
	"io"
	"math"
	"net/http"
	"os"
	"regexp"
	"strings"
	"time"

	"github.com/eientei/wsgraphql"
	"github.com/eientei/wsgraphql/mutcontext"
	"github.com/eientei/wsgraphql/proto"
	"github.com/gorilla/websocket"
	"github.com/opentracing/opentracing-go"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promauto"
	uuid "github.com/satori/go.uuid"
	auth "gitlab.mashplace.com/mashplace/backend/auth/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/generic/env"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/metrics"
	ws "gitlab.mashplace.com/mashplace/backend/common/graphql/websocket"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/connections"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/gwctx"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/resolvers"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/rest"
	"gitlab.mashplace.com/mashplace/backend/gateway/impl/schema"
	geo "gitlab.mashplace.com/mashplace/backend/geo/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
)

const (
	SitemapCacheDirEnv     = "SITEMAP_CACHE_DIR"
	SitemapCacheDirDefault = "/sitemaps"
)

type graphqlStartTimeType struct{}

var (
	graphqlStartTime  = graphqlStartTimeType{}
	regexMutation     = regexp.MustCompile(`(?i)^\s*mutation`)
	regexSubscription = regexp.MustCompile(`(?i)^\s*subscription`)
)

var SitemapCacheDir = env.ResolveString(SitemapCacheDirEnv, SitemapCacheDirDefault)
var AdminSiteURL = env.ResolveString("ADMIN_SITE_URL", "admin.dev.env.mashroom.com")

// supports $property and $view as substitutes for property id and view id
var envViewingActionSuccessURL = env.ResolveString("VIEWING_ACTION_SUCCESS_URL", "/dashboard/$property/overview")

// sets error text to "error" query-parmater
var envViewingActionFailureURL = env.ResolveString("VIEWING_ACTION_FAILURE_URL", "/error")

var page *template.Template

func init() {
	err := os.MkdirAll(connections.MediaCacheDir, os.FileMode(0755))
	if err != nil {
		panic(err)
	}

	err = os.MkdirAll(SitemapCacheDir, os.FileMode(0755))
	if err != nil {
		panic(err)
	}

	if b, err := Asset("playground.html"); err != nil {
		panic(err)
	} else {
		page = template.Must(template.New("graphql").Parse(string(b)))
	}
}

func Playground(title string, endpoint string) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		err := page.Execute(w, map[string]string{
			"title":    title,
			"endpoint": endpoint,
			"version":  "1.6.2",
		})
		if err != nil {
			http.Error(w, err.Error(), 401)
			return
		}
	}
}

func middleware(writer http.ResponseWriter, request *http.Request) {
	AccessControlAllow(writer, request)
}

func middlewareHandler(h http.Handler) http.HandlerFunc {
	return func(writer http.ResponseWriter, request *http.Request) {
		middleware(writer, request)
		if request.Method != http.MethodOptions {
			h.ServeHTTP(writer, request)
		}
	}
}

func middlewareFunc(f http.HandlerFunc) http.HandlerFunc {
	return func(writer http.ResponseWriter, request *http.Request) {
		middleware(writer, request)
		if request.Method != http.MethodOptions {
			f(writer, request)
		}
	}
}

func graphqlOperationType(query string) string {
	switch {
	case regexSubscription.MatchString(query):
		return "subscription"
	case regexMutation.MatchString(query):
		return "mutation"
	default:
		return "query"
	}
}

func extractRemote(r *http.Request) string {
	var remote string

	if h := r.Header.Get("X-Forwarded-For"); len(h) > 0 {
		remote = h
	} else {
		remote = r.RemoteAddr
	}

	if idx := strings.Index(remote, ","); idx > 0 {
		remote = remote[:idx]
	}

	if idx := strings.LastIndex(remote, ":"); idx > 0 {
		remote = remote[:idx]
	}

	return remote
}

// Serve implements gateway server
//
//nolint:gocognit, gocyclo // function has high cylomatic and cognitive complexity
func Serve(addr string) {
	s, errSchema := schema.CompileSchema()
	if errSchema != nil {
		panic(errSchema)
	}

	var metricGraphQLHandlingTime *prometheus.HistogramVec

	var metricGraphQLStartedTotal, metricGraphQLHandledTotal *prometheus.CounterVec

	if metrics.EnvMetricsEnable {
		if metrics.EnvMetricsHandlingTime {
			metricGraphQLHandlingTime = promauto.NewHistogramVec(prometheus.HistogramOpts{
				Name:    "graphql_server_handling_seconds",
				Help:    "Histogram of response latency (seconds) of graphql operation handled by the server.",
				Buckets: []float64{0.005, 0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1, 2.5, 5, 10, math.Inf(1)},
			}, []string{"operation_type", "operation_name"})
		}

		metricGraphQLStartedTotal = promauto.NewCounterVec(prometheus.CounterOpts{
			Name: "graphql_server_started_total",
			Help: "Total number of operations started on the server.",
		}, []string{"operation_type", "operation_name"})
		metricGraphQLHandledTotal = promauto.NewCounterVec(prometheus.CounterOpts{
			Name: "graphql_server_handled_total",
			Help: "Total number of operations completed on the server.",
		}, []string{"operation_type", "operation_name"})
	}

	graphqlHandler, errServe := wsgraphql.NewServer(wsgraphql.Config{
		Schema: &s,
		Upgrader: &websocket.Upgrader{
			CheckOrigin:  func(r *http.Request) bool { return true },
			Subprotocols: []string{"graphql-ws"},
		},
		OnPlainInit: func(globalctx mutcontext.MutableContext, r *http.Request, w http.ResponseWriter) {
			globalctx.Set("globalctx", globalctx)

			host := r.Host
			if strings.Contains(host, ":") {
				host = r.Host[:strings.IndexByte(r.Host, ':')]
			}

			log.Debugf("Host: [%s] AdminSiteURL: [%s]", host, AdminSiteURL)
			globalctx.Set("adminSite", strings.EqualFold(host, AdminSiteURL))
			globalctx.Set(grpcctx.CtxRemote, extractRemote(r))

			if v := r.Header.Get("Authorization"); len(v) > 0 {
				globalctx.Set(ws.CtxAccessToken, v)
			}

			if v := r.Header.Get("authToken"); len(v) > 0 {
				globalctx.Set(ws.CtxAccessToken, v)
			}

			if v := r.Header.Get("session"); len(v) > 0 {
				if uuid.FromStringOrNil(strings.TrimPrefix(v, "transient:")) != uuid.Nil {
					globalctx.Set(ws.CtxTransientUser, &ws.TransientUser{
						Session: v,
					})
					globalctx.Set(grpcctx.SessionID, v)
				}
			}
		},
		OnDisconnect: func(globalctx mutcontext.MutableContext) error {
			_ = globalctx.Cancel()

			return nil
		},
		OnConnect: func(globalctx mutcontext.MutableContext, parameters interface{}) error {
			globalctx.Set(ws.CtxUser, (*auth.AuthResponse)(nil))
			globalctx.Set(grpcctx.CtxUserID, "")
			globalctx.Set(grpcctx.CtxIsTestUser, false)
			globalctx.Set(grpcctx.CtxIsAdminUser, false)
			if parameters != nil {
				mapparams := parameters.(map[string]interface{})
				if v, ok := mapparams["authToken"].(string); ok {
					globalctx.Set(ws.CtxAccessToken, v)
				}
				if v, ok := mapparams["session"].(string); ok {
					if uuid.FromStringOrNil(strings.TrimPrefix(v, "transient:")) != uuid.Nil {
						globalctx.Set(ws.CtxTransientUser, &ws.TransientUser{
							Session: v,
						})
						globalctx.Set(grpcctx.SessionID, v)
					}
				}
			}

			if globalctx.Value(ws.CtxTransientUser) == nil {
				clientID := "transient:" + uuid.NewV4().String()

				globalctx.Set(ws.CtxTransientUser, &ws.TransientUser{
					Session: clientID,
				})
				globalctx.Set(grpcctx.SessionID, clientID)
			}

			rawtoken := globalctx.Value(ws.CtxAccessToken)
			if rawtoken == nil {
				return nil
			}

			token, ok := rawtoken.(string)
			if !ok {
				return nil
			}

			adminSite, _ := globalctx.Value("adminSite").(bool)
			resp, err := connections.AuthService.Auth(globalctx, &auth.AuthRequest{
				AccessToken: token,
				AdminSite:   adminSite,
			})
			if err != nil {
				resp = nil
			}

			tuser := resolvers.TryGetTransientUser(token)
			if tuser != nil && tuser.Session != "" {
				globalctx.Set(ws.CtxTransientUser, tuser)
				globalctx.Set(grpcctx.SessionID, tuser.Session)
			}

			globalctx.Set(ws.CtxUser, resp)
			globalctx.Set(grpcctx.CtxIsTestUser, resp.GetAccount().GetTestValue())
			globalctx.Set(grpcctx.CtxIsAdminUser, resp.GetAccount().GetAdminValue())
			globalctx.Set(grpcctx.CtxUserID, resp.GetAccount().GetIdValue())

			return nil
		},
		OnOperation: func(globalctx, opctx mutcontext.MutableContext, operation *proto.PayloadOperation) error {
			opctx.Set(gwctx.OperationContextKey, opctx)

			operationName := operation.OperationName
			operationType := graphqlOperationType(operation.Query)

			if operationName == "" {
				hs := sha1.Sum(([]byte)(operation.Query))
				operationName = hex.EncodeToString(hs[:])
			}

			span := opentracing.GlobalTracer().StartSpan(operationType + " " + operationName)

			_ = opentracing.ContextWithSpan(opctx, span)

			opctx.Set(gwctx.ActiveSpanKey, span)

			if metricGraphQLHandlingTime != nil {
				opctx.Set(graphqlStartTime, time.Now())
			}

			if metricGraphQLStartedTotal != nil {
				metricGraphQLStartedTotal.WithLabelValues(operationType, operationName).Inc()
			}

			return nil
		},
		OnOperationDone: func(globalctx, opctx mutcontext.MutableContext, operation *proto.PayloadOperation) error {
			var operationType, operationName string

			if metrics.EnvMetricsEnable {
				operationName = operation.OperationName
				operationType = graphqlOperationType(operation.Query)
			}

			if metricGraphQLHandledTotal != nil {
				metricGraphQLHandledTotal.WithLabelValues(operationType, operationName).Inc()
			}

			span := opentracing.SpanFromContext(opctx)
			if span != nil {
				span.Finish()
			}

			if metricGraphQLHandlingTime != nil {
				v, ok := opctx.Value(graphqlStartTime).(time.Time)

				if ok {
					metricGraphQLHandlingTime.WithLabelValues(operationType, operationName).
						Observe(float64(time.Since(v)) / float64(time.Second))
				}
			}

			return nil
		},
	})
	if errServe != nil {
		panic(errServe)
	}

	http.HandleFunc("/", middlewareHandler(Playground("GraphQL playground", "/query")))
	http.HandleFunc("/query", middlewareHandler(graphqlHandler))
	http.HandleFunc("/media", middlewareFunc(MediaUploader))
	http.HandleFunc("/media/", middlewareFunc(MediaID))
	http.HandleFunc("/media/bundle", middlewareFunc(MediaBundle))
	http.HandleFunc("/sitemaps/", middlewareFunc(SitemapID))
	http.HandleFunc("/callback/hellosign", middlewareHandler(rest.AppHandler(rest.HellosignHandler)))
	http.HandleFunc("/callback/stripe", middlewareHandler(rest.AppHandler(rest.StripeHandler)))
	http.HandleFunc("/callback/stripe/webhooks", middlewareFunc(rest.StripeWebhooksHandler))
	http.HandleFunc("/callback/stripe/success", middlewareFunc(rest.StripeSuccessHandler))
	http.HandleFunc("/callback/stripe/cancel", middlewareFunc(rest.StripeCancelHandler))
	http.HandleFunc("/api/property/lookup/", middlewareFunc(rest.IvrHandler))
	http.HandleFunc("/callback/twilio", middlewareFunc(rest.TwilioIvrHandler))
	http.HandleFunc("/callback/twilio/queue", middlewareFunc(rest.TwilioAssignmentCallbackHandler))
	http.HandleFunc("/callback/twilio/complete-task", middlewareFunc(rest.TwilioCompleteStatusHandler))
	http.HandleFunc("/callback/twilio/ivr-lookup/", middlewareFunc(rest.TwilioIvrLookupHandler))
	http.HandleFunc("/callback/tokenio-payment", middlewareFunc(rest.TokenioPaymentHandler))
	http.HandleFunc("/callback/tokenio-tokeninfo", middlewareFunc(rest.TokenioTokeninfoHandler))
	http.HandleFunc("/callback/tokenio-webhook", middlewareFunc(rest.TokenioWebhookHandler))
	http.HandleFunc("/callback/plaid-webhook", middlewareFunc(rest.PlaidWebhookHandler))
	http.HandleFunc("/callback/veriff/decision", middlewareFunc(rest.VeriffHandlerResult))
	http.HandleFunc("/callback/veriff/event", middlewareFunc(rest.VeriffEventHandlerResult))
	http.HandleFunc("/callback/rentprofile/", middlewareHandler(
		http.StripPrefix("/callback/rentprofile", rest.RentprofileHandler)),
	)
	http.HandleFunc("/action/property/viewing/", middlewareFunc(
		rest.PropertyViewing(
			"/action/property/viewing/",
			envViewingActionSuccessURL,
			envViewingActionFailureURL,
		)))
	http.HandleFunc("/api/grabdata/letting_agents", middlewareFunc(rest.LettingAgents))
	http.HandleFunc("/api/grabdata/area_guides", middlewareFunc(rest.AreaGuides))

	http.HandleFunc("/api/applied/heartbeat", middlewareFunc(rest.AppliedHeartbeat))
	http.HandleFunc("/api/applied/schemeConfiguration", middlewareFunc(rest.AppliedSchemeConfiguration))
	http.HandleFunc("/api/applied/newbusiness/quote", middlewareFunc(rest.AppliedNewBusinessQuote))

	go func() {
		log.Infof("Starting metrics exporter on %s%s", metrics.EnvMetricsListen, metrics.EnvMetricsPath)

		_ = metrics.ListenAndServe()
	}()

	log.Fatal(http.ListenAndServe(addr, nil))
}

func ID2Filepath(id string) string {
	var path []rune
	if len(id) == 64 {
		for i := 0; i < 64; i += 4 {
			path = append(path,
				'/',
				rune(id[i]),
				rune(id[i+1]),
				rune(id[i+2]),
				rune(id[i+3]),
			)
		}
	} else {
		for i := 0; i < 70; i += 5 {
			path = append(path,
				'/',
				rune(id[i]),
				rune(id[i+1]),
				rune(id[i+2]),
				rune(id[i+3]),
				rune(id[i+4]),
			)
		}
	}
	return string(path)
}

func SitemapID(writer http.ResponseWriter, request *http.Request) {
	id := strings.Replace(request.URL.Path, "/sitemaps/", "", 1)
	if len(id) == 0 {
		writer.WriteHeader(400)
		return
	}

	if request.Method == http.MethodGet {
		SitemapDownloader(writer, request)
	}
}

// SitemapDownloader - is ...
//
//nolint:gocyclo, gocognit
func SitemapDownloader(writer http.ResponseWriter, request *http.Request) {
	dynamicSitemap := false
	err := os.MkdirAll(SitemapCacheDir, os.FileMode(0755))
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	fileName := strings.Replace(request.URL.Path, "/sitemaps/", "", 1)
	path := SitemapCacheDir + "/" + fileName
	contentType := "application/xml"
	if strings.HasSuffix(fileName, ".gz") {
		contentType = "application/x-gzip"
	}

	switch fileName {
	case "sitemapgeneral.xml.gz", "sitemapgeneral.xml", "sitemaparticles.xml.gz", "sitemaparticles.xml":
		dynamicSitemap = true
	}

	_, err = os.Stat(path)
	if os.IsNotExist(err) || dynamicSitemap {
		success := false
		client, err := connections.GeoService.SitemapDownload(
			request.Context(),
			&geo.SitemapDownloadRequest{FileName: fileName},
		)
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
		ev, err := client.Recv()
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
		if ev.Init == nil {
			http.Error(writer, "missing init", http.StatusInternalServerError)
			return
		}
		init := ev.Init
		f, err := os.OpenFile(path, os.O_RDWR|os.O_CREATE|os.O_TRUNC, os.FileMode(0644))
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
		defer func() {
			_ = f.Close()
			if !success {
				_ = os.Remove(path)
			}
		}()
		h := sha256.New()
		for {
			ev, err := client.Recv()
			if errors.Is(err, io.EOF) {
				break
			}
			if err != nil {
				http.Error(writer, err.Error(), http.StatusInternalServerError)
				return
			}
			if ev.Data == nil {
				http.Error(writer, "missing data", http.StatusInternalServerError)
				return
			}

			_, err = f.Write(ev.Data.Data)
			if err != nil {
				http.Error(writer, err.Error(), http.StatusInternalServerError)
				return
			}

			_, err = h.Write(ev.Data.Data)
			if err != nil {
				http.Error(writer, err.Error(), http.StatusInternalServerError)
				return
			}
		}

		digest := h.Sum(nil)
		if !bytes.Equal(digest, init.Sha256) {
			http.Error(writer, "digest mismatch", http.StatusInternalServerError)
			return
		}
		success = true
	}

	writer.Header().Set("Content-Type", contentType)
	http.ServeFile(writer, request, path)
}

func AccessControlAllow(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Vary", "Authorization, authToken, Origin")
	w.Header().Set("Content-Type", "application/json")

	headers := []string{
		"Authorization",
		"authToken",
		"Origin",
		"Authorization",
		"authToken",
		"Access-Control-Request-Headers",
		"Access-Control-Request-Method",
		"X-Forwarded-For",
		"X-Forwarded-Port",
		"X-Forwarded-Proto",
		"X-Requested-With",
		"Upgrade",
		"Connection",
		"Via",
		"Media-Range",
		"Optional",
		"If",
		"If-Match",
		"If-Modified-Since",
		"If-None-Match",
		"If-Range",
		"If-Unmodified-Since",
		"Keep-Alive",
		"Last-Modified",
		"MIME-Version",
		"ETag",
		"Accept",
		"Accept-CH",
		"Accept-Charset",
		"Accept-Datetime",
		"Accept-Encoding",
		"Accept-Ext",
		"Accept-Features",
		"Accept-Language",
		"Accept-Params",
		"Accept-Ranges",
		"User-Agent",
		"*",
	}
	if req := r.Header.Get("Access-Control-Request-Headers"); req != "" {
		found := false
		for _, m := range headers {
			if strings.EqualFold(m, req) {
				found = true
				break
			}
		}
		if !found {
			headers = append(headers, req)
		}
	}
	w.Header().Set("Access-Control-Allow-Headers", strings.Join(headers, ", "))

	methods := []string{
		"GET",
		"HEAD",
		"POST",
		"PUT",
		"DELETE",
		"CONNECT",
		"OPTIONS",
		"TRACE",
		"PATCH",
		"*",
	}
	if req := r.Header.Get("Access-Control-Request-Method"); req != "" {
		found := false
		for _, m := range methods {
			if strings.EqualFold(m, req) {
				found = true
				break
			}
		}
		if !found {
			methods = append(methods, req)
		}
	}
	w.Header().Set("Access-Control-Allow-Methods", strings.Join(methods, ", "))

	if origin := r.Header.Get("Origin"); origin != "" {
		w.Header().Set("Access-Control-Allow-Origin", origin)
	}
}

func updateDraftContract(ctx context.Context, draftContractID string) (err error) {
	var listFilterResponse *offer.ListFilterResponse
	listFilterResponse, err = connections.OfferService.ListFilter(ctx, &offer.ListFilterRequest{
		Params: &offer.ListFilterRequest_ParamsValue{
			ParamsValue: &query.QueryParams{
				Filter: []*query.QueryParamsFilter{
					{
						Fields:   []string{"draft_contract"},
						Value:    []string{draftContractID},
						Operator: query.QueryFilterOperator_EQUALS,
					},
				},
			},
		},
	})
	if err != nil {
		return err
	}

	if len(listFilterResponse.GetOffers()) == 0 {
		return nil // it is not draft contract
	}

	of := listFilterResponse.GetOffers()[0]
	if of.State != offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES {
		return nil // the draft contract can only be updated if the offer state is CONTRACT_REVIEW_AND_CLAUSES
	}

	_, err = connections.OfferService.UpdateDraftContract(ctx, &offer.UpdateDraftContractRequest{
		Offer: of.GetId(),
	})
	if err != nil {
		return err
	}

	return nil
}
